package jp.riken.brain.ni.samuraigraph.application;

import java.awt.Window;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.awt.dnd.DnDConstants;
import java.awt.dnd.DropTargetDropEvent;
import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import jp.riken.brain.ni.samuraigraph.application.SGDataCreator.FileColumn;
import jp.riken.brain.ni.samuraigraph.base.SGBufferedFileReader;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfoSet;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.data.SGDataTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGIDataColumnTypeConstants;

import org.w3c.dom.DOMImplementation;

import ucar.nc2.NetcdfFile;

/**
 * 
 * This class provides static methods for the application.
 */
public class SGApplicationUtility implements SGIApplicationTextConstants, SGIDataColumnTypeConstants,
	SGIImageConstants {

    /**
     * Delete all files recursively.
     * 
     * @param f
     *            The abstract pathname to be deleted.
     */
    public static void deleteRecursively(final File f) {
        if (f.isDirectory()) {
            String fs = SGIConstants.FILE_SEPARATOR;
            String[] fList = f.list();
            StringBuffer sb = new StringBuffer();
            sb.append(f.toString());
            sb.append(fs);
            final int len = sb.length();
            for (int ii = 0; ii < fList.length; ii++) {
                sb.append(fList[ii].toString());
                File f_ = new File(sb.toString());
                deleteRecursively(f_);
                sb.setLength(len);
            }
        }
        f.delete();
    }

    /**
     * Delete all files recursively when the virtual machine terminates.
     * 
     * @param f
     *            The abstract pathname to be deleted.
     */
    public static void deleteOnExitRecursively(final File f) {
        if (f.isDirectory()) {
            String fs = SGIConstants.FILE_SEPARATOR;
            String[] fList = f.list();
            StringBuffer sb = new StringBuffer();
            sb.append(f.toString());
            sb.append(fs);
            final int len = sb.length();
            for (int ii = 0; ii < fList.length; ii++) {
                sb.append(fList[ii].toString());
                File f_ = new File(sb.toString());
                deleteOnExitRecursively(f_);
                sb.setLength(len);
            }
        }
        f.deleteOnExit();
    }

    /**
     * Copy file from f1 to f2.
     * 
     * @param f1
     * @param f2
     * @return true:success, false:failure
     */
    public static boolean copyBinaryFile(final File f1, final File f2)
            throws IOException {
        if (f1 == null || f2 == null) {
            return false;
        }

        FileInputStream in = null;
        FileOutputStream out = null;
        try {
            in = new FileInputStream(f1);
            out = new FileOutputStream(f2);
        } catch (FileNotFoundException ex) {
            if (in != null) {
                in.close();
            }
            if (out != null) {
                out.close();
            }
            return false;
        }

        BufferedInputStream bis = new BufferedInputStream(in);
        BufferedOutputStream bos = new BufferedOutputStream(out);

        final int bufferLength = 1024;
        byte[] buffer = new byte[bufferLength];
        int len;
        while ((len = bis.read(buffer, 0, bufferLength)) != -1) {
            bos.write(buffer, 0, len);
        }

        // close I/O streams
        bis.close();
        bos.close();

        return true;
    }

    /**
     * Copy file from f1 to f2.
     * 
     * @param f1
     * @param f2
     * @return true:success, false:failure
     */
    public static boolean copyBinaryFile(final String f1, final String f2)
            throws IOException {
        if (f1 == null || f2 == null) {
            return false;
        }
        return copyBinaryFile(new File(f1), new File(f2));
    }

    /**
     * Returns a DOMImplementation instance.
     * 
     * @return
     */
    public static DOMImplementation getDOMImplementation() {
        DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        DocumentBuilder builder = null;
        try {
            builder = factory.newDocumentBuilder();
        } catch (ParserConfigurationException ex) {
            return null;
        }
        DOMImplementation domImpl = builder.getDOMImplementation();

        return domImpl;
    }

    /**
     * Find files with the given name.
     * 
     * @param baseDir -
     *            base directory to start to search
     * @param name -
     *            the file name to search
     * @param list -
     *            a list to add the found files
     */
    public static void findFiles(final File baseDir, final String name,
            final ArrayList list) {
        if (baseDir.isDirectory() == false) {
            return;
        }

        File[] fArray = baseDir.listFiles();
        for (int ii = 0; ii < fArray.length; ii++) {
            if (fArray[ii].isDirectory()) {
                findFiles(fArray[ii], name, list);
            } else {
                String name_ = fArray[ii].getName();
                if (name_.equals(name)) {
                    list.add(fArray[ii].getAbsolutePath());
                }
            }
        }

    }

    /**
     * Check whether one version is released later than the other version.
     * 
     * @param m1
     * @param m2
     * @param m3
     * @param n1
     * @param n2
     * @param n3
     * @return whether the version n is released later than version m
     */
    public static boolean compareVersionNumber(final int m1, final int m2,
            final int m3, final int n1, final int n2, final int n3) {
        boolean b;
        if (m1 < n1) {
            b = true;
        } else if (m1 == n1) {
            if (m2 < n2) {
                b = true;
            } else if (m2 == n2) {
                b = (m3 < n3);
            } else {
                b = false;
            }
        } else {
            b = false;
        }

        return b;
    }

    public static final void showDataFileInvalidMessageDialog(final Window owner) {
        SGUtility.showErrorMessageDialog(owner, MSG_DATA_FILE_INVALID,
                TITLE_ERROR);
    }

    public static final void showDataTypeInvalidMessageDialog(final Window owner) {
        SGUtility.showErrorMessageDialog(owner, MSG_DATA_TYPE_INVALID,
                TITLE_ERROR);
    }

    /**
     * Returns a list of dropped files.
     * @param dtde
     *             a drop event
     * @return
     *             A file list. If some error occur, returns null.
     */
    public static List getDroppedFileList(final DropTargetDropEvent dtde) {
        List fileList = new ArrayList();
        dtde.acceptDrop(DnDConstants.ACTION_COPY_OR_MOVE);
        try {
            if ((dtde.getDropAction() & DnDConstants.ACTION_COPY_OR_MOVE) != 0) {
                Transferable trans = dtde.getTransferable();
                if (dtde.isDataFlavorSupported(DataFlavor.javaFileListFlavor)) {
                    List lst = (List) trans.getTransferData(DataFlavor.javaFileListFlavor);
                    fileList.addAll(lst);
                } else if (dtde.isDataFlavorSupported(DataFlavor.stringFlavor)) {
                    String str;
                    String ss = (String) trans
                            .getTransferData(DataFlavor.stringFlavor);
                    StringTokenizer st = new StringTokenizer(ss, "\n");
                    while (st.hasMoreTokens()) {
                        str = st.nextToken();
                        File file = new File(new URI(str.trim()));
                        fileList.add(file);
                    }
                }
            }
        } catch (Exception ex) {
            // notify drop failed
            dtde.dropComplete(false);
            return null;
        }

        // notify drop succeeded
        dtde.dropComplete(true);

        return fileList;
    }
    
    /**
     * Create a file path name.
     * @param parent
     *              The parent path name string
     * @param child
     *              The child path name string
     * @return
     *              The full path name string
     */
    public static String getPathName(final String parent, final String child) {
        StringBuffer sb = new StringBuffer();
        sb.append(parent);
        sb.append(SGIConstants.FILE_SEPARATOR);
        sb.append(child);
        String path = sb.toString();
        return path;
    }

    /**
     * Append the extension to the file name.
     * @param filename
     *                 The file name
     * @param extension
     *                 The extension
     * @return
     *                 The file name with extension
     */
    public static String appendExtension(final String filename, final String extension) {
        StringBuffer sb = new StringBuffer();
        sb.append(filename);
        sb.append('.');
        sb.append(extension);
        String path = sb.toString();
        return path;
    }

    /**
     * Returns whether a given file name has the given extension.
     * 
     * @param fileName
     *           the file name
     * @param extension
     *           the extension
     * @return true if a given file name has the given extension
     */
    public static boolean hasExtension(final String fileName, final String extension) {
    	return fileName.toLowerCase().endsWith("." + extension.toLowerCase());
    }
    
    /**
     * Returns whether a file with given location is a netCDF file.
     * 
     * @param location
     * @return
     */
    public static boolean isNetCDFFile(final String location) {
        try {
            boolean isNetCDFFile = NetcdfFile.canOpen(location);
            return isNetCDFFile;
        } catch (Exception e) {
            return false;
        }
    }

    static SGDataColumnInfoSet getArrayDataColInfoSet(FileColumn[] fileCols,
			String dataType, Map<String, Object> infoMap) {

		final int colNum = fileCols.length;
		SGDataColumnInfo[] colInfoArray = new SGDataColumnInfo[colNum];
		for (int ii = 0; ii < colNum; ii++) {
			colInfoArray[ii] = new SGDataColumnInfo(fileCols[ii].title,
					fileCols[ii].valueType);
		}
		SGDataColumnInfo[] aColInfoArray = getAdditionalInfoArray(dataType,
				infoMap, colInfoArray);

		SGDataColumnInfo[] allInfoArray = new SGDataColumnInfo[colInfoArray.length + aColInfoArray.length];
		for (int ii = 0; ii < colInfoArray.length; ii++) {
			allInfoArray[ii] = colInfoArray[ii];
		}
		for (int ii = 0; ii < aColInfoArray.length; ii++) {
			allInfoArray[ii + colInfoArray.length] = aColInfoArray[ii];
		}
		
		SGDataColumnInfoSet colInfoSet = new SGDataColumnInfoSet(allInfoArray);
		return colInfoSet;
	}

    static SGDataColumnInfo[] getAdditionalInfoArray(final String dataType, 
    		final Map<String, Object> infoMap, final SGDataColumnInfo[] colInfoArray) {

    	List<SGDataColumnInfo> aColInfoList = new ArrayList<SGDataColumnInfo>();
    	
		// create a column for the sampling rate
		Double samplingRate = (Double) infoMap.get(KEY_SAMPLING_RATE);
		if (samplingRate != null) {
			// create column info for sampling
			SGDataColumnInfo samplingRateColumnInfo = new SGDataColumnInfo(
					SGDataUtility.createSamplingRateTitle(samplingRate
							.doubleValue()),
					SGIDataColumnTypeConstants.VALUE_TYPE_SAMPLING_RATE);

			// set X_VALUE to the column type
			samplingRateColumnInfo.setColumnType(X_VALUE);

			// add to the list
			aColInfoList.add(samplingRateColumnInfo);
		}
		
//		// create a column for the tick label of the date
//		List<Integer> dateIndexList = new ArrayList<Integer>();
//		for (int ii = 0; ii < colInfoArray.length; ii++) {
//			final String valueType = colInfoArray[ii].getValueType();
//			if (VALUE_TYPE_DATE.equals(valueType)) {
//				dateIndexList.add(Integer.valueOf(ii));
//			}
//		}
//		for (int ii = 0; ii < dateIndexList.size(); ii++) {
//			Integer index = dateIndexList.get(ii);
//			String title = "Labels for date column of No." + (index + 1);
//			SGDataColumnInfo dateColumnInfo = new SGDataColumnInfo(title, 
//					SGIDataColumnTypeConstants.VALUE_TYPE_TEXT);
//			aColInfoList.add(dateColumnInfo);
//		}

		// create an array of columns
		SGDataColumnInfo[] addInfoArray = new SGDataColumnInfo[aColInfoList.size()];
		addInfoArray = aColInfoList.toArray(addInfoArray);

		return addInfoArray;
	}

    /**
     * Return the replaced data type of text data.
     * <p>
     * Replace the data type SXY_SAMPLING to SXY_MULTIPLE and SXY_DATE to SXY.
     * 
     * @param dataType
     * @return replaced data type
     */
    static String getArrayDataType(final String dataType) {
        String type = dataType;
        // replace the data type
        if (SGDataTypeConstants.SXY_SAMPLING_DATA.equals(dataType)) {
            type = SGDataTypeConstants.SXY_MULTIPLE_DATA;
        } else if (SGDataTypeConstants.SXY_DATE_DATA.equals(dataType)) {
            type = SGDataTypeConstants.SXY_DATA;
        }
        return type;

    }
    
    /**
     * Return exact string that is one of the specified types.
     * 
     * @param types type strings
     * @param str string which contains not letter or digit characters.
     * @return exact string that is one of the specified types. or null.
     */
    static String getExactTypeString(final String[] types, final String str) {
        if (null==str) {
            return null;
        }
        if (types.length==0) {
            return null;
        }
        for (int i = 0; i < types.length; i++) {
            if (SGUtilityText.isEqualString(types[i], str)) {
                return types[i];
            }
        }
        return null;
    }
    
    /**
     * Creates a byte array from given file object.
     * 
     * @param f
     *          the file
     * @return a byte array
     */
    public static byte[] toByteArray(final File f) {
		byte[] byteArray = null;
		ByteArrayOutputStream bos = null;
		BufferedInputStream bis = null;
		FileInputStream fis = null;
		try {
			// read image byte data
			try {
				fis = new FileInputStream(f);
			} catch (FileNotFoundException e) {
				return null;
			} finally {
			}
			bis = new BufferedInputStream(fis);
			bos = new ByteArrayOutputStream();
			byte[] buf = new byte[1024];
			while (bis.read(buf) != -1) {
				bos.write(buf);
			}

			// get byte data
			byteArray = bos.toByteArray();

		} catch (IOException e) {
			return null;
		} finally {
			if (fis != null) {
				try {
					fis.close();
				} catch (IOException e) {
				}
			}
			if (bis != null) {
				try {
					bis.close();
				} catch (IOException e) {
				}
			}
			if (bos != null) {
				try {
					bos.close();
				} catch (IOException e) {
				}
			}
		}
		return byteArray;
    }

    /**
     * Returns the extension of an image file.
     * 
     * @param f
     *          image file
     * @return the extension of an image file
     */
    public static String getExtension(final File f) {
    	final String path = f.getPath();
    	final int index = path.lastIndexOf('.');
    	if (index == -1) {
    		return null;
    	}
    	if (index == path.length() - 1) {
    		return null;
    	}
    	final String ext = path.substring(index + 1);
    	return findExtension(ext);
    }

    /**
     * Finds and returns the extension of an image file.
     * 
     * @param str
     *            a text string to compare
     * @return a text string of the extension
     */
    public static String findExtension(final String str) {
    	for (int ii = 0; ii < DRAWABLE_IMAGE_EXTENSIONS.length; ii++) {
    		if (DRAWABLE_IMAGE_EXTENSIONS[ii].equalsIgnoreCase(str)) {
    			return DRAWABLE_IMAGE_EXTENSIONS[ii];
    		}
    	}
    	return null;
    }
}
