package jp.riken.brain.ni.samuraigraph.application;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.text.ParseException;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGBufferedFileReader;
import jp.riken.brain.ni.samuraigraph.base.SGCSVTokenizer.Token;
import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfoSet;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGIDataSource;
import jp.riken.brain.ni.samuraigraph.base.SGIProgressControl;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.data.SGDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGDataSourceObserver;
import jp.riken.brain.ni.samuraigraph.data.SGDataTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGDate;
import jp.riken.brain.ni.samuraigraph.data.SGDateDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGDateVariable;
import jp.riken.brain.ni.samuraigraph.data.SGIDataColumnTypeConstants;
import jp.riken.brain.ni.samuraigraph.data.SGISXYTypeData;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFFile;
import jp.riken.brain.ni.samuraigraph.data.SGNumberDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGSXYData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleDimensionNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYNetCDFDateData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYZData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYZNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSamplingDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGTextDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGTextDataFile;
import jp.riken.brain.ni.samuraigraph.data.SGVXYData;
import jp.riken.brain.ni.samuraigraph.data.SGVXYNetCDFData;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFile;

/**
 * A class to create data objects.
 */
public class SGDataCreator implements SGIConstants, SGIDataColumnTypeConstants {

    private static final String PROGRESS_MESSAGE_READFILE = "Read File";

    private static final String PROGRESS_MESSAGE_CREATEDATA = "Create Data";

    /**
     * A data source observer.
     */
    private SGDataSourceObserver mDataSourceObserver = new SGDataSourceObserver();

    /**
     * The default constructor.
     * 
     */
    public SGDataCreator() {
        super();
    }

    /**
     * Creates a data object.
     * 
     * @param path
     *            the path name of the file
     * @param colInfoSet
     *            a set of column information
     * @param infoMap
     *            information of data
     * @param progress
     *            the progress monitor
     * @param versionNumber
     *            the versionNumber of the property file if a property file is used
     * @param mode
     *            the mode of loading properties
     * @return a set of created data. null if failed.
     */
    public CreatedDataSet create(final String path, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap, final SGIProgressControl progress,
            final String versionNumber, final int mode) throws FileNotFoundException {

        progress.setIndeterminateProgress(true);
    	CreatedDataSet cdSet = null;
        progress.startProgress();
        try {
            if (path == null || infoMap == null) {
            	return null;
            }

            // when information is empty, returns null
            if (infoMap.size() == 0) {
                return null;
            }

            SGIDataSource dataSource = null;
            final String dataType = (String) infoMap.get(KEY_DATA_TYPE);
            if (SGDataUtility.isArrayData(dataType)) {
				
                progress.setProgressMessage(PROGRESS_MESSAGE_READFILE);

                // read text data from file
                List<String>[] listArray = this.createListArray(path);
                if (listArray == null) {
                    return null;
                }

                // allocate memory
                progress.setProgressMessage(PROGRESS_MESSAGE_CREATEDATA);

                // create columns
                final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();
                if (colInfo.length < listArray.length) {
                    return null;
                }
                SGDataColumn[] columns = this.createColumnArray(colInfo, listArray, infoMap);
                if (columns == null) {
                    return null;
                }
		        
                if (versionNumber != null && mode == LOAD_PROPERTIES_FROM_DATA_SET) {
                    // if version of the property file is older than 2.0.0
                    if ("".equals(versionNumber)) {
                    	// if columns contains that of error bars, swap them
                    	// because old versions of Samurai Graph exports data columns 
                    	// into a data set file in reverse order
                    	int lowerIndex = -1;
                    	int upperIndex = -1;
                    	for (int ii = 0; ii < colInfo.length; ii++) {
                    		String columnType = colInfo[ii].getColumnType();
                    		if (columnType.startsWith(LOWER_ERROR_VALUE)) {
                    			lowerIndex = ii;
                    		}
                    		if (columnType.startsWith(UPPER_ERROR_VALUE)) {
                    			upperIndex = ii;
                    		}
                    	}
                    	if (lowerIndex != -1 && upperIndex != -1) {
                    		SGDataColumn tmp = columns[lowerIndex];
                    		columns[lowerIndex] = columns[upperIndex];
                    		columns[upperIndex] = tmp;
                    	}
                    }
                }

                // create data source
                dataSource = new SGTextDataFile(path, columns);

            } else if (SGDataUtility.isNetCDFData(dataType)) {

                NetcdfFile ncFile = null;
                try {
                        ncFile = NetcdfFile.open(path);
                } catch (IOException e) {
                    return null;
                }

                // create data source
                dataSource = new SGNetCDFFile(ncFile);

            } else {
                return null;
            }
			
            cdSet = this.createData(dataSource, colInfoSet, infoMap, progress);
			
        } finally {
            progress.endProgress();
        }
        
        return cdSet;
    }
    
    /**
     * A created data object.
     *
     * This has data and its title.
     */
    public static class CreatedData {
        private SGData mData = null;
        private String mTitle = null;

        /**
         * Creates a data object with given identifier.
         * 
         * @param data
         *            the data
         * @param title
         *            the title
         */
        public CreatedData(final SGData data, final String title) {
            super();
            if (data == null) {
                throw new IllegalArgumentException("data == null");
            }
            this.mTitle = title;
            this.mData = data;
        }
        public String getTitle() {
            return this.mTitle;
        }
        public SGData getData() {
            return this.mData;
        }
    }
    
    public static class CreatedDataSet {
        private List<CreatedData> mDataList = new ArrayList<CreatedData>();
        private String mMiddleString = "\\_";
        
        public CreatedDataSet() {
            super();
        }
        public boolean addData(final SGData data, final String title) {
            this.mDataList.add(new CreatedData(data, title));
            return true;
        }
        public boolean addData(final SGData data) {
            return this.addData(data, null);
        }
        public int getDataLength() {
            return this.mDataList.size();
        }
        public CreatedData getData(final int n) {
            if (n < 0 || n >= this.mDataList.size()) {
                throw new IllegalArgumentException("Index out of bounds: " + n);
            }
            return this.mDataList.get(n);
        }
        public String getMiddleString() {
            return this.mMiddleString;
        }
        public void setMiddleString(final String middleString) {
            this.mMiddleString = middleString;
        }
    }

    /**
     * Create an array of data column objects.
     */
    private SGDataColumn[] createColumnArray(SGDataColumnInfo[] colInfo,
    		List<String>[] listArray, Map<String, Object> infoMap) {
    	
    	// get data type
        final String dataType = (String) infoMap.get(KEY_DATA_TYPE);
        
        // check data length
        final int len = listArray[0].size();
        for (int ii = 1; ii < listArray.length; ii++) {
            if (listArray[ii].size() != len) {
            	return null;
            }
        }
        
        SGDataColumn[] columns = new SGDataColumn[colInfo.length];
        for (int ii = 0; ii < columns.length; ii++) {
            String title = colInfo[ii].getTitle();
            String valueType = colInfo[ii].getValueType();
            if (SGUtilityText.isEqualString(VALUE_TYPE_NUMBER, valueType)) {
                List<String> valueList = listArray[ii];
                if (SGDataUtility.hasTickLabels(dataType)) {
                    for (int jj = 0; jj < len; jj++) {
                        String value = valueList.get(jj);
                        Double d = SGUtilityText.getDouble(value);
                        if (d == null) {
                            return null;
                        }
                    }
                    String[] sArray = new String[len];
                    for (int jj = 0; jj < len; jj++) {
                        sArray[jj] = valueList.get(jj);
                    }
                    columns[ii] = new SGNumberDataColumn(title, sArray);
                } else {
                    double[] dArray = new double[len];
                    for (int jj = 0; jj < len; jj++) {
                        String value = valueList.get(jj);
                        Double d = SGUtilityText.getDouble(value);
                        if (d == null) {
                            return null;
                        }
                        dArray[jj] = d.doubleValue();
                    }
                    columns[ii] = new SGNumberDataColumn(title, dArray);
                }
            } else if (SGUtilityText.isEqualString(VALUE_TYPE_TEXT, valueType)) {
                List<String> valueList = listArray[ii];
                String[] sArray = new String[len];
                for (int jj = 0; jj < len; jj++) {
                    sArray[jj] = valueList.get(jj);
                }
                columns[ii] = new SGTextDataColumn(title, sArray);
            } else if (SGUtilityText.isEqualString(VALUE_TYPE_DATE, valueType)) {
                List<String> valueList = listArray[ii];
                SGDate[] dArray = new SGDate[len];
                for (int jj = 0; jj < len; jj++) {
                    String value = valueList.get(jj);
                    try {
                        dArray[jj] = new SGDate(value);
                    } catch (ParseException e) {
                        return null;
                    }
                }
                columns[ii] = new SGDateDataColumn(title, dArray);
            } else if (SGUtilityText.isEqualString(VALUE_TYPE_SAMPLING_RATE, valueType)) {
                Number num = this.getSamplingRate(infoMap);
                if (num == null) {
                	return null;
                }
                final double samplingRate = num.doubleValue();
                columns[ii] = new SGSamplingDataColumn(samplingRate, len);
            } else {
                throw new Error("Illegal value type: " + valueType);
            }
        }
        return columns;
    }

    /**
     * Creates an array of SXY Data objects.
     * 
     * @return an array of created data objects
     */
    private CreatedDataSet createSingleSXYData(final SGIDataSource dataSource,
    		final SGIProgressControl progress, 
    		final SGDataColumnInfoSet colInfoSet, final Map<String, Object> infoMap) {
    	
        SGTextDataFile file = (SGTextDataFile) dataSource;
        final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column indices
        List<Integer> xIndexList = new ArrayList<Integer>();
        List<Integer> yIndexList = new ArrayList<Integer>();
        Map<Integer, Integer> lIndexMap = new HashMap<Integer, Integer>();
        Map<Integer, Integer> uIndexMap = new HashMap<Integer, Integer>();
        Map<Integer, Integer> tIndexMap = new HashMap<Integer, Integer>();
        if (SGDataUtility.getSXYColumnType(colInfo, xIndexList, yIndexList,
				lIndexMap, uIndexMap, tIndexMap) == false) {
        	return null;
		}
        
        if (xIndexList.size() != 1 && yIndexList.size() != 1) {
        	return null;
        }

        // create data objects
        List<SGISXYTypeData> dataList = new ArrayList<SGISXYTypeData>();
        List<String> titleList = new ArrayList<String>();
        if (xIndexList.size() == 1 && yIndexList.size() == 1) {
        	final Integer xIndex = xIndexList.get(0);
        	final Integer yIndex = yIndexList.get(0);
        	Integer xlIndex = lIndexMap.get(xIndex);
        	Integer ylIndex = lIndexMap.get(yIndex);
        	Integer xuIndex = uIndexMap.get(xIndex);
        	Integer yuIndex = uIndexMap.get(yIndex);
        	Integer lIndex, uIndex, ehIndex;
        	if (xlIndex != null && xuIndex != null) {
        		lIndex = xlIndex;
        		uIndex = xuIndex;
        		ehIndex = xIndex;
        	} else 	if (ylIndex != null && yuIndex != null) {
        		lIndex = ylIndex;
        		uIndex = yuIndex;
        		ehIndex = yIndex;
        	} else {
        		lIndex = null;
        		uIndex = null;
        		ehIndex = null;
        	}
        	Integer xtIndex = tIndexMap.get(xIndex);
        	Integer ytIndex = tIndexMap.get(yIndex);
        	Integer tIndex, thIndex;
        	if (xtIndex != null) {
        		tIndex = xtIndex;
        		thIndex = xIndex;
        	} else if (ytIndex != null) {
        		tIndex = ytIndex;
        		thIndex = yIndex;
        	} else {
        		tIndex = null;
        		thIndex = null;
        	}
            SGISXYTypeData data = new SGSXYData(file, this.mDataSourceObserver,
            		xIndex, yIndex, lIndex, uIndex, ehIndex, tIndex, thIndex);
            dataList.add(data);
            titleList.add(null);
        } else if (xIndexList.size() == 1) {
        	Integer xIndex = xIndexList.get(0);
            for (int ii = 0; ii < yIndexList.size(); ii++) {
            	final Integer yIndex = yIndexList.get(ii);
            	Integer lIndex = lIndexMap.get(yIndex);
            	Integer uIndex = uIndexMap.get(yIndex);
            	Integer tIndex = tIndexMap.get(yIndex);
                SGISXYTypeData data = new SGSXYData(file, this.mDataSourceObserver,
                		xIndex, yIndex, lIndex, uIndex, yIndex, tIndex, yIndex);
                dataList.add(data);
                String title = colInfo[yIndex.intValue()].getTitle();
                if (title == null || "".equals(title)) {
                	title = Integer.valueOf(yIndex.intValue() + 1).toString();
                }
                titleList.add(title);
            }
        } else {
        	Integer yIndex = yIndexList.get(0);
            for (int ii = 0; ii < xIndexList.size(); ii++) {
            	final Integer xIndex = xIndexList.get(ii);
            	Integer lIndex = lIndexMap.get(xIndex);
            	Integer uIndex = uIndexMap.get(xIndex);
            	Integer tIndex = tIndexMap.get(xIndex);
                SGISXYTypeData data = new SGSXYData(file, this.mDataSourceObserver,
                		xIndex, yIndex, lIndex, uIndex, xIndex, tIndex, xIndex);
                dataList.add(data);
                String title = colInfo[xIndex.intValue()].getTitle();
                if (title == null || "".equals(title)) {
                	title = Integer.valueOf(xIndex.intValue() + 1).toString();
                }
                titleList.add(title);
            }
        }

        // modifies the titles
        List<String> titleListNew = new ArrayList<String>();
        for (int ii = 0; ii < titleList.size(); ii++) {
        	String title = titleList.get(ii);
        	if (title == null) {
        		titleListNew.add(null);
        		continue;
        	}
        	title = SGUtility.addEscapeChar(title);
        	titleListNew.add(title);
        }
        titleList = titleListNew;
        
        SGData[] dataArray = new SGData[dataList.size()];
        for (int ii = 0; ii < dataArray.length; ii++) {
        	SGSXYData data = (SGSXYData) dataList.get(ii);
        	dataArray[ii] = (SGData) data.toMultiple();
        }
        String[] titleArray = titleList.toArray(new String[titleList.size()]);
        CreatedDataSet cdSet = new CreatedDataSet();
        for (int ii = 0; ii < titleArray.length; ii++) {
        	cdSet.addData(dataArray[ii], titleArray[ii]);
        }
        
        // disposes data objects
        for (int ii = 0; ii < dataArray.length; ii++) {
        	SGData data = (SGData) dataList.get(ii);
        	data.dispose();
        }
        
        return cdSet;
    }
    
//    // append sampling data columns
//    private SGDataColumn[] appendSamplingColumn(SGDataColumnInfo[] colInfo, 
//    		SGDataColumn[] columns, Double samplingRate) {
//
//        for (int ii = 0; ii < colInfo.length; ii++) {
//        	if (SGUtilityText.isEqualString(VALUE_TYPE_SAMPLING_RATE, 
//        			colInfo[ii].getValueType())) {
//
//                // create a data column for sampling
//        		SGDataColumn[] columnsNew = new SGDataColumn[columns.length + 1];
//        		for (int jj = 0; jj < columns.length; jj++) {
//        			columnsNew[jj] = columns[jj];
//        		}
//        		final int len = columns[0].getLength();
//        		columnsNew[columnsNew.length - 1] = new SGSamplingDataColumn(
//        				samplingRate.doubleValue(), len);
//        		columns = columnsNew;
//        		
//        		break;
//        	}
//        }
//
//        return columns;
//    }

    /**
     * Creates a SXY Multiple data.
     * 
     * @return created data object
     */
    private CreatedDataSet createMultipleSXYData(final SGIDataSource dataSource,
    		final SGIProgressControl progress, 
    		final SGDataColumnInfoSet colInfoSet, final Map<String, Object> infoMap) {
    	
        SGTextDataFile file = (SGTextDataFile) dataSource;
		final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column indices
        List<Integer> xIndexList = new ArrayList<Integer>();
        List<Integer> yIndexList = new ArrayList<Integer>();
        Map<Integer, Integer> lIndexMap = new HashMap<Integer, Integer>();
        Map<Integer, Integer> uIndexMap = new HashMap<Integer, Integer>();
        Map<Integer, Integer> tIndexMap = new HashMap<Integer, Integer>();
        if (SGDataUtility.getSXYColumnType(colInfo, xIndexList, yIndexList,
				lIndexMap, uIndexMap, tIndexMap) == false) {
        	return null;
		}

        if (xIndexList.size() != 1 && yIndexList.size() != 1) {
        	return null;
        }

        // create a data object
        Integer[] xIndices = new Integer[xIndexList.size()];
        xIndices = xIndexList.toArray(xIndices);
        Integer[] yIndices = new Integer[yIndexList.size()];
        yIndices = yIndexList.toArray(yIndices);
        
        Integer[] lIndices = null;
        Integer[] uIndices = null;
        Integer[] ehIndices = null;
        Integer[] tIndices = null;
        Integer[] thIndices = null;
        if (xIndexList.size() == 1 && yIndexList.size() == 1) {
        	final Integer xIndex = xIndexList.get(0);
        	final Integer yIndex = yIndexList.get(0);
        	Integer xlIndex = lIndexMap.get(xIndex);
        	Integer ylIndex = lIndexMap.get(yIndex);
        	Integer xuIndex = uIndexMap.get(xIndex);
        	Integer yuIndex = uIndexMap.get(yIndex);
        	Integer lIndex = null;
        	Integer uIndex = null;
        	Integer ehIndex = null;
        	if (xlIndex != null && xuIndex != null) {
        		lIndex = xlIndex;
        		uIndex = xuIndex;
        		ehIndex = xIndex;
        	} else 	if (ylIndex != null && yuIndex != null) {
        		lIndex = ylIndex;
        		uIndex = yuIndex;
        		ehIndex = yIndex;
        	}
        	if (lIndex != null && uIndex != null && ehIndex != null) {
            	lIndices = new Integer[] { lIndex };
            	uIndices = new Integer[] { uIndex };
            	ehIndices = new Integer[] { ehIndex };
        	}
        	Integer xtIndex = tIndexMap.get(xIndex);
        	Integer ytIndex = tIndexMap.get(yIndex);
        	Integer tIndex = null;
        	Integer thIndex = null;
        	if (xtIndex != null) {
        		tIndex = xtIndex;
        		thIndex = xIndex;
        	} else if (ytIndex != null) {
        		tIndex = ytIndex;
        		thIndex = yIndex;
        	}
        	if (tIndex != null) {
        		tIndices = new Integer[] { tIndex };
        		thIndices = new Integer[] { thIndex };
        	}
        } else if (xIndexList.size() == 1) {
        	List<Integer> lIndexList = new ArrayList<Integer>();
        	List<Integer> uIndexList = new ArrayList<Integer>();
        	List<Integer> ehIndexList = new ArrayList<Integer>();
        	List<Integer> tIndexList = new ArrayList<Integer>();
        	List<Integer> thIndexList = new ArrayList<Integer>();
            for (int ii = 0; ii < yIndexList.size(); ii++) {
            	final Integer yIndex = yIndexList.get(ii);
            	Integer lIndex = lIndexMap.get(yIndex);
            	Integer uIndex = uIndexMap.get(yIndex);
            	if (lIndex != null && uIndex != null) {
            		lIndexList.add(lIndex);
            		uIndexList.add(uIndex);
                	ehIndexList.add(yIndex);
            	}
            	Integer tIndex = tIndexMap.get(yIndex);
            	if (tIndex != null) {
            		tIndexList.add(tIndex);
                	thIndexList.add(yIndex);
            	}
            }
            lIndices = lIndexList.toArray(new Integer[lIndexList.size()]);
            uIndices = uIndexList.toArray(new Integer[uIndexList.size()]);
            ehIndices = ehIndexList.toArray(new Integer[ehIndexList.size()]);
            tIndices = tIndexList.toArray(new Integer[tIndexList.size()]);
            thIndices = thIndexList.toArray(new Integer[thIndexList.size()]);
        } else {
        	List<Integer> lIndexList = new ArrayList<Integer>();
        	List<Integer> uIndexList = new ArrayList<Integer>();
        	List<Integer> ehIndexList = new ArrayList<Integer>();
        	List<Integer> tIndexList = new ArrayList<Integer>();
        	List<Integer> thIndexList = new ArrayList<Integer>();
            for (int ii = 0; ii < xIndexList.size(); ii++) {
            	final Integer xIndex = xIndexList.get(ii);
            	Integer lIndex = lIndexMap.get(xIndex);
            	Integer uIndex = uIndexMap.get(xIndex);
            	if (lIndex != null && uIndex != null) {
            		lIndexList.add(lIndex);
            		uIndexList.add(uIndex);
                	ehIndexList.add(xIndex);
            	}
            	Integer tIndex = tIndexMap.get(xIndex);
            	if (tIndex != null) {
            		tIndexList.add(tIndex);
            		thIndexList.add(xIndex);
            	}
            }
            lIndices = lIndexList.toArray(new Integer[lIndexList.size()]);
            uIndices = uIndexList.toArray(new Integer[uIndexList.size()]);
            ehIndices = ehIndexList.toArray(new Integer[ehIndexList.size()]);
            tIndices = tIndexList.toArray(new Integer[tIndexList.size()]);
            thIndices = thIndexList.toArray(new Integer[thIndexList.size()]);
        }

        SGSXYMultipleData data = new SGSXYMultipleData(file, this.mDataSourceObserver,
        		xIndices, yIndices, lIndices,
        		uIndices, ehIndices, tIndices, thIndices);
        
        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        
        return cdSet;
    }

    /**
     * Creates a VXY data.
     * 
     * @return created data object
     */
    private CreatedDataSet createVXYData(final SGIDataSource dataSource,
    		final SGIProgressControl progress, 
    		final SGDataColumnInfoSet colInfoSet, final Map<String, Object> infoMap) {

    	SGTextDataFile file = (SGTextDataFile) dataSource;
		final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get polar flag
        if (infoMap.size() <= 1) {
            return null;
        }
        Object obj = infoMap.get(KEY_POLAR_SELECTED);
        if ((obj instanceof Boolean) == false) {
            return null;
        }
        Boolean b = (Boolean) obj;
        final boolean isPolar = b.booleanValue();

        // get column indices
        final String type1 = isPolar ? MAGNITUDE : X_COMPONENT;
        final String type2 = isPolar ? ANGLE : Y_COMPONENT;
        Integer xIndex = null;
        Integer yIndex = null;
        Integer c1Index = null;
        Integer c2Index = null;
        for (int ii = 0; ii < colInfo.length; ii++) {
            String cType = colInfo[ii].getColumnType();
            if (SGUtilityText.isEqualString(X_COORDINATE, cType)) {
                xIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Y_COORDINATE, cType)) {
                yIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(type1, cType)) {
                c1Index = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(type2, cType)) {
                c2Index = Integer.valueOf(ii);
            } else if ("".equals(cType)) {
                continue;
            } else {
                throw new Error("Illegal Column Type: " + cType);
            }
        }

        // create a data object
        SGData data = new SGVXYData(file, 
        		this.mDataSourceObserver, xIndex, yIndex, c1Index, c2Index, isPolar);

        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        
        return cdSet;
    }

    /**
     * Creates a VXY data.
     * 
     * @return created data object
     */
    private CreatedDataSet createSXYZData(final SGIDataSource dataSource,
    		final SGIProgressControl progress, 
    		final SGDataColumnInfoSet colInfoSet, final Map<String, Object> infoMap) {
    	
    	SGTextDataFile file = (SGTextDataFile) dataSource;
		final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column indices
        Integer xIndex = null;
        Integer yIndex = null;
        Integer zIndex = null;
        for (int ii = 0; ii < colInfo.length; ii++) {
            String cType = colInfo[ii].getColumnType();
            if (SGUtilityText.isEqualString(X_VALUE, cType)) {
                xIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Y_VALUE, cType)) {
                yIndex = Integer.valueOf(ii);
            } else if (SGUtilityText.isEqualString(Z_VALUE, cType)) {
                zIndex = Integer.valueOf(ii);
            } else if ("".equals(cType)) {
                continue;
            } else {
                throw new Error("Illegal Column Type: " + cType);
            }
        }

        // create a data object
        SGData data = new SGSXYZData(file,
        		this.mDataSourceObserver, xIndex, yIndex, zIndex);

        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        
        return cdSet;
    }
    
    /**
     * Creates an array of the list.
     * <p>
     * Array size is number of rows in file.
     * @param path file path
     * @return
     * @throws FileNotFoundException
     */
    @SuppressWarnings("unchecked")
    private List<String>[] createListArray(final String path)
    throws FileNotFoundException {

        List<String>[] listArray = null;
    	int colNum = -1;

        // read the file
    	SGBufferedFileReader reader = null;
        try {
        	reader = new SGBufferedFileReader(path);
        	BufferedReader br = reader.getBufferedReader();
            boolean isFirstLine = true;

            while (true) {
                // read a line
                final String line = SGUtilityText.readLine(br);
                if (line == null) {
                    break;
                }

                // break the string into tokens
                final List<Token> tokenList = new ArrayList<Token>();
                // line is read from data file
                if (SGUtilityText.tokenize(line, tokenList, true) == false) {
                    return null;
                }
                
                // in case of an empty line or a comment line
                if (tokenList.size() == 0) {
                	continue;
                }

                if (colNum == -1) {
                	colNum = tokenList.size();
                	
                	// create list array
                	listArray = new List[colNum];
                    for (int ii = 0; ii < listArray.length; ii++) {
                        listArray[ii] = new ArrayList<String>();
                    }
                    
                } else {
                    // check the tokens
                    if (tokenList.size() != colNum) {
                        if (tokenList.size() != 0) {
                            // format error
                            return null;
                        }
                        // only contains spaces and tabs
                        continue;
                    }
                }

                // Checks whether all tokens consist of spaces.
                boolean spacesOnly = true;
                for (int ii = 0; ii < colNum; ii++) {
                    Token token = (Token) tokenList.get(ii);
                	String str = token.getString();
                	char[] cArray = str.toCharArray();
                	for (int jj = 0; jj < cArray.length; jj++) {
                		if (!Character.isSpaceChar(cArray[jj])) {
                			spacesOnly = false;
                			break;
                		}
                	}
                	if (!spacesOnly) {
                		break;
                	}
                }
                if (spacesOnly) {
                	// Skips a line if all tokens consists of spaces.
                	continue;
                }

                // format check
                List<Integer> indexList = SGDataUtility.getColumnIndexListOfNumber(tokenList);

                if (isFirstLine) {
                    isFirstLine = false;
                    if (this.evaluteTitleList(indexList)) {
                        continue;
                    }
                }

                // array of the tokens
                for (int ii = 0; ii < colNum; ii++) {
                    Token token = (Token) tokenList.get(ii);
                    listArray[ii].add(token.getString());
                }
            }
        } catch (IOException ex) {
            return null;
		} finally {
            if (reader != null) {
            	reader.close();
            }
        }

        // check array index
        final int dataLength = listArray[0].size();
        for (int ii = 1; ii < listArray.length; ii++) {
            if (listArray[ii].size() != dataLength) {
                return null;
            }
        }

        return listArray;
    }

    // check title list
    private boolean evaluteTitleList(final List<Integer> indexList) {
        boolean ret = true;
        if (indexList.size() == 0) {
            return false;
        }
        for (int ii = 0; ii < indexList.size(); ii++) {
            Integer val = (Integer) indexList.get(ii);
            if (val.intValue() != 0) {
                ret = false;
            }
        }
        return ret;
    }

    /**
     * Determine the data-type from the first line.
     * 
     * @param path
     *            the path of file
     * @param cList
     *            the list of candidates of data types
     * @return true if succeeded
     */
    public boolean getDataTypeCandidateList(final String path,
            final List<String> cList) throws FileNotFoundException {

    	if (this.getTextDataTypeCandidateList(path, cList) == false) {
    		return false;
    	}
    	
    	/*
        try {
            if (NetcdfFile.canOpen(path)) {
            	if (this.getNetCDFDataTypeCandidateList(path, cList) == false) {
            		return false;
            	}

            } else {
            	if (this.getTextDataTypeCandidateList(path, cList) == false) {
            		return false;
            	}
            }
            
        } catch (IOException e) {
        	return false;
        }
        */
    	
        return true;
    }

    /*
    private boolean getNetCDFDataTypeCandidateList(final String path,
            final List<String> cList) throws FileNotFoundException {
    	
    	// open the netCDF file
        NetcdfFile ncFile = null;
    	try {
			ncFile = NetcdfFile.open(path);
		} catch (IOException e) {
			return false;
        } finally {            
            if (ncFile != null) {
                try {
                    ncFile.close();
                } catch (IOException e) {
                }
            }
        }
        
    	SGNetCDFFile nc = new SGNetCDFFile(ncFile);
    	
        // create column info map
        Map<String, Object> infoMap = new HashMap<String, Object>();
        infoMap.put(KEY_DATA_SOURCE, nc);
        infoMap.put(KEY_MULTIPLE, Boolean.FALSE);
        infoMap.put(KEY_MULTIPLE_VARIABLE, Boolean.TRUE);
        infoMap.put(KEY_POLAR_SELECTED, Boolean.TRUE);

        // create column list
        List<SGVariable> varList = nc.getVariables();
        final int size = varList.size();
        SGNetCDFDataColumnInfo[] cols = new SGNetCDFDataColumnInfo[size];
        int cnt = 0;
        for (SGVariable var : varList) {
            String name = var.getNameInPriorityOrder();
            cols[cnt] = new SGNetCDFDataColumnInfo(var, name, var.getValueType());
            cnt++;
        }
        List<SGDataColumnInfo> columnInfoList = new ArrayList<SGDataColumnInfo>(
        		Arrays.asList(cols));

        // get available data types
        String[] dataTypes = new String[] {
        		SGDataTypeConstants.SXY_NETCDF_DATA,
//        		SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA,
        		SGDataTypeConstants.VXY_NETCDF_DATA,
        		SGDataTypeConstants.SXYZ_NETCDF_DATA
        };
        for (int ii = 0; ii < dataTypes.length; ii++) {
            SGDataUtility.DefaultColumnTypeResult result = SGDataUtility
					.getDefaultColumnTypes(dataTypes[ii],
							columnInfoList, infoMap);
            if (result.isSucceeded()) {
            	cList.add(dataTypes[ii]);
            }
        }

        return true;
    }
    */
    
    private boolean getTextDataTypeCandidateList(final String path,
            final List<String> cList) throws FileNotFoundException {

    	SGBufferedFileReader reader = null;
        try {
        	reader = new SGBufferedFileReader(path);
            BufferedReader br = reader.getBufferedReader();

            // get a list of tokens from the first line
            List<Token> firstTokenList = SGUtilityText.getFirstTokenList(br);
            if (firstTokenList == null) {
                return false;
            }

            // get number convertable columns
            List<Integer> firstIndexList = SGDataUtility.getColumnIndexListOfNumber(
            		firstTokenList);

            // get titles
            if (this.evaluteTitleList(firstIndexList)) {
                // this is title. read next line for indexList
                firstTokenList = SGUtilityText.getFirstTokenList(br);
                if (firstTokenList == null) {
                    return false;
                }
                firstIndexList = SGDataUtility.getColumnIndexListOfNumber(firstTokenList);
            }
                                
            // check all values
            List<String>[] listArray = this.createListArray(path);
            if (listArray == null) {
                return false;
            }
            List<Integer> indexList = new ArrayList<Integer>();
            for (int ii = 0; ii < listArray.length; ii++) {
                Integer firstIndex = (Integer) firstIndexList.get(ii);
                List<String> list = listArray[ii];
                if (firstIndex.intValue() != 0) {
                    // number
                    boolean err = false;
                    for (int jj = 0; jj < list.size(); jj++) {
                        String value = (String) list.get(jj);
                        // check the number
                        if (SGUtilityText.getDouble(value) == null) {
                            err = true;
                            break;
                        }
                    }
                    if (err) {
                        indexList.add(Integer.valueOf(0));
                    } else {
                        indexList.add(Integer.valueOf(1));                                
                    }
                } else {
                    indexList.add(Integer.valueOf(0));                            
                }
            }
            
            // get data type candidates
            if (SGDataUtility.getDataTypeCandidateList(
                    firstTokenList, indexList, cList) == false) {
                return false;
            }
            
        } catch (IOException ex) {
        	return false;
		} finally {
            if (reader != null) {
            	reader.close();
            }
        }

        return true;
    }
    
    /**
     * A class which has information of each column of a data file.
     */
    class FileColumn {
        // column index
        int index = -1;

        // title
        String title = null;

        // data type
        String valueType = null;

        /**
         * Returns a string representation.
         */
        public String toString() {
            StringBuffer sb = new StringBuffer();
            sb.append('[');
            sb.append("index=");
            sb.append(this.index);
            sb.append(", title=");
            sb.append(this.title);
            sb.append(", valueType=");
            sb.append(this.valueType);
            sb.append(']');
            return sb.toString();
        }
    }

    /**
     * Read a given file and parse the type of each column.
     * 
     * @param path
     *            the file path
     * @param dataType
     *            the type of data
     * @param isPropertyFileData
     *            set true if data is for the property file
     * @param versionNumber
     *            the version number of the property file
     * @return information of columns
     * @throws FileNotFoundException
     */
    public FileColumn[] parseFileComlumnType(final String path, final String dataType,
            final boolean isPropertyFileData, final String versionNumber)
            throws FileNotFoundException {
        FileColumn[] info = null;
        SGBufferedFileReader reader = null;
        try {
        	reader = new SGBufferedFileReader(path);
            BufferedReader br = reader.getBufferedReader();
        	
            List<Token> tokenList = SGUtilityText.getFirstTokenList(br);
            if (tokenList == null) {
                return null;
            }
            final int colNum = tokenList.size();
            info = new FileColumn[colNum];

            // get number convertable columns
            List<Integer> indexList = SGDataUtility.getColumnIndexListOfNumber(tokenList);

            // get titles
            String[] titles = new String[colNum];
            if (this.evaluteTitleList(indexList)) {
                // this line represents titles
                for (int ii = 0; ii < colNum; ii++) {
                    Token token = (Token) tokenList.get(ii);
                    titles[ii] = token.getString();
                }

                // read next line
                tokenList = SGUtilityText.getFirstTokenList(br);
                if (tokenList == null) {
                    return null;
                }
                if (tokenList.size() != colNum) {
                    return null;
                }
            } else {
                for (int ii = 0; ii < colNum; ii++) {
                    titles[ii] = null;
                }
            }

            for (int ii = 0; ii < colNum; ii++) {
                info[ii] = new FileColumn();
                info[ii].index = ii;
                info[ii].title = titles[ii];
                String valueType = null;
                Token token = (Token) tokenList.get(ii);
                if (token.isDoubleQuoted()) {
                    Date d = SGUtilityText.getDate(token.getString());
                    if (d != null) {
                        valueType = VALUE_TYPE_DATE;
                    } else {
                        valueType = VALUE_TYPE_TEXT;
                    }
                } else {
                    Number num = SGUtilityText.getDouble(token.getString());
                    if (num != null) {
                        valueType = VALUE_TYPE_NUMBER;
                    } else {
                        Date d = SGUtilityText.getDate(token.getString());
                        if (d != null) {
                            valueType = VALUE_TYPE_DATE;
                        } else {
                            valueType = VALUE_TYPE_TEXT;
                        }
                    }                    
                }
                info[ii].valueType = valueType;
            }
            
            // check values in all number columns and date columns
            List<Integer> numColList = new ArrayList<Integer>();
            List<Integer> dateColList = new ArrayList<Integer>();
            for (int ii = 0; ii < colNum; ii++) {
                if (SGUtilityText.isEqualString(VALUE_TYPE_NUMBER, info[ii].valueType)) {
                    numColList.add(Integer.valueOf(ii));
                } else if (SGUtilityText.isEqualString(VALUE_TYPE_DATE, info[ii].valueType)) {
                    dateColList.add(Integer.valueOf(ii));
                }
            }
            
            if (numColList.size() != 0 || dateColList.size() != 0) {
                    
                // get all values
            	List<String>[] listArray = this.createListArray(path);
                if (listArray == null) {
                    return null;
                }
                
                // check number columns
                List<Integer> errNumColList = new ArrayList<Integer>();
                for (int ii = 0; ii < numColList.size(); ii++) {
                    Integer colIndex = (Integer) numColList.get(ii);
                    List<String> list = listArray[colIndex.intValue()];
                    boolean valid = true;
                    for (int jj = 0; jj < list.size(); jj++) {
                        String value = list.get(jj);
                        Double d = SGUtilityText.getDouble(value);
                        if (d == null) {
                            valid = false;
                        }
                    }
                    if (!valid) {
                        errNumColList.add(colIndex);
                    }
                }

                // check date columns
                List<Integer> errDateColList = new ArrayList<Integer>();
                for (int ii = 0; ii < dateColList.size(); ii++) {
                    Integer colIndex = (Integer) dateColList.get(ii);
                    List<String> list = listArray[colIndex.intValue()];
                    boolean valid = true;
                    for (int jj = 0; jj < list.size(); jj++) {
                        String value = list.get(jj);
                        Date d = SGUtilityText.getDate(value);
                        if (d == null) {
                            valid = false;
                        }
                    }
                    if (!valid) {
                            errDateColList.add(colIndex);
                    }
                }
                
                // replace the column type from number or date to text
                for (int ii = 0; ii < errNumColList.size(); ii++) {
                    Integer colIndex = (Integer) errNumColList.get(ii);
                    info[colIndex.intValue()].valueType = VALUE_TYPE_TEXT;
                }                   
                for (int ii = 0; ii < errDateColList.size(); ii++) {
                    Integer colIndex = (Integer) errDateColList.get(ii);
                    info[colIndex.intValue()].valueType = VALUE_TYPE_TEXT;
                }                   
            }

            // measures for version older than 2.0.0
            if (isPropertyFileData) {
                if (versionNumber != null) {
                    // if version of the property file is older than 2.0.0
                    if ("".equals(versionNumber)) {
                        // data with tick labels
                        if (SGDataUtility.hasTickLabels(dataType)) {
                            // the third and fifth columns are forced to be a text column
                            if (colNum == 3 || colNum == 5) {
                                info[colNum - 1].valueType = VALUE_TYPE_TEXT;
                            }
                        }
                    }
                }
            }
            
        } catch (IOException ex) {
            return null;
		} finally {
            if (reader != null) {
            	reader.close();
            }
        }
        return info;
    }

    
    /**
     * Creates an array of SXY NetCDF Data.
     * 
     * @return created data
     */
    private CreatedDataSet createSingleVariableSXYNetCDFData(final SGIDataSource dataSource,
            final SGIProgressControl progress, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap) {

        SGNetCDFFile file = (SGNetCDFFile) dataSource;
        NetcdfFile ncFile = file.getNetcdfFile();
        final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column names
        List<String> xNameList = new ArrayList<String>();
        List<String> yNameList = new ArrayList<String>();
        Map<String, String> lNameMap = new HashMap<String, String>();
        Map<String, String> uNameMap = new HashMap<String, String>();
        Map<String, String> tickNameMap = new HashMap<String, String>();
        List<String> timeNameList = new ArrayList<String>();
        List<String> indexNameList = new ArrayList<String>();
        if (!SGDataUtility.getSXYNetCDFColumnType(colInfo, xNameList,
                yNameList, lNameMap, uNameMap, tickNameMap, timeNameList,
                indexNameList)) {
            return null;
        }
        if (xNameList.size() != 1 && yNameList.size() != 1) {
            return null;
        }

        String timeName = null;
        if (timeNameList.size() != 0) {
            timeName = timeNameList.get(0);
        }
        String indexName = null;
        if (indexNameList.size() != 0) {
            indexName = indexNameList.get(0);
        }

        // create data objects
        List<SGISXYTypeData> dataList = new ArrayList<SGISXYTypeData>();
        List<String> titleList = new ArrayList<String>();
        if (xNameList.size() == 1 && yNameList.size() == 1) {
            final String xName = xNameList.get(0);
            final String yName = yNameList.get(0);
            String xlName = lNameMap.get(xName);
            String ylName = lNameMap.get(yName);
            String xuName = uNameMap.get(xName);
            String yuName = uNameMap.get(yName);
            String lName, uName, ehName;
            if (xlName != null && xuName != null) {
                lName = xlName;
                uName = xuName;
                ehName = xName;
            } else 	if (ylName != null && yuName != null) {
                lName = ylName;
                uName = yuName;
                ehName = yName;
            } else {
                lName = null;
                uName = null;
                ehName = null;
            }
            String xTickName = tickNameMap.get(xName);
            String yTickName = tickNameMap.get(yName);
            String tName, thName;
            if (xTickName != null) {
                tName = xTickName;
                thName = xName;
            } else if (yTickName != null) {
                tName = yTickName;
                thName = yName;
            } else {
                tName = null;
                thName = null;
            }
            SGSXYNetCDFData data;
            if (file.findVariable(xName) instanceof SGDateVariable || file.findVariable(yName) instanceof SGDateVariable) {
                data = new SGSXYNetCDFDateData(file,
                        this.mDataSourceObserver, xName, yName, lName, uName, ehName,
                        tName, thName, timeName, indexName);
            } else {
                data = new SGSXYNetCDFData(file,
                        this.mDataSourceObserver, xName, yName, lName, uName, ehName,
                        tName, thName, timeName, indexName);
            }
            dataList.add(data);

            // set the title
            titleList.add(null);

            // set origins
            this.setOrigin(colInfo, ncFile, data);

        } else if (xNameList.size() == 1) {
            String xName = xNameList.get(0);
            for (int ii = 0; ii < yNameList.size(); ii++) {
                final String yName = yNameList.get(ii);
                String lName = lNameMap.get(yName);
                String uName = uNameMap.get(yName);
                String tName = tickNameMap.get(yName);
                SGSXYNetCDFData data = new SGSXYNetCDFData(file,
                        this.mDataSourceObserver, xName, yName, lName, uName, yName,
                        tName, yName, timeName, indexName);
                dataList.add(data);

                // set the title
                titleList.add(yName);

                // set origins
                this.setOrigin(colInfo, ncFile, data);
            }

        } else {
            String yName = yNameList.get(0);
            for (int ii = 0; ii < xNameList.size(); ii++) {
                final String xName = xNameList.get(ii);
                String lName = lNameMap.get(xName);
                String uName = uNameMap.get(xName);
                String tName = tickNameMap.get(yName);
                SGSXYNetCDFData data = new SGSXYNetCDFData(file,
                        this.mDataSourceObserver, xName, yName, lName, uName, xName,
                        tName, xName, timeName, indexName);
                dataList.add(data);

                // set the title
                titleList.add(xName);

                // set origins
                this.setOrigin(colInfo, ncFile, data);
            }
        }

        // modifies the titles
        List<String> titleListNew = new ArrayList<String>();
        for (int ii = 0; ii < titleList.size(); ii++) {
        	String title = titleList.get(ii);
        	if (title == null) {
        		titleListNew.add(null);
        		continue;
        	}
        	title = SGUtility.addEscapeChar(title);
        	titleListNew.add(title);
        }
        titleList = titleListNew;
        
        SGData[] dataArray = new SGData[dataList.size()];
        for (int ii = 0; ii < dataArray.length; ii++) {
            SGSXYNetCDFData data = (SGSXYNetCDFData) dataList.get(ii);
            dataArray[ii] = (SGData) data.toMultiple();
        }
        String[] titleArray = titleList.toArray(new String[titleList.size()]);
        CreatedDataSet cdSet = new CreatedDataSet();
        for (int ii = 0; ii < titleArray.length; ii++) {
            cdSet.addData(dataArray[ii], titleArray[ii]);
        }

        // disposes data objects
        for (int ii = 0; ii < dataArray.length; ii++) {
            SGData data = (SGData) dataList.get(ii);
            data.dispose();
        }

        return cdSet;
    }

    /**
     * Creates SXY NetCDF Data with multiple variables.
     * 
     * @return created data
     */
    private CreatedDataSet createMultipleVariableSXYNetCDFData(final SGIDataSource dataSource,
            final SGIProgressControl progress, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap) {

        SGNetCDFFile file = (SGNetCDFFile) dataSource;
        NetcdfFile ncFile = file.getNetcdfFile();
        final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        SGSXYMultipleNetCDFData data = SGDataUtility.createVariableNetCDFData(file, this.mDataSourceObserver, colInfo);
        
//        // get column names
//        List<String> xNameList = new ArrayList<String>();
//        List<String> yNameList = new ArrayList<String>();
//        Map<String, String> lNameMap = new HashMap<String, String>();
//        Map<String, String> uNameMap = new HashMap<String, String>();
//        Map<String, String> tNameMap = new HashMap<String, String>();
//        List<String> timeNameList = new ArrayList<String>();
//        List<String> indexNameList = new ArrayList<String>();
//        if (!SGDataUtility.getSXYNetCDFColumnType(colInfo, xNameList,
//                yNameList, lNameMap, uNameMap, tNameMap, timeNameList, indexNameList)) {
//            return null;
//        }
//        if (lNameMap.size() != uNameMap.size()) {
//            return null;
//        }
//        String[] xNames = xNameList.toArray(new String[xNameList.size()]);
//        String[] yNames = yNameList.toArray(new String[yNameList.size()]);
//
//        Iterator<String> eItr = lNameMap.keySet().iterator();
//        List<String> lNameList = new ArrayList<String>();
//        List<String> uNameList = new ArrayList<String>();
//        List<String> ehNameList = new ArrayList<String>();
//        while (eItr.hasNext()) {
//            String eh = eItr.next();
//            ehNameList.add(eh);
//            String lName = lNameMap.get(eh);
//            if (lName != null) {
//                lNameList.add(lName);
//            }
//            String uName = uNameMap.get(eh);
//            if (uName != null) {
//                uNameList.add(uName);
//            }
//        }
//        if (lNameList.size() != uNameList.size()) {
//            return null;
//        }
//        if (lNameList.size() != ehNameList.size()) {
//            return null;
//        }
//        String[] lNames = lNameList.toArray(new String[lNameList.size()]);
//        String[] uNames = uNameList.toArray(new String[uNameList.size()]);
//        String[] ehNames = ehNameList.toArray(new String[ehNameList.size()]);
//
//        Iterator<String> tItr = tNameMap.keySet().iterator();
//        List<String> tNameList = new ArrayList<String>();
//        List<String> thNameList = new ArrayList<String>();
//        while (tItr.hasNext()) {
//            String th = tItr.next();
//            thNameList.add(th);
//            String tName = tNameMap.get(th);
//            if (tName != null) {
//                tNameList.add(tName);
//            }
//        }
//        if (tNameList.size() != thNameList.size()) {
//            return null;
//        }
//        String[] tNames = tNameList.toArray(new String[tNameList.size()]);
//        String[] thNames = thNameList.toArray(new String[thNameList.size()]);
//
//        String timeName = null;
//        if (timeNameList.size() != 0) {
//            timeName = timeNameList.get(0);
//        }
//        String indexName = null;
//        if (indexNameList.size() != 0) {
//            indexName = indexNameList.get(0);
//        }
//
//        // create data object
//        SGSXYMultipleNetCDFData data = new SGSXYMultipleVariableNetCDFData(
//                file, this.mDataSourceObserver, xNames, yNames, lNames, uNames, ehNames, 
//                tNames, thNames, timeName, indexName);

        // set origins
        this.setOrigin(colInfo, ncFile, data);

        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        return cdSet;
    }

    /**
     * Creates SXY NetCDF Data with multiple dimension indices.
     * 
     * @return created data
     */
    private CreatedDataSet createDimensionSXYNetCDFData(final SGIDataSource dataSource,
            final SGIProgressControl progress, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap, final boolean multiple) {
    	
        SGNetCDFFile file = (SGNetCDFFile) dataSource;
        NetcdfFile ncFile = file.getNetcdfFile();
        final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column names
        List<String> xNameList = new ArrayList<String>();
        List<String> yNameList = new ArrayList<String>();
        List<String> lNameList = new ArrayList<String>();
        List<String> uNameList = new ArrayList<String>();
        List<String> tNameList = new ArrayList<String>();
        List<String> mNameList = new ArrayList<String>();
        List<String> timeNameList = new ArrayList<String>();
        List<String> indexNameList = new ArrayList<String>();
        if (SGDataUtility.getSXYNetCDFColumnType(
                colInfo, xNameList, yNameList,
                lNameList, uNameList, tNameList, mNameList, timeNameList,
                indexNameList) == false) {
            return null;
        }
        if (xNameList.size() != 1) {
            return null;
        }
        if (yNameList.size() != 1) {
            return null;
        }
        if (mNameList.size() != 1) {
            return null;
        }
        if (!(lNameList.size() == 0 && uNameList.size() == 0) 
                && !(lNameList.size() == 1 && uNameList.size() == 1)) {
            return null;
        }
        String xName = xNameList.get(0);
        String yName = yNameList.get(0);
        String lName = null;
        if (lNameList.size() > 0) {
            lName = lNameList.get(0);
        }
        String uName = null;
        if (uNameList.size() > 0) {
            uName = uNameList.get(0);
        }
        String tName = null;
        if (tNameList.size() > 0) {
            tName = tNameList.get(0);
        }
        String dimName = mNameList.get(0);
        String timeName = null;
        if (timeNameList.size() != 0) {
            timeName = timeNameList.get(0);
        }
        String indexName = null;
        if (indexNameList.size() != 0) {
            indexName = indexNameList.get(0);
        }

        // create an array of dimension indices
        Dimension dim = ncFile.findDimension(dimName);
        if (dim == null) {
            return null;
        }
        final int dimLen = dim.getLength();
        Integer dimStart = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_START);
        if (dimStart == null) {
            return null;
        }
        if (dimStart.intValue() < 0 || dimStart.intValue() >= dimLen) {
            return null;
        }
        Integer dimEnd = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_END);
        if (dimEnd == null) {
            return null;
        }
        if (dimEnd.intValue() < 0 || dimEnd.intValue() >= dimLen || dimEnd.intValue()<dimStart.intValue()) {
            return null;
        }
        Integer dimStep = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_STEP);
        if (dimStep == null) {
            return null;
        }

        CreatedDataSet cdSet = new CreatedDataSet();

        if (multiple) {
            // create data object
            SGSXYMultipleNetCDFData data = new SGSXYMultipleDimensionNetCDFData(
                    file, this.mDataSourceObserver, xName, yName, lName, uName, 
                    tName, dimName, dimStart.intValue(), dimEnd.intValue(), dimStep.intValue(),
                    timeName, indexName);

            // set origins
            this.setOrigin(colInfo, ncFile, data);

            cdSet.addData(data);

        } else {

            Integer[] dimIndices = SGDataUtility.calcDimIndices(dim, dimStart, dimEnd, dimStep);
            for (int ii = 0; ii < dimIndices.length; ii++) {
                // create data object
                SGSXYMultipleDimensionNetCDFData data =
                    new SGSXYMultipleDimensionNetCDFData(
                            file, this.mDataSourceObserver, xName, yName, lName, uName, 
                            tName, dimName, dimIndices[ii], dimIndices[ii], 0,
                            timeName, indexName);

                // set origins
                this.setOrigin(colInfo, ncFile, data);

                // the suffix for the data name
                String suffix = this.getDimensionDataName(data, ii, dimName, dimIndices[ii]);

                cdSet.addData(data, "("+suffix+")");
            }
            cdSet.setMiddleString(" ");
        }

        return cdSet;
    }
    
    /**
     * Return the suffix for the data name.
     * <p>
     * If dimIndex is null or data has no value for dimIndex, return index.
     * Otherwise, return "dimName=value".
     * if same string exists in cdSet.
     * @param data
     * @param index
     * @param dimName
     * @param dimIndex
     * @return the suffix for the data name
     */
    private String getDimensionDataName(
            final SGSXYMultipleDimensionNetCDFData data, final int index,
            final String dimName, final Integer dimIndex) {
        if (null==dimIndex) {
            return Integer.toString(index);
        }
        Double dvalue = data.getDimensionValue(dimIndex.intValue());
        if (null==dvalue) {
            return Integer.toString(index);
        } else {
            double val = dvalue.doubleValue();
            String name = dimName+"="+SGUtility.getNumberName(val);
            return name;
        }
    }

    /**
     * Creates VXY NetCDF Data.
     * 
     * @return a data object
     */
    private CreatedDataSet createVXYNetCDFData(final SGIDataSource dataSource,
            final SGIProgressControl progress, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap) {
    	
    	SGNetCDFFile file = (SGNetCDFFile) dataSource;
    	NetcdfFile ncFile = file.getNetcdfFile();
		final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get polar flag
        if (infoMap.size() <= 1) {
            return null;
        }
        Object obj = infoMap.get(KEY_POLAR_SELECTED);
        if ((obj instanceof Boolean) == false) {
            return null;
        }
        Boolean b = (Boolean) obj;
        final boolean isPolar = b.booleanValue();

        // allocate memory
        progress.setProgressMessage(PROGRESS_MESSAGE_CREATEDATA);

        // get column indices
        final String type1 = isPolar ? MAGNITUDE : X_COMPONENT;
        final String type2 = isPolar ? ANGLE : Y_COMPONENT;
        String xName = null;
        String yName = null;
        String fName = null;
        String sName = null;
        String tName = null;
        String iName = null;
        for (int ii = 0; ii < colInfo.length; ii++) {
        	SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[ii];
            String cType = col.getColumnType();
            String name = col.getVariableName();
            if (SGUtilityText.isEqualString(X_COORDINATE, cType)) {
                xName = name;
            } else if (SGUtilityText.isEqualString(Y_COORDINATE, cType)) {
                yName = name;
            } else if (SGUtilityText.isEqualString(type1, cType)) {
                fName = name;
            } else if (SGUtilityText.isEqualString(type2, cType)) {
                sName = name;
            } else if (SGUtilityText.isEqualString(TIME, cType)) {
                tName = name;
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, cType)) {
                iName = name;
            } else if ("".equals(cType)) {
                continue;
            } else {
                throw new Error("Illegal Column Type: " + cType);
            }
        }

        // create data object
        SGVXYNetCDFData data = new SGVXYNetCDFData(file, 
        		this.mDataSourceObserver, xName, yName, fName, sName, isPolar,
        		tName, iName);
        
        // set origins
        this.setOrigin(colInfo, ncFile, data);
        
        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        return cdSet;
    }

    /**
     * Creates SXYZ NetCDF Data.
     * 
     * @return a data object
     */
    private CreatedDataSet createSXYZNetCDFData(final SGIDataSource dataSource,
            final SGIProgressControl progress, final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap) {
    	
    	SGNetCDFFile file = (SGNetCDFFile) dataSource;
    	NetcdfFile ncFile = file.getNetcdfFile();
		final SGDataColumnInfo[] colInfo = colInfoSet.getDataColumnInfoArray();

        // get column indices
        String xName = null;
        String yName = null;
        String zName = null;
        String tName = null;
        String iName = null;
        for (int ii = 0; ii < colInfo.length; ii++) {
        	SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[ii];
            String cType = col.getColumnType();
            String name = col.getVariableName();
            if (SGUtilityText.isEqualString(X_VALUE, cType)) {
                xName = name;
            } else if (SGUtilityText.isEqualString(Y_VALUE, cType)) {
                yName = name;
            } else if (SGUtilityText.isEqualString(Z_VALUE, cType)) {
                zName = name;
            } else if (SGUtilityText.isEqualString(TIME, cType)) {
                tName = name;
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, cType)) {
                iName = name;
            } else if ("".equals(cType)) {
                continue;
            } else {
                throw new Error("Illegal Column Type: " + cType);
            }
        }

        // create data object
        SGSXYZNetCDFData data = new SGSXYZNetCDFData(file,
        		this.mDataSourceObserver, xName, yName, zName, tName, iName);
        
        // set origins
        this.setOrigin(colInfo, ncFile, data);
        
        CreatedDataSet cdSet = new CreatedDataSet();
        cdSet.addData(data);
        return cdSet;
    }
    
    // Returns the sampling rate.
    private Number getSamplingRate(Map<String, Object> infoMap) {
        Object obj = infoMap.get(KEY_SAMPLING_RATE);
        if (obj == null) {
        	return null;
        }
        if (!(obj instanceof Number)) {
            return null;
        }
        return (Number) obj;
    }
    
    // Sets the origin to data.
    private void setOrigin(SGDataColumnInfo[] colInfo, NetcdfFile ncFile, 
            SGNetCDFData ncData) {
        for (int ii = 0; ii < colInfo.length; ii++) {
            SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[ii];
            final String name = col.getVariableName();
            final int origin = col.getOrigin();
            if (ncFile.findDimension(name) != null) {
                ncData.setOrigin(name, origin);
            }
        }
    }

    /**
     * Creates a data object.
     * 
     * @param dataSource
     *           the data source
     * @param colInfoSet
     *           a set of column information
     * @param infoMap
     *            information of data
     * @param progress
     *            the progress monitor
     * @return a set of created data
     */
    public CreatedDataSet create(SGIDataSource dataSource,
            final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap, final SGIProgressControl progress) {
//        progress.setIndeterminateProgress(true);
        return this.createData(dataSource, colInfoSet, infoMap, progress);
    }
    
    /**
     * Creates a data object.
     */
    private CreatedDataSet createData(SGIDataSource dataSource,
            final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap, final SGIProgressControl progress) {
    	
        progress.setProgressMessage(PROGRESS_MESSAGE_CREATEDATA);

    	CreatedDataSet cdSet = null;
        progress.startProgress();
        try {
            if (dataSource == null || colInfoSet == null || infoMap == null) {
            	return null;
            }

            // when information is empty, returns null
            if (infoMap.size() == 0) {
                return null;
            }
            
            final String dataType = (String) infoMap.get(KEY_DATA_TYPE);
            
            if (SGDataUtility.isArrayData(dataType)) {
                SGTextDataFile file = (SGTextDataFile) dataSource;
                if (SGDataUtility.isSXYTypeData(dataType)) {
                    Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
                    if (multiple == null) {
                        return null;
                    }
                    if (multiple) {
                        cdSet = this.createMultipleSXYData(
                                file, progress, colInfoSet, infoMap);
                    } else {
                        cdSet = this.createSingleSXYData(
                                file, progress, colInfoSet, infoMap);
                    }
                } else if (SGUtilityText.isEqualString(dataType,
                        SGDataTypeConstants.VXY_DATA)) {
                    cdSet = this.createVXYData(file, progress, colInfoSet, infoMap);
                } else if (SGUtilityText.isEqualString(dataType,
                        SGDataTypeConstants.SXYZ_DATA)) {
                    cdSet = this.createSXYZData(file, progress, colInfoSet, infoMap);
                }
				
            } else if (SGDataUtility.isNetCDFData(dataType)) {
                SGNetCDFFile file = (SGNetCDFFile) dataSource;
                if (SGDataUtility.isSXYTypeData(dataType)) {
                    Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
                    if (multiple == null) {
                        return null;
                    }
                    Boolean multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
                    if (multipleVariable == null) {
                        return null;
                    }
                    if (multipleVariable.booleanValue()) {
                        if (multiple.booleanValue()) {
                            cdSet = this.createMultipleVariableSXYNetCDFData(
                                    file, progress, colInfoSet, infoMap);
                        } else {
                            cdSet = this.createSingleVariableSXYNetCDFData(
                                    file, progress, colInfoSet, infoMap);
                        }
                    } else {
                        cdSet = this.createDimensionSXYNetCDFData(
                                file, progress, colInfoSet, infoMap, multiple.booleanValue());
                    }
                } else if (SGUtilityText.isEqualString(dataType,
                        SGDataTypeConstants.VXY_NETCDF_DATA)) {
                    cdSet = this.createVXYNetCDFData(file, progress, colInfoSet, infoMap);
                } else if (SGUtilityText.isEqualString(dataType,
                        SGDataTypeConstants.SXYZ_NETCDF_DATA)) {
                    cdSet = this.createSXYZNetCDFData(file, progress, colInfoSet, infoMap);
                }
            } else {
                return null;
            }

        } finally {
            progress.endProgress();
        }
        
        return cdSet;
    }
    
    /**
     * Creates a data object for netCDF dataset.
     * @param dataSource
     *           the data source.
     *           dataSource must be SGNetCDFFile because netCDF dataset.
     * @param dataType
     *           the data type of created data.
     * @param colInfoSet
     *           a set of column information
     * @param infoMap
     *            information of data
     * @param progress
     *            the progress monitor
     * @return a set of created data
     */
    public CreatedDataSet createForNetCDFDataSet(
            final SGIDataSource dataSource,
            final String dataType,
            final SGDataColumnInfoSet colInfoSet,
            final Map<String, Object> infoMap, final SGIProgressControl progress) {
        
        progress.setProgressMessage(PROGRESS_MESSAGE_CREATEDATA);

        CreatedDataSet cdSet = null;
        progress.startProgress();
        
        try {
            if (dataSource == null || colInfoSet == null || infoMap == null) {
                return null;
            }

            // when information is empty, returns null
            if (infoMap.size() == 0) {
                return null;
            }
            
            if (SGDataUtility.isArrayData(dataType)) {
                Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
                if (multiple == null) {
                    infoMap.put(KEY_MULTIPLE, Boolean.FALSE);
                }
                Boolean multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
                if (multipleVariable == null) {
                    infoMap.put(KEY_MULTIPLE_VARIABLE, Boolean.TRUE);
                }
            }
            
            SGNetCDFFile file = (SGNetCDFFile) dataSource;
            if (SGDataUtility.isSXYTypeData(dataType)) {
                Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
                if (multiple == null) {
                    return null;
                }
                Boolean multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
                if (multipleVariable == null) {
                    return null;
                }
                if (multipleVariable.booleanValue()) {
                    if (multiple.booleanValue()) {
                        cdSet = this.createMultipleVariableSXYNetCDFData(
                                file, progress, colInfoSet, infoMap);
                    } else {
                        cdSet = this.createSingleVariableSXYNetCDFData(
                                file, progress, colInfoSet, infoMap);
                    }
                } else {
                    cdSet = this.createDimensionSXYNetCDFData(
                            file, progress, colInfoSet, infoMap, multiple.booleanValue());
                }
            } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.VXY_NETCDF_DATA) ||
                    SGUtilityText.isEqualString(dataType, SGDataTypeConstants.VXY_DATA)) {
                cdSet = this.createVXYNetCDFData(file, progress, colInfoSet, infoMap);
            } else if (SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXYZ_NETCDF_DATA) ||
                    SGUtilityText.isEqualString(dataType, SGDataTypeConstants.SXYZ_DATA)) {
                cdSet = this.createSXYZNetCDFData(file, progress, colInfoSet, infoMap);
            }
        } finally {
            progress.endProgress();
        }
        
        return cdSet;
    }
}
