package jp.riken.brain.ni.samuraigraph.application;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingWindow;
import jp.riken.brain.ni.samuraigraph.base.SGFigure;
import jp.riken.brain.ni.samuraigraph.base.SGIDataSource;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElement;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementString;
import jp.riken.brain.ni.samuraigraph.base.SGIProgressControl;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.data.SGDataColumn;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGISXYTypeMultipleData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleDimensionNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGSXYMultipleVariableNetCDFData;
import jp.riken.brain.ni.samuraigraph.data.SGVariable;

class SGMainFunctionsSplitMerge {

    static boolean splitSXYDataSub(SGDrawingWindow wnd) {
        List<SGFigure> fList = wnd.getVisibleFigureList();

        SGIProgressControl progress = (SGIProgressControl) wnd;
        progress.endProgress();
        progress.startProgress();
        progress.setProgressMessage("Split Data");

        Map<SGFigure, List<SGData>> dataMap = new HashMap<SGFigure, List<SGData>>();
        try {
            for (SGFigure f : fList) {
                // get focused data objects
                SGIFigureElementGraph gElement = f.getGraphElement();
                List<SGData> dataList = gElement.getFocusedDataList();

                List<SGData> dataListAdded = splitData(f, dataList, progress);
                if (null!=dataListAdded) {
                    dataMap.put(f, dataListAdded);
                }
            }

        } finally {
            // end the progress bar
            progress.endProgress();
        }

        // hide the original data
        if (wnd.hideSelectedObjects() == false) {
            return false;
        }
        
        // make the new data focused
        Iterator<SGFigure> itr = dataMap.keySet().iterator();
        while (itr.hasNext()) {
            SGFigure f = itr.next();
            List<SGData> dList = dataMap.get(f);
            SGIFigureElementGraph gElement = f.getGraphElement();
            for (int ii = 0; ii < dList.size(); ii++) {
                SGData d = dList.get(ii);
                gElement.setDataFocused(d, true);
            }
        }
        
        wnd.notifyToRoot();

        return true;
    }
    
    /**
     * Split given data in a given figure.
     * 
     * @param figure
     * @param dataList
     * @param progress
     * @return list of data which is added by splitting multiple data.
     * or null if failed to split.
     */
    private static List<SGData> splitData(
            final SGFigure figure, final List<SGData> dataList, final SGIProgressControl progress) {
        // get figure elements
        SGIFigureElement[] elArray = figure.getIFigureElementArray();

        SGIFigureElementGraph gElement = figure.getGraphElement();
        SGIFigureElementString fes = (SGIFigureElementString) figure.getIFigureElement(SGIFigureElementString.class);
        
        List<SGData> dList = new ArrayList<SGData>();
        
        for (SGData data : dataList) {

            if (data instanceof SGISXYTypeMultipleData) {

                final boolean bArray = SGDataUtility.isArrayData(data);
                final boolean nArray = SGDataUtility.isNetCDFData(data);
                if ((!bArray && !nArray) || (bArray && nArray)) {
                    throw new Error("Unsupported data type: " + data.getDataType());
                }
                SGISXYTypeMultipleData dataMult = (SGISXYTypeMultipleData) data;
                String dataNameMult = gElement.getDataName(data);
                
                SGISXYTypeMultipleData[] subArray = dataMult.getSXYTypeMultipleDataArray();
                if (subArray.length == 1) {
                	subArray[0].dispose();
                    continue;
                }
                SGData[] dataArray = new SGData[subArray.length];
                for (int ii = 0; ii < dataArray.length; ii++) {
                    dataArray[ii] = (SGData) subArray[ii];
                }
                
                String[] nameArray = getNameOfSplittedData(dataMult, dataArray, dataNameMult, bArray);
                SGProperties pMult = gElement.getDataProperties((SGData) dataMult);
                
                final float diff = 1.0f / (dataArray.length + 1);
                for (int ii = 0; ii < dataArray.length; ii++) {
                    if (null!=progress) {
                        progress.setProgressValue((ii + 1) * diff);
                    }

                    // get data properties
                    SGProperties dp = dataArray[ii].getProperties();

                    Map<Class<? extends SGIFigureElement>, SGProperties> propertiesMap =
                        new HashMap<Class<? extends SGIFigureElement>, SGProperties>();
                    for (int jj = 0; jj < elArray.length; jj++) {

                        // synchronize the properties for duplicated data object
                        SGProperties sp = elArray[jj].synchronizeDataProperties(pMult, dp);
                        if (sp == null) {
                            continue;
                        }

                        // put into the map
                        propertiesMap.put(elArray[jj].getClass(), sp);
                    }
                    if (figure.addData(dataArray[ii], nameArray[ii], propertiesMap) == false) {
                        return null;
                    }
                }
                
                for (int ii = 0; ii < dataArray.length; ii++) {
                    dList.add(dataArray[ii]);
                }
                
                gElement.hideNetCDFLabels(fes, (SGData) dataMult);
            }
        }
        
        return dList;
    }
    
    static List<SGData> splitDataInFigure(final SGFigure figure, final int[] dataIdArray) {
        List<SGData> dataList = new ArrayList<SGData>();
        for (int ii = 0; ii < dataIdArray.length; ii++) {
            SGData data = figure.getData(dataIdArray[ii]);
            if (data == null) {
                return null;
            }
            dataList.add(data);
        }
        
        List<SGData> dataListAdded = splitData(figure, dataList, null);
        
        figure.hideData(dataIdArray);
        if (figure.isChangedRoot()) {
            figure.notifyToRoot();
        }
        figure.repaint();
        
        return dataListAdded;
    }
    
    /**
     * Returns the names of split data.
     * 
     * @param dataMult
     * @param dataArray
     * @param dataNameMult
     * @param bArray
     * @return the names of split data
     */
    static String[] getNameOfSplittedData(
            final SGISXYTypeMultipleData dataMult,
            final SGData[] dataArray,
            final String dataNameMult, final boolean bArray) {
        
        String[] nameArray = new String[dataArray.length];
        
        for (int ii = 0; ii < dataArray.length; ii++) {
            StringBuffer sb = new StringBuffer();
            sb.append(dataNameMult);
            String suffix = "";
            if (bArray) {
                SGSXYMultipleData aData = (SGSXYMultipleData) dataMult;
                SGDataColumn[] cols = aData.getMultipleColumns();
                String title = cols[ii].getTitle();
                if (title == null || "".equals(title)) {
                    suffix = Integer.toString(ii + 1);
                } else {
                    suffix = SGUtility.addEscapeChar(title);
                }
                
                sb.append("\\_");
                sb.append(suffix);
            } else {
                if (dataMult instanceof SGSXYMultipleVariableNetCDFData) {
                    SGSXYMultipleVariableNetCDFData nData = (SGSXYMultipleVariableNetCDFData) dataMult;
                    SGVariable[] vars = nData.getMultipleVariables();
                    suffix = SGUtility.addEscapeChar(vars[ii].getName());
                    
                    sb.append("\\_");
                    sb.append(suffix);
                } else if (dataMult instanceof SGSXYMultipleDimensionNetCDFData) {
                    SGSXYMultipleDimensionNetCDFData nData = (SGSXYMultipleDimensionNetCDFData) dataMult;
                    final int start = nData.getDimensionStart();
                    final int end = nData.getDimensionEnd();
                    final int step = Math.abs(nData.getDimensionStep());
                    int index = start;
                    if (start <= end) {
                        index = start + step * ii;
                        if (index > end) {
                            index = end;
                        }
                    } else {
                        index = start - step * ii;
                        if (index < end) {
                            index = end;
                        }
                    }
                    Double dvalue = nData.getDimensionValue(index);
                    if (null==dvalue) {
                        suffix = Integer.toString(index);
                        
                        sb.append("\\_");
                        sb.append(suffix);
                    } else {
                        double val = dvalue.doubleValue();
                        sb.append(" (");
                        sb.append(SGUtility.addEscapeChar(nData.getDimensionName()));
                        sb.append('=');
                        sb.append(SGUtility.getNumberName(val));
                        sb.append(")");
                    }
                }
            }
            nameArray[ii] = sb.toString();
        }
        
        return nameArray;
    }
    
    static boolean mergeSXYDataSub(SGDrawingWindow wnd) {
        List<SGFigure> fList = wnd.getVisibleFigureList();

        SGIProgressControl progress = (SGIProgressControl) wnd;
        progress.endProgress();
        progress.startProgress();
        progress.setProgressMessage("Merge Data");

        Map<SGFigure, List<SGData>> dataMap = new HashMap<SGFigure, List<SGData>>();
        try {
            for (SGFigure f : fList) {
                // get focused data objects
                SGIFigureElementGraph gElement = f.getGraphElement();
                List<SGData> dataList = gElement.getFocusedDataList();

                List<SGData> dataListAdded = mergeData(f, dataList, progress);
                if (null==dataListAdded || dataListAdded.size()==0) {
                    continue;
                }
                dataMap.put(f, dataListAdded);
            }
        } finally {
            // end the progress bar
            progress.endProgress();
        }

        // hide the original data
        if (wnd.hideSelectedObjects() == false) {
            return false;
        }

        // make the new data focused
        Iterator<SGFigure> itr = dataMap.keySet().iterator();
        while (itr.hasNext()) {
            SGFigure f = itr.next();
            List<SGData> dList = dataMap.get(f);
            SGIFigureElementGraph gElement = f.getGraphElement();
            for (int ii = 0; ii < dList.size(); ii++) {
                SGData d = dList.get(ii);
                gElement.setDataFocused(d, true);
            }
        }

        wnd.notifyToRoot();

        return true;
    }
    
    private static List<SGData> mergeData(
            final SGFigure figure, final List<SGData> dataList, final SGIProgressControl progress) {
        
        // get figure elements
        SGIFigureElement[] elArray = figure.getIFigureElementArray();

        SGIFigureElementGraph gElement = figure.getGraphElement();
//        SGIFigureElementString fes = (SGIFigureElementString) figure.getIFigureElement(SGIFigureElementString.class);
        
        Map<SGIDataSource, List<SGData>> map = new HashMap<SGIDataSource, List<SGData>>();
        for (SGData data : dataList) {
            if (data instanceof SGISXYTypeMultipleData) {
                SGIDataSource dataSource = data.getDataSource();
                List<SGData> store = map.get(dataSource);
                if (null==store) {
                    store = new ArrayList<SGData>();
                    map.put(dataSource, store);
                }
                store.add(data);
            }
        }
        
//        gElement.hideNetCDFLabelsOfFocusedObjects(fes);
        
        List<SGData> dList = new ArrayList<SGData>();
        
        int count = 0;
        final float diff = 1.0f / (map.size() + 1);
        for (SGIDataSource source : map.keySet()) {
            if (null!=progress) {
                progress.setProgressValue((count + 1) * diff);
            }
            
            List<SGData> store = map.get(source);
            if (store.size()<=1) {
                continue;
            }
            
            SGISXYTypeMultipleData dataMult = null;
            
            SGData dataLast = store.get(store.size()-1);
            SGProperties pLast = gElement.getDataProperties(dataLast);
            final boolean bArray = SGDataUtility.isArrayData(dataLast);
            
            Map<Class<? extends SGIFigureElement>, SGProperties> propertiesMap =
                new HashMap<Class<? extends SGIFigureElement>, SGProperties>();
            
            if (bArray) {
                dataMult = SGSXYMultipleData.merge(store);
            } else {
                if (dataLast instanceof SGSXYMultipleVariableNetCDFData) {
                    dataMult = SGSXYMultipleVariableNetCDFData.merge(store);
                } else if (dataLast instanceof SGSXYMultipleDimensionNetCDFData) {
                    dataMult = SGSXYMultipleDimensionNetCDFData.merge(store);
                }
            }
            String name = getNameOfMergedData((SGData)dataMult);
            
            if (null!=dataMult) {
                // get data properties
                SGProperties pMult = ((SGData)dataMult).getProperties();

                for (int jj = 0; jj < elArray.length; jj++) {
                    // synchronize the properties for duplicated data object
                    SGProperties sp = elArray[jj].synchronizeDataProperties(pLast, pMult);
                    if (sp == null) {
                        continue;
                    }
                    propertiesMap.put(elArray[jj].getClass(), sp);
                }
                if (figure.addData((SGData)dataMult, name, propertiesMap) == false) {
                    return null;
                }

                dList.add((SGData)dataMult);
            }
            
            count++;
        }
        
        return dList;
    }
    
    static List<SGData> mergeDataInFigure(final SGFigure figure, final int[] dataIdArray) {
        List<SGData> dataList = new ArrayList<SGData>();
        for (int ii = 0; ii < dataIdArray.length; ii++) {
            SGData data = figure.getData(dataIdArray[ii]);
            if (data == null) {
                return null;
            }
            dataList.add(data);
        }
        
        // set focus to data
        SGIFigureElementGraph gElement = figure.getGraphElement();
        gElement.clearFocusedObjects();
        for (int ii = 0; ii < dataList.size(); ii++) {
            gElement.setDataFocused(dataList.get(ii), true);
        }
        
        List<SGData> dataListAdded = mergeData(figure, dataList, null);
        
        figure.hideData(dataIdArray);
        if (figure.isChangedRoot()) {
            figure.notifyToRoot();
        }
        figure.repaint();
        
        return dataListAdded;
    }
    
    /**
     * Returns the name of the merged data.
     * 
     * @param data merged data
     * @return the name of the merged data.
     */
    static String getNameOfMergedData(final SGData data) {
        String path = null;
        if (data instanceof SGSXYMultipleData) {
            path = ((SGSXYMultipleData)data).getDataFile().getPath();
        } else if (data instanceof SGSXYMultipleVariableNetCDFData) {
            path = ((SGSXYMultipleVariableNetCDFData)data).getNetcdfFile().getNetcdfFile().getLocation();
        } else if (data instanceof SGSXYMultipleDimensionNetCDFData) {
            path = ((SGSXYMultipleDimensionNetCDFData)data).getNetcdfFile().getNetcdfFile().getLocation();
        }
        if (null!=path) {
            return SGUtility.createDataNameBase(path);
        } else {
            return "MergedData";
        }
    }

}
