package jp.riken.brain.ni.samuraigraph.base;

import java.lang.reflect.InvocationTargetException;
import java.util.ArrayList;
import java.util.List;

import javax.swing.SwingUtilities;

/**
 * A class of the animation.
 * 
 */
public class SGAnimationThread extends Thread implements SGIDisposable {

    /**
     * An object which has data sources of the animation.
     */
    protected SGIAnimation mAnimation = null;

    /**
     * A flag whether to play the animation.
     */
    protected boolean mPlayFlag = false;

    /**
     * A flag whether to loop playback.
     */
    protected boolean mLoopPlaybackFlag = false;

    /**
     * The step size.
     */
    protected int mStep = 1;

    /**
     * The time interval.
     */
    protected int mTimeInterval = 1000;

    /**
     * Builds an animation thread.
     * 
     * @param animation
     *           an animation data source
     */
    public SGAnimationThread(SGIAnimation animation) {
        super();
        if (animation == null) {
            throw new IllegalArgumentException("animation == null");
        }
        this.mAnimation = animation;
    }

    /**
     * Builds an animation thread with the same as given animation thread.
     * 
     * @param th
     *          an animation thread to copy
     */
    public SGAnimationThread(SGAnimationThread th) {
        super();
        this.mAnimation = th.mAnimation;
        this.mTimeInterval = th.mTimeInterval;
        this.mStep = th.mStep;
        this.mLoopPlaybackFlag = th.mLoopPlaybackFlag;
        this.mPlayFlag = th.mPlayFlag;
        this.mObserverList = new ArrayList<AnimationThreadObserver>(th.mObserverList);
    }

    /**
     * Returns the animation data source.
     * 
     * @return the animation data source
     */
    public SGIAnimation getAnimation() {
        return this.mAnimation;
    }
    
    private static final long MIN_SLEEP_TIME = 10L;

    /**
     * Run this thread.
     */
    public void run() {

        // get the number of frame
        final int fSize = this.mAnimation.getFrameNumber();

        PlayInfo info = this.getPlayInfo();
        
        // endless loop for the animation
        while (this.mPlayFlag) {

            // current index
            final int curIndex = this.mAnimation.getCurrentFrameIndex();

            // increment the counter
            int nextIndex = curIndex + this.mStep;

            // when reached to the max / min index
            boolean outOfRange = false;
            if (this.mStep > 0) {
                if (nextIndex >= fSize) {
                    outOfRange = true;
                }
            } else {
                if (nextIndex < 0) {
                    outOfRange = true;
                }
            }
            if (outOfRange) {
                if (this.mLoopPlaybackFlag) {
                    if (this.mStep > 0) {
                        nextIndex %= fSize;
                    } else {
                        final int num = Math.abs(nextIndex) / fSize + 1;
                        nextIndex += num * fSize;
                    }
                } else {
                    this.mPlayFlag = false;
                    continue;
                }
            }

            // set the frame index
            final int nIndex = nextIndex;
            try {
                SwingUtilities.invokeAndWait(new Runnable() {
                    public void run() {
                        mAnimation.setCurrentFrameIndex(nIndex);
                    }
                });
            } catch (InterruptedException e1) {
            } catch (InvocationTargetException e1) {
            }

            // notify to the observers
            for (AnimationThreadObserver obs : this.mObserverList) {
                obs.updated(this);
            }
            
            // sleep seconds for time step
            try {
                final long initTime = System.currentTimeMillis();
                
                // loops of minimum sleeping time
                boolean endFlag = false;
                for (int ii = 0; ii < info.repeatNum; ii++) {
                    info = this.getPlayInfo();
                    if (!this.mPlayFlag) {
                        break;
                    }
                    Thread.sleep(MIN_SLEEP_TIME);
                    final long curTime = System.currentTimeMillis();
                    final long spentTime = curTime - initTime;
                    if (spentTime > this.mTimeInterval) {
                        endFlag = true;
                        break;
                    }
                }

                if (endFlag) {
                    continue;
                }

                info = this.getPlayInfo();
                if (!this.mPlayFlag) {
                    break;
                }
                
                // sleep remained time
                Thread.sleep(info.remainedTime);
                
            } catch (InterruptedException e) {
            }
        }
    }

    // a class for the information of playing
    private static class PlayInfo {
        int repeatNum;
        long remainedTime;
    }
    
    private PlayInfo getPlayInfo() {
        int repeatNum;
        long rTime;
        if (this.mTimeInterval < MIN_SLEEP_TIME) {
            repeatNum = 0;
            rTime = this.mTimeInterval;
        } else {
            repeatNum = (int) (this.mTimeInterval / MIN_SLEEP_TIME);
            rTime = this.mTimeInterval - repeatNum * MIN_SLEEP_TIME;
        }
        PlayInfo info = new PlayInfo();
        info.repeatNum = repeatNum;
        info.remainedTime = rTime;
        return info;
    }

    /**
     * Returns whether the animation is playing.
     */
    public boolean isPlaying() {
        return this.mPlayFlag;
    }

    /**
     * Sets whether to play the animation.
     * 
     * @param b
     *          true to play the animation
     */
    public void setPlaying(boolean b) {
        this.mPlayFlag = b;
    }

    /**
     * Returns whether loop playback is available.
     * 
     * @return true the loop playback is available
     */
    public boolean isLoopPlaybackEnabled() {
        return this.mLoopPlaybackFlag;
    }

    /**
     * Sets whether loop playback is available.
     * 
     * @param b
     *           true to enbale the loop playback
     */
    public void setLoopPlaybackEnabled(final boolean b) {
        this.mLoopPlaybackFlag = b;
    }

    /**
     * Returns the step number.
     * 
     * @return the step number
     */
    public int getStep() {
        return this.mStep;
    }

    /**
     * Sets the step number.
     * 
     * @param s
     *          a number to set to the step number
     */
    public void setStep(final int s) {
        if (s == 0) {
            throw new IllegalArgumentException("s == 0");
        }
        this.mStep = s;
    }

    /**
     * Returns the time interval.
     * 
     * @return the sleeping time
     */
    public long getTimeInterval() {
        return this.mTimeInterval;
    }

    /**
     * Sets the sleeping time.
     * 
     * @param t
     *          a value to set to the sleeping time
     */
    public void setTimeInterval(final int t) {
        if (t < 0) {
            throw new IllegalArgumentException("t < 0");
        }
        this.mTimeInterval = t;
    }

    /**
     * Disposes of this object.
     */
    public void dispose() {
    	this.mDisposed = true;
        this.mObserverList.clear();
    }
    
    // The flag whether this object is already disposed of.
    private boolean mDisposed = false;

    /**
     * Returns whether this object is already disposed of.
     * 
     * @return true if this object is already disposed of
     */
    public boolean isDisposed() {
    	return this.mDisposed;
    }
    
    /**
     * Returns the number of frames.
     * 
     * @return the number of frames
     */
    public int getFrameNumber() {
        return this.mAnimation.getFrameNumber();
    }

    /**
     * Returns the current frame index.
     * 
     * @return the current frame index
     */
    public int getCurrentFrameIndex() {
        return this.mAnimation.getCurrentFrameIndex();
    }

    /**
     * Sets the current frame index.
     * 
     * @param index the frame index to be set
     */
    public void setCurrentFrameIndex(final int index) {
        this.mAnimation.setCurrentFrameIndex(index);
    }

    /**
     * An interface of the observer of this thread.
     * @author kuromaru
     *
     */
    public static interface AnimationThreadObserver {
        /**
         * Called when a thread cycle is looped.
         * 
         * @param th
         *          an animation thread
         */
        public void updated(SGAnimationThread th);
    }

    /**
     * List of observers of this thread.
     */
    private List<AnimationThreadObserver> mObserverList = new ArrayList<AnimationThreadObserver>();
    
    /**
     * Add an observer of this thread.
     * 
     * @param obs
     *           an observer
     */
    public void addAnimationThreadObserver(AnimationThreadObserver obs) {
        this.mObserverList.add(obs);
    }
    
    /**
     * Remove an observer of this thread.
     * 
     * @param obs
     *           an observer
     */
    public void removeAnimationThreadObserver(AnimationThreadObserver obs) {
        this.mObserverList.remove(obs);
    }
}
