package jp.riken.brain.ni.samuraigraph.base;

import java.awt.Color;

/**
 * A model for the color bar.
 * A color bar connects an array of colors and a range of double values.
 * @author kuromaru
 *
 */
public abstract class SGColorBarModel implements Cloneable, SGIDisposable {

    /**
     * The axis.
     */
    protected SGAxis mAxis = null;

    /**
     * An array of colors.
     */
    protected Color[] mColors = null;
    
    /**
     * A flag whether the order of colors is revered.
     */
    protected boolean mReversedOrderFlag = false;
    
    /**
     * The default constructor.
     *
     */
    public SGColorBarModel() {
        super();
        
        // initialize
        this.init();
    }
    
    /**
     * Builds a color map. 
     * 
     * @param colors
     *          an array of colors
     */
    public SGColorBarModel(final Color[] colors) {
        
        super();

        if (colors == null) {
            throw new IllegalArgumentException("Color array is null.");
        }
        if (colors.length == 0) {
            throw new IllegalArgumentException("Length of a given color array is zero.");
        }
        for (int ii = 0; ii < colors.length; ii++) {
            if (colors[ii] == null) {
                throw new IllegalArgumentException("Null value is included in a given array: " + ii);
            }
        }
        
        // set to attributes
        this.mColors = (Color[]) colors.clone();
        
        // initialize
        this.init();
    }
    
    private void init() {
        this.mReversedOrderFlag = false;
        this.mAxis = new SGAxis(0.0, 1.0);
    }
    
    /**
     * Disposes of this object.
     */
    public void dispose() {
    	this.mDisposed = true;
        this.mColors = null;
    }
    
    // The flag whether this object is already disposed of.
    private boolean mDisposed = false;

    /**
     * Returns whether this object is already disposed of.
     * 
     * @return true if this object is already disposed of
     */
    public boolean isDisposed() {
    	return this.mDisposed;
    }
    
    /**
     * Clones this color bar model.
     * 
     * @return
     *         a copy of this object
     */
    public final Object clone() {
        try {
            SGColorBarModel model = (SGColorBarModel) super.clone();
            model.mAxis = (SGAxis) this.mAxis.clone();
            if (this.mColors != null) {
                model.mColors = (Color[]) this.mColors.clone();
            }
            return model;
        } catch (CloneNotSupportedException ex) {
            // this shouldn't happen, since we are Cloneable
            throw new InternalError();
        }
    }

    /**
     * Returns the axis.
     * 
     * @return the axis
     */
    public SGAxis getAxis() {
        return this.mAxis;
    }

    /**
     * Returns the minimum value.
     * 
     * @return the minimum value
     */
    public double getMinValue() {
        return this.mAxis.getMinValue();
    }
    
    /**
     * Returns the maximum value.
     * 
     * @return the maximum value
     */
    public double getMaxValue() {
        return this.mAxis.getMaxValue();
    }

    /**
     * Returns the scale type.
     * 
     * @return the scale type
     */
    public int getScaleType() {
        return this.mAxis.getScaleType();
    }
    
    /**
     * Returns whether the coordinate is inverted.
     * 
     * @return true if the coordinate is inverted
     */
    public boolean isInvertCoordinates() {
        return this.mAxis.isInvertCoordinates();
    }
    
    /**
     * Sets the value range. The minimum value corresponds to the first element 
     * and the maximum value corresponds to the last element respectively of given color array.
     * 
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     */
    public void setValueRange(final double min, final double max) {
        this.setValueRange(min, max, this.mAxis.getScaleType());
    }
    
    /**
     * Sets the value range and the scale type.
     * 
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     * @param scaleType
     *           the scale type
     */
    public void setValueRange(final double min, final double max, 
            final int scaleType) {
        this.mAxis.setScale(min, max, scaleType);
    }

    /**
     * Sets whether the coordinate is inverted.
     * 
     * @param b
     *          true to be inverted
     */
    public void setInvertCoordinates(final boolean b) {
        this.mAxis.setInvertCoordinates(b);
    }
    
    /**
     * Returns an array of colors.
     * 
     * @return an array of colors
     */
    public Color[] getColors() {
        if (this.mColors != null) {
            return (Color[]) this.mColors.clone();
        } else {
            return null;
        }
    }
    
    /**
     * Returns whether the order of colors is reversed.
     * 
     * @return whether the order of colors is reversed
     */
    public boolean isReversedOrder() {
        return this.mReversedOrderFlag;
    }
    
    /**
     * Sets whether the order of colors is reversed.
     * 
     * @param b
     *          true to reverse the order of colors
     */
    public void setReversedOrder(final boolean b) {
        this.mReversedOrderFlag = b;
    }

    /**
     * Evaluate a given value with the color map and returns a corresponding color.
     * 
     * @param value
     *           a value to evaluate
     * @return corresponding color
     */
    public Color evaluate(final double value) {
        final int scaleType = this.getScaleType();
        return this.evaluate(value, scaleType);
    }
    
    /**
     * Evaluate a given value with the color map and returns a corresponding color.
     * 
     * @param value
     *           a value to evaluate
     * @param scaleType
     *           the scale type
     * @return a corresponding color
     */
    public Color evaluate(final double value, final int scaleType) {
        if (Double.isNaN(value) || Double.isInfinite(value)) {
            throw new IllegalArgumentException("Invalid input value: " + value);
        }
        final double minValue = this.getMinValue();
        final double maxValue = this.getMaxValue();
        final double valueInScale, minValueInScale, maxValueInScale;
        switch (scaleType) {
        case SGAxis.LINEAR_SCALE :
            valueInScale = value;
            minValueInScale = minValue;
            maxValueInScale = maxValue;
            break;
        case SGAxis.LOG_SCALE :
            if (value <= 0.0) {
                throw new IllegalArgumentException("Not positive value for log scale: " + value);
            }
            valueInScale = Math.log10(value);
            minValueInScale = Math.log10(minValue);
            maxValueInScale = Math.log10(maxValue);
            break;
        default :
            throw new IllegalArgumentException("Invalid scale type: " + scaleType);
        }
        final double ratioAll = (valueInScale - minValueInScale)
                / (maxValueInScale - minValueInScale);
        return this.eval(ratioAll);
    }
    
    /**
     * Returns the color in an array of colors of attribute.
     * 
     * @param index
     *           the array index
     * @return the array element
     */
    protected Color getColor(final int index) {
        if (this.mReversedOrderFlag) {
            // return the last element
            return this.mColors[this.mColors.length - 1 - index];
        } else {
            // returns the first element
            return this.mColors[index];
        }
    }
    
    /**
     * Returns the corresponding color for a given value.
     * The input value should be in a scale that the minumum value and the maximum value
     * in the current scale is given by 0.0 and 1.0, respectively.
     * 
     * @param value
     *           a value
     * @return the corresponding color
     */
    public abstract Color eval(final double value);

}
