package jp.riken.brain.ni.samuraigraph.base;

import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.HashMap;
import java.util.Map;

import org.w3c.dom.Element;

/**
 * The base class of all data classes.
 */
public abstract class SGData implements Cloneable, SGICopiable, SGIData, SGIDisposable {

    /**
     * The default constructor.
     */
    public SGData() {
        super();
    }
    
    /**
     * Clones this data object.
     * 
     * @return
     *         shallow copy of this data object
     */
    public final Object clone() {
        try {
            return super.clone();
        } catch (CloneNotSupportedException ex) {
            // this shouldn't happen, since we are Cloneable
            throw new InternalError();
        }
    }
    
    /**
     * Calling the clone method.
     * 
     * @return
     *         shallow copy of this data object
     */
    public Object copy() {
        return this.clone();
    }
    
    /**
     * Set values with a given data.
     * 
     * @param data
     *             data to be copied
     * @return
     *             true if succeeded
     */
    public abstract boolean setData(final SGData data);

    /**
     * Write this data to a given output stream.
     * @param w
     *          an output stream
     * @param mode
     *          the mode of saving data
     * @return
     *          true if succeeded
     */
    protected abstract boolean writeData(Writer w, int mode) throws IOException;

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
    	this.mDisposed = true;
    }

    // The flag whether this object is already disposed of.
    private boolean mDisposed = false;

    /**
     * Returns whether this object is already disposed of.
     * 
     * @return true if this object is already disposed of
     */
    public boolean isDisposed() {
    	return this.mDisposed;
    }
    
    /**
     * Write properties of this data to a given Element.
     * @param el
     *          an Element
     * @param type
     *            type of the method to save properties
     * @return
     *          true if succeeded
     */
    public abstract boolean writeProperty(Element el, int type);
    
    /**
     * Read properties from a given Element and set to this data.
     * @param el
     *          an Element
     * @return
     *          true if succeeded
     */
    public abstract boolean readProperty(Element el);

    /**
     * Save values to given file.
     * 
     * @param data
     *             a data
     * @param mode
     *             the mode of saving data
     * @return
     *             true if succeeded
     */
    public abstract boolean saveData(final File file, final int mode);
    
    /**
     * Returns a map which has data information.
     * This method returns an empty map by default.
     * Subclasses overrode this method if necessary.
     * @return
     *        a map which has data information
     */
    public Map<String, Object> getInfoMap() {
        return new HashMap<String, Object>();
    }
    
    /**
     * Returns properties of this data.
     * @return
     *         properties of this data
     */
    public abstract SGProperties getProperties();

    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public abstract boolean getProperties(final SGProperties p);
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public abstract boolean setProperties(final SGProperties p);

    /**
     * A class for data properties.
     */
    public static abstract class DataProperties extends SGProperties {

        /**
         * The default constructor.
         */
        public DataProperties() {
            super();
        }
    }
    
    /**
     * Sets the column information.
     * 
     * @param colInfo
     *           the column information
     * @param true if succeeded
     */
    public boolean setColumnInfo(SGDataColumnInfo[] colInfo) {
        // set the column type
        String[] types = new String[colInfo.length];
        for (int ii = 0; ii < types.length; ii++) {
            types[ii] = colInfo[ii].getColumnType();
        }
        if (this.setColumnType(types) == false) {
        	return false;
        }
        return true;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *              an array of column types
     * @return true if succeeded
     */
    public abstract boolean setColumnType(String[] columns);

//    /**
//     * Returns whether all given data columns are valid.
//     * 
//     * @param colInfo
//     *           data columns
//     * @return true if all given data columns are valid
//     */
//    public boolean isValidColumn(SGDataColumnInfo[] colInfo) {
//        String[] types = new String[colInfo.length];
//        for (int ii = 0; ii < types.length; ii++) {
//            types[ii] = colInfo[ii].getColumnType();
//        }
//    	return this.isValidColumn(types);
//    }
//
//    /**
//     * Returns whether all given data column types are valid.
//     * 
//     * @param types
//     *           type of data columns
//     * @return true if all given data columns are valid
//     */
//    public boolean isValidColumn(String[] types) {
//    	// returns true by default
//    	return true;
//    }

}
