package jp.riken.brain.ni.samuraigraph.base;

import java.util.ArrayList;
import java.util.List;

/**
 * An object to manage axes.
 */
public interface SGIFigureElementAxis extends SGIFigureElement {

    /**
     * An axis value that is used when the minimum value and the maximum value 
     * is very close.
     * 
     */
    public static final double ALTERNATIVE_AXIS_VALUE = 1.0;

    /**
     * Code for axis location.
     */
    public static final int AXIS_HORIZONTAL_1 = 0;

    public static final int AXIS_HORIZONTAL_2 = 1;

    public static final int AXIS_PERPENDICULAR_1 = 2;

    public static final int AXIS_PERPENDICULAR_2 = 3;
    
    public static final int AXIS_NORMAL = 4;
    
    /**
     * Brief description of axis location
     */
    public static final String AXIS_TOP = "Top";

    public static final String AXIS_BOTTOM = "Bottom";

    public static final String AXIS_LEFT = "Left";

    public static final String AXIS_RIGHT = "Right";
    
    public static final String AXIS_COLOR_BAR = "ColorBar";

    /**
     * The table connecting a number code and a brief description of each axis location.
     */
    public static final String[] AXIS_LOCATION_TABLE = {
        AXIS_BOTTOM, AXIS_TOP, AXIS_LEFT, AXIS_RIGHT, AXIS_COLOR_BAR
    };
    
    /**
     * Brief description of axis pairs
     */
    public static final String LEFT_TOP = "Left-Top";

    public static final String LEFT_BOTTOM = "Left-Bottom";

    public static final String RIGHT_TOP = "Right-Top";

    public static final String RIGHT_BOTTOM = "Right-Bottom";

    /**
     * Command to show or hide an axis
     */
    public static final String MENUCMD_SHOW_AXIS = "Show Axis";

    /**
     * Command to draw image later
     */
    public static final String MENUCMD_DRAW_LATER = "Draw Later";

    /**
     * Command to hide the color bar
     */
    public static final String MENUCMD_HIDE_COLOR_BAR = "Hide";

    /**
     * Returns the list of all axes.
     * 
     * @return the list of all axes
     */
    public ArrayList getAxisList();

    /**
     * Returns the list of horizontal axes.
     * 
     * @return the list of horizontal axes
     */
    public ArrayList getHorizontalAxisList();

    /**
     * Returns the list of perpendicular axes.
     * 
     * @return the list of perpendicular axes
     */
    public ArrayList getPerpendicularAxisList();

    /**
     * Returns the list of normal axes.
     * 
     * @return the list of normal axes
     */
    public ArrayList getNormalAxisList();

    /**
     * Returns whether a given axis is horizontal.
     * 
     * @param axis
     *            an axis
     * @return true if a given axis is horizontal
     */
    public boolean isHorizontal(final SGAxis axis);

    /**
     * Returns whether a given axis is perpendicular.
     * 
     * @param axis
     *            an axis
     * @return true if a given axis is perpendicular
     */
    public boolean isPerpendicular(final SGAxis axis);

    /**
     * Returns whether a given axis is normal.
     * 
     * @param axis
     *           an axis
     * @return true if a given axis is normal
     */
    public boolean isNormal(final SGAxis axis);

    /**
     * Returns string representation of the axis location.
     * 
     * @param locationInPlane -
     *            One of the following parameters: AXIS_HORIZONTAL_1,
     *            AXIS_HORIZONTAL_2, AXIS_PERPENDICULAR_1 and
     *            AXIS_PERPENDICULAR_2
     * @return String representation of the axis location
     */
    public String getLocationName(final int locationInPlane);

    /**
     * Returns string representation of the axis location.
     * 
     * @param axis
     *            an axis
     * @return String representation of the axis location
     */
    public String getLocationName(final SGAxis axis);

    /**
     * Returns code of the location in a plane of a given axis.
     * 
     * @param axis
     *            an axis
     * @return the axis location
     */
    public int getLocationInPlane(final SGAxis axis);

    /**
     * Returns an axis at a given location.
     * 
     * @param locationInPlane
     *            string representation of the location of an axis in a plane
     * @return an axis object
     */
    public SGAxis getAxisInPlane(final int locationInPlane);

    /**
     * Returns an axis at a given location.
     * 
     * @param locationInPlane
     *            the location of an axis in a plane
     * @return an axis object
     */
    public SGAxis getAxis(final String str);

    /**
     * Returns the z-axis.
     * 
     * @return the z-axis
     */
    public SGAxis getZAxis();
    
    /**
     * Set the space between the scale lines and the scale numbers.
     * @param space
     *       a value to set the space in units of pt
     * @return
     *       true if succeeded
     */
    public boolean setSpaceAxisLineAndNumber(float space);
    
    /**
     * Set the space between the scale lines and the scale numbers in a given unit.
     * @param space
     *       a value to set the space
     * @param unit
     *       unit of length
     * @return
     *       true if succeeded
     */
    public boolean setSpaceAxisLineAndNumber(float space, String unit);

    /**
     * Set the space between the scale numbers and the title.
     * @param space
     *       a value to set the space in units of pt
     * @return
     *       true if succeeded
     */
    public boolean setSpaceNumberAndTitle(float space);
    
    /**
     * Set the space between the scale numbers and the title in a given unit.
     * @param space
     *       a value to set the space
     * @param unit
     *       unit of length
     * @return
     *       true if succeeded
     */
    public boolean setSpaceNumberAndTitle(float space, String unit);

    /**
     * @uml.property name="spaceAxisLineAndNumber"
     */
    public float getSpaceAxisLineAndNumber();

    /**
     * @uml.property name="spaceNumberAndTitle"
     */
    public float getSpaceNumberAndTitle();

    /**
     * 
     */
    public float getScaleFontSize();

    /**
     * 
     */
    public float getTitleFontSize();

    /**
     * Returns whether the color bar is available.
     * 
     * @return true if the color bar is available
     */
    public boolean isColorBarAvailable();
    
    /**
     * Returns the color bar model.
     * 
     * @return the color bar model
     */
    public SGColorBarModel getColorBarModel();

    /**
     * Returns the color bar if it exists.
     * 
     * @return the color bar
     */
    public SGIColorBar getColorBar();

    /**
     * Returns whether the color bar is visible.
     * 
     * @return true if visible
     */
    public boolean isColorBarVisible();

    /**
     * Sets whether the color bar is visible.
     * 
     * @param true to set visible
     * 
     * @return true if succeeded
     */
    public boolean setColorBarVisible(final boolean b);

    /**
     * Fits the range of axes range to the focused data.
     * 
     * @param element
     *           a figure element for data
     * @return true if succeeded
     */
    public boolean fitAxisRangeToFocusedData(SGIFigureElementForData element);

    /**
     * Fits the range of axes range to the given data.
     * 
     * @param element
     *           a figure element for data
     * @param dataList
     *           a list of data
     * @return true if succeeded
     */
    public boolean fitAxisRangeToData(SGIFigureElementForData element, 
    		List<SGData> dataList);
    
    /**
     * 
     * @param config
     * @param x
     * @param y
     * @return
     */
    public double getValue(final int config, final int x, final int y);

    public static final String TAG_NAME_AXES = "Axes";

    public static final String KEY_LINE_COLOR = "LineColor";

    public static final String KEY_AXIS_LINE_WIDTH = "AxisLineWidth";

    public static final String KEY_TICK_MARK_WIDTH = "TickMarkWidth";

    public static final String KEY_TICK_MARK_LENGTH = "TickMarkLength";

    public static final String KEY_FRAME_LINE_VISIBLE = "FrameLineVisible";

    public static final String KEY_FRAME_LINE_WIDTH = "FrameLineWidth";

    public static final String KEY_FRAME_LINE_COLOR = "FrameLineColor";

    public static final String KEY_AXIS_STRING_COLOR = "StringColor";

    public static final String KEY_AXIS_TITLE_FONT_NAME = "TitleFontName";

    public static final String KEY_AXIS_TITLE_FONT_SIZE = "TitleFontSize";

    public static final String KEY_AXIS_TITLE_FONT_STYLE = "TitleFontStyle";

    public static final String KEY_AXIS_SCALE_FONT_NAME = "ScaleFontName";

    public static final String KEY_AXIS_SCALE_FONT_SIZE = "ScaleFontSize";

    public static final String KEY_AXIS_SCALE_FONT_STYLE = "ScaleFontStyle";

    public static final String TAG_NAME_AXIS = "Axis";

    public static final String KEY_POSITION = "Position";

    public static final String KEY_AXIS_MIN_VALUE = "MinValue";

    public static final String KEY_AXIS_MAX_VALUE = "MaxValue";

    public static final String KEY_AXIS_INVERT_COORDINATES = "InvertCoordinates";

    public static final String KEY_AXIS_SCALE_TYPE = "ScaleType";

    public static final String KEY_STEP_VALUE = "StepValue";

    public static final String KEY_BASELINE_VALUE = "BaselineValue";

    public static final String KEY_AUTO_CALC_NUMBER = "AutomaticCalculationOfTick";

    public static final String KEY_TITLE_TEXT = "Title";

    public static final String KEY_AXIS_VISIBLE = "AxisVisible";

    public static final String KEY_TITLE_VISIBLE = "TitleVisible";

    public static final String KEY_TICK_MARK_VISIBLE = "TickMarkVisible";

    public static final String KEY_NUMBER_VISIBLE = "NumberVisible";

    public static final String KEY_TICK_MARK_INSIDE = "TickMarkInside";

    public static final String KEY_NUMBER_INTEGER = "NumberInteger";

    public static final String KEY_EXPONENT_VISIBLE = "Exponent";

    public static final String KEY_EXPONENT_VALUE = "ExponentValue";
    
    public static final String KEY_NUMBER_ANGLE = "NumberAngle";

}
