package jp.riken.brain.ni.samuraigraph.data;

import java.io.BufferedWriter;
import java.io.File;
import java.io.IOException;
import java.io.Writer;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGBufferedFileWriter;
import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGIConstants;
import jp.riken.brain.ni.samuraigraph.base.SGIDataSource;
import jp.riken.brain.ni.samuraigraph.base.SGITextDataConstants;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

import org.w3c.dom.Element;

/**
 * An abstract data class which has arrays of numbers and strings.
 *
 */
public abstract class SGArrayData extends SGData implements SGITextDataConstants,
        SGIConstants, SGIDataColumnTypeConstants {

    /**
     * The text data file.
     */
    protected SGTextDataFile mDataFile = null;

    /**
     * An observer for array data.
     */
    protected SGDataSourceObserver mObserver = null;

    /**
     * The default constructor.
     */
    public SGArrayData() {
        super();
    }

    /**
     * Build a data with given titles and numbers.
     * 
     * @param dataFile
     *           the text data file
     * @param obs
     *           an observer for array data
     */
    public SGArrayData(final SGTextDataFile dataFile, SGDataSourceObserver obs) {
        
        super();

        // null check
        if (dataFile == null || obs == null) {
            throw new IllegalArgumentException("dataFile == null || obs == null");
        }

        // set to the attribute
        this.mDataFile = dataFile;
        this.mObserver = obs;

        // add to the observer
        obs.addData(this);
    }

    /**
     * Returns the number of columns.
     * 
     * @return
     *        the number of columns or -1 if data columns do not exist
     */
    public int getColNum() {
    	return this.mDataFile.getColNum();
    }
    
    /**
     * Returns the number of rows of each column.
     * 
     * @return
     *        the number of rows of each column or -1 if data columns do not exist
     */
    public int getRowNum() {
    	return this.mDataFile.getRowNum();
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     * 
     */
    public void dispose() {
        super.dispose();
        
        // call a method of netCDF data observer
        this.mObserver.dataDisposed(this);

        // clear attributes
        this.mDataFile = null;
    }
    
    /**
     * Save values to given file.
     * 
     * @param data
     *             a data
     * @param mode
     *             the mode of saving data
     * @return
     *             true if succeeded
     */
    public boolean saveData(final File file, final int mode) {
        boolean status = false;
        SGDataColumn[] colArray = this.getDataSetColumns();
        SGBufferedFileWriter writer = null;
        try {
            writer = new SGBufferedFileWriter(file);
			BufferedWriter bw = writer.getBufferedWriter();
            boolean titleFlag = true;
            for (int ii = 0; ii < colArray.length; ii++) {
                String title = colArray[ii].getTitle();
                if (title == null) {
                    titleFlag = false;
                    break;
                }
            }
            if (titleFlag) {
                for (int ii = 0; ii < colArray.length; ii++) {
                    String title = colArray[ii].getTitle();
                    String str = SGUtilityText.getCSVString(title);
                    if (ii != 0) {
                    	bw.write(',');
                    }
                    bw.write(str);
                }
                bw.write('\n');
            }
            status = this.writeData(bw, mode);
        } catch (IOException ex) {
            return false;
        } finally {
        	if (writer != null) {
        		writer.close();
        	}
        }
        return status;
    }

    /**
     * Returns an array of data columns to save to a data set.
     * 
     * @return an array of data columns to save to a data set
     */
    protected abstract SGDataColumn[] getDataSetColumns();
    
    public SGDataColumn[] getDataSetColumnsClone() {
        SGDataColumn[] cols = this.getDataSetColumns();
        SGDataColumn[] columns = new SGDataColumn[cols.length];
        for (int i = 0; i < cols.length; i++) {
            columns[i] = (SGDataColumn)cols[i].clone();
        }
        return columns;
    }
    
    /**
     * Returns an array of data columns that are used in groupset.
     * 
     * @return an array of data columns that are used in groupset
     */
    public abstract SGDataColumn[] getUsedDataColumnsClone();
    
    /**
     * Returns the number of data columns of a given class.
     * 
     * @param cl
     *           a class object
     * @return the number of data columns of a given class
     */
    public int getColNum(Class cl) {
    	return this.mDataFile.getColNum(cl);
    }
    
    /**
     * Returns the number of data points.
     * 
     * @return the number of data points
     */
    public abstract int getPointsNumber();

    /**
     * Returns an array of data column indices of given class.
     * 
     * @param cl
     *           a class object
     * @return an array of column indices
     */
    public int[] getIndexArray(Class cl) {
    	return this.mDataFile.getIndexArray(cl);
    }
    
    /**
     * Set the data.
     * 
     * @param data
     *             data set to this object
     * @return true if succeeded            
     */
    public boolean setData(final SGData data) {
        if (!(data instanceof SGArrayData)) {
            throw new IllegalArgumentException("!(data instanceof SGArrayData)");
        }
        SGArrayData aData = (SGArrayData) data;
        
    	this.mDataFile = aData.mDataFile;
    	this.mObserver = aData.mObserver;
        this.mObserver.addData(this);

        return true;
    }
    
    /**
     * Returns an array of titles of the columns.
     * 
     * @return an array of titles of the columns
     */
    public String[] getTitles() {
    	return this.mDataFile.getTitles();
    }
    
    /**
     * Returns the title of a column at given index.
     * 
     * @param colIndex
     *                the column index
     * @return the title
     */
    public String getTitle(final int colIndex) {
    	return this.mDataFile.getTitle(colIndex);
    }

    /**
     * Returns an array of value types of the columns.
     * 
     * @return an array of value types of the columns
     */
    public String[] getValueTypes() {
    	return this.mDataFile.getValueTypes();
    }

    /**
     * Returns an array of current column types.
     * @return
     *        an array of current column types
     */
    public abstract String[] getCurrentColumnType();
    
    /**
     * Get properties of this data.
     * @param p
     *          properties to set values
     * @return
     *          true if succeeded
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof ArrayDataProperties) == false) {
            return false;
        }
        return true;
    }
    
    /**
     * Set properties to this data.
     * @param p
     *          properties that have values to set to this data
     * @return
     *          true if succeeded
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof ArrayDataProperties) == false) {
            return false;
        }
        return true;
    }
    
    /**
     * A class for array data properties.
     */
    public static abstract class ArrayDataProperties extends DataProperties {

        /**
         * The default constructor.
         */
        public ArrayDataProperties() {
            super();
        }
    }
    
    /**
     * Read and returns the column index.
     * 
     * @param el
     *            an Element object
     * @param key
     *            the key of index property
     * @return the index if it exists
     */
    protected Integer readIndex(Element el, String key) {
        Integer num = null;
        String str = el.getAttribute(key);
        if (str.length() != 0) {
            num = SGUtilityText.getInteger(str);
        }
        return num;
    }

    /**
     * Read and returns the column indices.
     * 
     * @param el
     *            an Element object
     * @param key
     *            the key of index property
     * @return the indices it exists
     */
    protected Integer[] readIndices(Element el, String key) {
    	return SGUtility.readIndices(el, key);
    }

    /**
     * Write this data to a given output stream.
     * 
     * @param w
     *          an output stream
     * @param mode
     *          the mode of writing data
     * @return
     *          true if succeeded
     */
    protected boolean writeData(Writer w, final int mode) throws IOException {
        SGDataColumn[] colArray = this.getDataSetColumns();
        StringBuffer sb = new StringBuffer();
        Object[][] array = this.getDataSetValues();
        for (int ii = 0; ii < array.length; ii++) {
            final int colNum = array[ii].length;
            
            if (mode == SGIConstants.SAVE_PROPERTIES_TO_DATA_SET_OLDER) {
                // check whether NaN values exist
                boolean isNaN = false;
                for (int jj = 0; jj < colNum; jj++) {
                    if (colArray[jj] instanceof SGNumberDataColumn) {
                        if (array[ii][jj] instanceof Double) {
                            Double d = (Double) array[ii][jj];
                            if (d.isNaN()) {
                                isNaN = true;
                                break;
                            }
                        }
                    }
                }

                // when NaN values exist
                if (isNaN) {
                    sb.append(DATA_COMMENT_PREFIX);
                    sb.append(DATA_COMMENT_HEADER_PREFIX);
                    sb.append(HEADER_NOT_A_COMMENT_LINE);
                    sb.append(DATA_COMMENT_HEADER_SUFFIX);
                }
            }
            
            // export values in a row
            for (int jj = 0; jj < colNum; jj++) {
                sb.append(array[ii][jj]);
                if (jj != colNum - 1) {
                    sb.append(',');
                }
            }
            sb.append('\n');
        }
        w.write(sb.toString());
        return true;
    }
    
    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    protected Object[][] getDataSetValues() {
        final int dataNum = this.getPointsNumber();
        SGDataColumn[] colArray = this.getDataSetColumns();
        final int colNum = colArray.length;
        Object[][] array = new Object[dataNum][];
        for (int ii = 0; ii < dataNum; ii++) {
            array[ii] = new Object[colNum];
            for (int jj = 0; jj < colNum; jj++) {
                SGDataColumn col = colArray[jj];
                array[ii][jj] = col.getExportedValue(ii);
            }
        }
        return array;
    }
    
    /**
     * Checks whether a given column index is within array bounds.
     * 
     * @param index
     *              the column index
     * @return true if the given column index is within array bounds
     */
    protected boolean checkColumnIndexRange(final Integer index) {
    	return this.mDataFile.checkColumnIndexRange(index);
    }
    
    /**
     * Sets the data column.
     * 
     * @param col
     *            the column to be set
     * @param colIndex
     *            the column index
     */
    protected void setDataColumn(final SGDataColumn col, final int colIndex) {
    	this.mDataFile.setDataColumn(col, colIndex);
    }
    
    /**
     * Returns the data type for given type.
     * 
     * @param type
     *           the type of use
     * @return the data type
     */
    public String getDataType(final int type) {
        // by default, returns the data type
        return this.getDataType();
    }

    /**
     * Writes properties of this data object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    public boolean writeProperty(Element el, final int type) {

        switch (type) {
        case SAVE_PROPERTIES_FOR_DUPLICATION:
        case SAVE_PROPERTIES_FOR_COPYING:
        case SAVE_PROPERTIES_FOR_CUTTING:
        case SAVE_PROPERTIES_TO_PROPERTY_FILE:
            // write indices of attributes
            if (this.writeAttributeColumnIndices(el) == false) {
                return false;
            }
            break;
        case SAVE_PROPERTIES_TO_DATA_SET:
        case SAVE_PROPERTIES_TO_DATA_SET_OLDER:
            // write sequential column indices
            if (this.writeSequentialColumnIndices(el) == false) {
                return false;
            }
            break;
        case SAVE_PROPERTIES_TO_DATA_SET_NETCDF:
            if (this.writeSequentialColumnName(el) == false) {
                return false;
            }
            break;
        default:
            // do nothing
            break;
        }
        
        return true;
    }

    /**
     * Writes data column indices in the attributes.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected abstract boolean writeAttributeColumnIndices(Element el);

    /**
     * Writes data column indices as sequential numbers.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected abstract boolean writeSequentialColumnIndices(Element el);
    
    /**
     * Writes data column name for netCDF file as sequential numbers.
     * 
     * @param el
     *            the Element object
     * @return true if succeeded
     */
    protected abstract boolean writeSequentialColumnName(Element el);
    
    /**
     * Return data column name for netCDF file as sequential numbers.
     * 
     * @param index
     * @return sequential data column name for netCDF file
     */
    protected String getSequentialColumnName(final int index) {
        return "column"+index;
    }

	protected SGDataColumn getColumn(final Integer index) {
		if (index != null) {
			return this.mDataFile.mDataColumns[index.intValue()];
		} else {
			return null;
		}
	}

	protected SGDataColumn[] getColumns(final Integer[] indices) {
		if (indices != null) {
			SGDataColumn[] columns = new SGDataColumn[indices.length];
			for (int ii = 0; ii < columns.length; ii++) {
				columns[ii] = this.mDataFile.mDataColumns[indices[ii]];
			}
			return columns;
		} else {
			return null;
		}
	}

    /**
     * Returns a map which has data information.
     * Overrode to set the data type.
     * 
     * @return a map which has data information
     */
    public Map<String, Object> getInfoMap() {
        Map<String, Object> map = super.getInfoMap();
        map.put(KEY_DATA_TYPE, this.getDataType());
        return map;
    }

    /**
     * Copies of this data object.
     * 
     * @return copy of this data object
     */
    public Object copy() {
        SGArrayData aData = (SGArrayData) super.copy();
        this.mObserver.addData(aData);
        return aData;
    }
    
    /**
     * Returns the data file.
     * 
     * @return the data file
     */
    public SGTextDataFile getDataFile() {
    	return this.mDataFile;
    }

    /**
     * Returns the data source.
     * 
     * @return the data source
     */
    @Override
    public SGIDataSource getDataSource() {
    	return this.getDataFile();
    }

}
