package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Calendar;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeSet;

import jp.riken.brain.ni.samuraigraph.base.SGCSVTokenizer.Token;
import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2d;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;

import org.w3c.dom.NamedNodeMap;

import ucar.ma2.DataType;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

/**
 * An utility class for data.
 *
 */
public class SGDataUtility implements SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

	private static final String[] ARRAY_EMPTY = { "" };

	private static final String[] ARRAY_NUMBER_SXYZ = { "", X_VALUE, Y_VALUE,
			Z_VALUE };

	private static final String[] ARRAY_NUMBER_NETCDF_SXYZ = { "", X_VALUE,
			Y_VALUE, Z_VALUE, TIME, SERIAL_NUMBER };

	private static final String[] ARRAY_NUMBER_VXY_POLAR = { "", X_COORDINATE,
			Y_COORDINATE, MAGNITUDE, ANGLE };

	private static final String[] ARRAY_NUMBER_VXY_ORTHOGONAL = { "",
			X_COORDINATE, Y_COORDINATE, X_COMPONENT, Y_COMPONENT };

	private static final String[] ARRAY_NUMBER_NETCDF_VXY_POLAR = { "",
			X_COORDINATE, Y_COORDINATE, MAGNITUDE, ANGLE, TIME, SERIAL_NUMBER };

	private static final String[] ARRAY_NUMBER_NETCDF_VXY_ORTHOGONAL = { "",
			X_COORDINATE, Y_COORDINATE, X_COMPONENT, Y_COMPONENT, TIME, SERIAL_NUMBER };

    private static final String MID_COLUMN_NO = " for No.";
    
    private static final String MID_COLUMN_NETCDF = " for ";
    
    /**
     * The default number of multiple dimension.
     */
    private static final int DEFAULT_MULTIPLE_DIMENSION_NUM = 4;
    
    /**
     * The key name of attribute on saving text data to netCDF dataset. 
     */
    public static final String ATTRIBUTE_KEY_VALUE_TYPE = "value_type";
    
    /**
     * Appends the column number to a text string of the column type.
     * 
     * @param colType
     *              the column type
     * @param index
     *              array index of the column
     */
    public static final String appendColumnNo(final String colType, final int index) {
    	StringBuffer sb = new StringBuffer();
    	sb.append(colType);
    	sb.append(MID_COLUMN_NO);
    	sb.append(index + 1);
    	return sb.toString();
    }
    
    /**
     * Appends the variable name to a text string of the column type on the netCDF data.
     * 
     * @param colType
     * @param variableName
     * @return
     */
    public static final String appendColumnTitle(final String colType, final String variableName) {
        StringBuffer sb = new StringBuffer();
        sb.append(colType);
        sb.append(MID_COLUMN_NETCDF);
        sb.append(variableName);
        return sb.toString();
    }
    
    public static final String appendColumnNoOrTitle(
            final String colType, final int index, final boolean isRepeated, final String title) {
        if (isRepeated) {
            return appendColumnNo(colType, index);
        } else {
            return appendColumnTitle(colType, title);
        }
    }
    
    /**
     * Appends the column title to a text string of the column type on the netCDF data.
     * <p>
     * When using the netCDF data, column title exists and is not doubled.
     * @param colType
     * @param index
     * @param colInfo
     * @return
     */
    public static final String appendNetCDFColumnTitle(
            final String colType, final int index, final SGDataColumnInfo[] colInfo) {
        StringBuffer sb = new StringBuffer();
        sb.append(colType);
        sb.append(MID_COLUMN_NETCDF);
        if (colInfo[index] instanceof SGNetCDFDataColumnInfo) {
            sb.append(((SGNetCDFDataColumnInfo)colInfo[index]).getVariableName());
        } else {
            sb.append(colInfo[index].getTitle());
        }
        return sb.toString();
    }
    
    public static final String appendColumnType(
            final String colType, final int index, final boolean isNetCDF,
            final SGDataColumnInfo[] colInfo, final boolean isRepeated) {
        if (isNetCDF) {
            return appendNetCDFColumnTitle(colType, index, colInfo);
        } else {
            return appendColumnNoOrTitle(colType, index, isRepeated, colInfo[index].getTitle());
        }
    }
    
    /**
     * Extracts the column number from a given text string.
     * 
     * @param str
     *            a text string
     * @return the column number if it exists and otherwise null
     */
    public static final Integer getAppendedColumnIndex(final String str) {
        final String forSeparator = "for";
        final String noSeparator = "No";
        int idx = str.toUpperCase().indexOf(forSeparator.toUpperCase());
        if (idx == -1) {
            return null;
        }
        idx = idx + forSeparator.length();
        String substr = str.substring(idx).trim().toUpperCase();
        
        if (SGUtilityText.startsWith(substr, noSeparator.toUpperCase())==false) {
            return null;
        }
        substr = SGUtilityText.getCharString(substr);
        final String numStr = substr.substring(noSeparator.length());
        Integer index = SGUtilityText.getInteger(numStr);
        if (index==null) {
            return null;
        }
        return Integer.valueOf(index.intValue() - 1);
    }
    
    public static final Integer getAppendedColumnIndex(
            final String str, final SGDataColumnInfo[] colInfo) {
        Integer value = getColumnIndexOfNetCDFAppendedColumnTitle(str, colInfo);
        if (null==value) {
            value = getAppendedColumnIndex(str);
        }
        return value;
    }
    
    public static final Integer getAppendedColumnIndex(
            final String str, final String[] columnTitles) {
        int idx = str.toUpperCase().indexOf(MID_COLUMN_NETCDF.trim().toUpperCase());
        if (idx == -1) {
            return null;
        }
        idx = idx + MID_COLUMN_NETCDF.trim().length();
        final String titleStr = str.substring(idx).trim();
        for (int i = 0; i < columnTitles.length; i++) {
            if (columnTitles[i]!=null) {
                if (titleStr.equals(columnTitles[i].trim())) {
                    return Integer.valueOf(i);
                }
            }
        }
        return getAppendedColumnIndex(str);
    }
    
    /**
     * Extracts the column number from a given text string.
     * 
     * @param str
     * @param colInfo
     * @return the column number if it exists and otherwise null
     */
    public static final Integer getColumnIndexOfNetCDFAppendedColumnTitle(
            final String str, final SGDataColumnInfo[] colInfo) {
        int idx = str.toUpperCase().indexOf(MID_COLUMN_NETCDF.trim().toUpperCase());
        if (idx == -1) {
            return null;
        }
        idx = idx + MID_COLUMN_NETCDF.trim().length();
        final String titleStr = str.substring(idx).trim();
        for (int i = 0; i < colInfo.length; i++) {
            if (colInfo[i] instanceof SGNetCDFDataColumnInfo) {
                if (titleStr.equals(((SGNetCDFDataColumnInfo)colInfo[i]).getVariableName())) {
                    return Integer.valueOf(i);
                }
            } else {
                if (titleStr.equals(colInfo[i].getTitle())) {
                    return Integer.valueOf(i);
                }
            }
        }
        return null;
    }
    
    public static final Integer getColumnIndexOfNetCDFAppendedColumnTitle(
            final String str, SGNetCDFFile file) {
        int idx = str.toUpperCase().indexOf(MID_COLUMN_NETCDF.toUpperCase().trim());
        if (idx == -1) {
            return null;
        }
        idx = idx + MID_COLUMN_NETCDF.trim().length();
        final String titleStr = str.substring(idx).trim();
        int index = file.getVariableIndex(titleStr);
        if (index == -1) {
            return null;
        }
        return Integer.valueOf(index);
    }
    
    public static final Integer getColumnIndexOfAppendedColumnType(
            final String str, final boolean isNetCDF, final SGDataColumnInfo[] colInfo) {
        if (isNetCDF) {
            return getColumnIndexOfNetCDFAppendedColumnTitle(str, colInfo);
        } else {
            return getAppendedColumnIndex(str, colInfo);
        }
    }
    
    /**
     * Create items for the combo box.
     * <p>
     * Return the enabled candidates for column type of Scalar-XY data type.
     * 
     * @param dataType
     *            the type of data
     * @param infoMap
     *            the map of data information
     * @param valueType
     *            the type of value
     * @return an array of items
     */
    private static String[] getSXYColumnTypeCandidates(
            final String dataType,
            final Map<String, Object> infoMap, final String valueType) {
        
        final boolean isNetCDF = isNetCDFData(dataType);
        
        // for netCDF data
        Boolean multiple = null;
        Boolean multipleVariable = null;
        if (isNetCDF) {
            multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
            if (multiple == null) {
                return null;
            }
            multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
            if (multipleVariable == null) {
                return null;
            }
        }

        List<String> itemList = new ArrayList<String>();
        itemList.add("");
        if (!VALUE_TYPE_TEXT.equals(valueType)) {
            // other than text column
            itemList.add(X_VALUE);
            itemList.add(Y_VALUE);
        }

        // current row index
        Integer rowIndex = (Integer) infoMap.get(KEY_CURRENT_ROW_INDEX);

        // all columns
        SGDataColumnInfo[] colInfo = (SGDataColumnInfo[]) infoMap.get(KEY_COLUMN_INFO);
        boolean[] isRepeatedTitle = isEmptyOrRepeatedColumnTitle(colInfo);

        // get column indices of x and y values
        List<Integer> xIndexList = new ArrayList<Integer>();
        List<Integer> yIndexList = new ArrayList<Integer>();
        for (int ii = 0; ii < colInfo.length; ii++) {
            String colType = colInfo[ii].getColumnType();
            final Integer index = Integer.valueOf(ii);
            if (X_VALUE.equals(colType)) {
                xIndexList.add(index);
            } else if (Y_VALUE.equals(colType)) {
                yIndexList.add(index);
            }
        }

        // counts the number of number type rows
        int numberRowNum = 0;
        for (int ii = 0; ii < colInfo.length; ii++) {
            String vType = colInfo[ii].getValueType();
            if (VALUE_TYPE_NUMBER.equals(vType)) {
                numberRowNum++;
            } else if (VALUE_TYPE_DATE.equals(vType)) {
                if (X_VALUE.equals(colInfo[ii].getColumnType()) || Y_VALUE.equals(colInfo[ii].getColumnType())) {
                    numberRowNum++;
                }
            }
        }

        // add time, pickup and serial number columns
        if (isNetCDF && VALUE_TYPE_NUMBER.equals(valueType)) {
            itemList.add(TIME);
            itemList.add(PICKUP);
            itemList.add(SERIAL_NUMBER);
        }

        if (isNetCDF && !multipleVariable.booleanValue()) {
            // netCDF multiple for dimension
            itemList.add(LOWER_ERROR_VALUE);
            itemList.add(UPPER_ERROR_VALUE);
            itemList.add(LOWER_UPPER_ERROR_VALUE);
            itemList.add(TICK_LABEL);
        } else {
            // text data or netCDF variable data.

            // get the list of column index those are assignable
            // for the error bar or tick labels
            List<Integer> indexList = getOptionalColumnsAssignableIndexList(
                    dataType, colInfo, xIndexList, yIndexList);
            if (indexList != null) {
                // remove the current row if it exists
                indexList.remove(rowIndex);

                if (VALUE_TYPE_NUMBER.equals(valueType)) {
                    // only for the number type column

                    // add items of error bars to the list
                    for (int ii = 0; ii < indexList.size(); ii++) {
                        Integer index = indexList.get(ii);
                        String vType = colInfo[index.intValue()].getValueType();
                        if (VALUE_TYPE_SAMPLING_RATE.equals(vType)) {
                            continue;
                        }
                        final int num = index.intValue();
                        if (numberRowNum >= 4) {
                            itemList.add(appendColumnType(LOWER_ERROR_VALUE, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                            itemList.add(appendColumnType(UPPER_ERROR_VALUE, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                            itemList.add(appendColumnType(LOWER_UPPER_ERROR_VALUE, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                        } else if (numberRowNum >= 3) {
                            itemList.add(appendColumnType(LOWER_UPPER_ERROR_VALUE, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                        }
                    }
                }

                if (!VALUE_TYPE_SAMPLING_RATE.equals(colInfo[rowIndex].getValueType())) {
                    // add items of tick labels to the list
                    List<Integer> dateIndexList = new ArrayList<Integer>();
                    List<Integer> notDateIndexList = new ArrayList<Integer>();
                    for (int ii = 0; ii < indexList.size(); ii++) {
                        Integer index = indexList.get(ii);
                        String vType = colInfo[index.intValue()].getValueType();
                        if (VALUE_TYPE_DATE.equals(vType)) {
                            dateIndexList.add(index);
                        } else {
                            notDateIndexList.add(index);
                        }
                    }
                    if (dateIndexList.size() != 0) {
                        // tick label is used for "not date type" columns
                        for (int ii = 0; ii < notDateIndexList.size(); ii++) {
                            Integer index = notDateIndexList.get(ii);
                            final int num = index.intValue();
                            itemList.add(appendColumnType(TICK_LABEL, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                        }
                    } else {
                        for (int ii = 0; ii < indexList.size(); ii++) {
                            Integer index = indexList.get(ii);
                            final int num = index.intValue();
                            itemList.add(appendColumnType(TICK_LABEL, num, isNetCDF, colInfo, isRepeatedTitle[num]));
                        }
                    }
                }
            }
        }

        return itemList.toArray(new String[itemList.size()]);
    }
    
    /**
     * Create items for the combo box.
     * <p>
     * Return the enabled candidates for column type.
     * 
     * @param dataType
     *            the type of data
     * @param infoMap
     *            the map of data information
     * @param valueType
     *            the type of value
     * @return an array of items
     */
    public static String[] getColumnTypeCandidates(final String dataType,
            final Map<String, Object> infoMap, final String valueType) {

        String[] items = null;
        if (isSXYTypeData(dataType)) {
            // SXY
        	if (VALUE_TYPE_BYTE_DATA.equals(valueType)) {
        		items = ARRAY_EMPTY;
        	} else {
                items = getSXYColumnTypeCandidates(dataType, infoMap, valueType);
        	}
        } else if (SGDataTypeConstants.SXYZ_DATA.equals(dataType)) {
            // SXYZ
            if (VALUE_TYPE_NUMBER.equals(valueType)) {
                items = ARRAY_NUMBER_SXYZ;
            } else if (VALUE_TYPE_TEXT.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_DATE.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_BYTE_DATA.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else {
                throw new IllegalArgumentException("Invalid value type: " + valueType);
            }
        } else if (SGDataTypeConstants.SXYZ_NETCDF_DATA.equals(dataType)) {
            // SXYZ netCDF
            if (VALUE_TYPE_NUMBER.equals(valueType)) {
                items = ARRAY_NUMBER_NETCDF_SXYZ;
            } else if (VALUE_TYPE_TEXT.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_DATE.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_BYTE_DATA.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else {
                throw new IllegalArgumentException("Invalid value type: " + valueType);
            }
        } else if (SGDataTypeConstants.VXY_DATA.equals(dataType)) {
            // VXY
            if (VALUE_TYPE_NUMBER.equals(valueType)) {
                if (isPolar(infoMap)) {
                    items = ARRAY_NUMBER_VXY_POLAR;
                } else {
                    items = ARRAY_NUMBER_VXY_ORTHOGONAL;
                }
            } else if (VALUE_TYPE_TEXT.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_DATE.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_BYTE_DATA.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else {
                throw new IllegalArgumentException("Invalid value type: " + valueType);
            }
        } else if (SGDataTypeConstants.VXY_NETCDF_DATA.equals(dataType)) {
            // VXY for netCDF
            if (VALUE_TYPE_NUMBER.equals(valueType)) {
                if (isPolar(infoMap)) {
                    items = ARRAY_NUMBER_NETCDF_VXY_POLAR;
                } else {
                    items = ARRAY_NUMBER_NETCDF_VXY_ORTHOGONAL;
                }
            } else if (VALUE_TYPE_TEXT.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_DATE.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else if (VALUE_TYPE_BYTE_DATA.equals(valueType)) {
                items = ARRAY_EMPTY;
            } else {
                throw new IllegalArgumentException("Invalid value type: " + valueType);
            }
        } else {
            throw new IllegalArgumentException("Invalid data type: " + dataType);
        }
        
        return (String[]) items.clone();
    }

    /**
     * Checks whether given data columns are valid for given data type.
     * This method checks only the number of selected columns.
     * 
     * @param dataType
     *            data type
     * @param columnInfoArray
     *            an array of columns information
     * @param infoMap
     *            information map for data
     */
    public static boolean checkDataColumns(final String dataType,
            final SGDataColumnInfo[] columnInfoArray,
            final Map<String, Object> infoMap) {

        final int len = columnInfoArray.length;
        String[] columnTypes = new String[len];
        for (int ii = 0; ii < len; ii++) {
        	columnTypes[ii] = columnInfoArray[ii].getColumnType();
        }

        // other than multiple and sampling data, check the duplication
        if (!isSXYTypeData(dataType)
//                && !dataType.equals(SGDataTypeConstants.SXY_SAMPLING_DATA)
                ) {
            for (int ii = 0; ii < len - 1; ii++) {
                if (columnTypes[ii] == null || "".equals(columnTypes[ii])) {
                    continue;
                }
                for (int jj = ii + 1; jj < len; jj++) {
                    if (columnTypes[ii].equals(columnTypes[jj])) {
                        return false;
                    }
                }
            }
        }

        if (isSXYTypeData(dataType)) {

        	// one array for x values and multiple arrays for y values
        	// or one array for y values and multiple arrays for x values
        	// are permitted
            final int xCnt = count(columnTypes, X_VALUE);
            final int yCnt = count(columnTypes, Y_VALUE);

			final boolean byMultiple = (xCnt == 1 && yCnt > 1);
			final boolean bxMultiple = (xCnt > 1 && yCnt == 1);
			final boolean bothSingle = (xCnt == 1 && yCnt == 1);

			if (!byMultiple && !bxMultiple && !bothSingle) {
				return false;
			}
			String multipleColumnType = null;
			String singleColumnType = null;
			if (byMultiple) {
				multipleColumnType = Y_VALUE;
				singleColumnType = X_VALUE;
			} else if (bxMultiple) {
				multipleColumnType = X_VALUE;
				singleColumnType = Y_VALUE;
			}

            if (!bothSingle) {
    			Set<String> multipleValueTypeSet = new HashSet<String>();
    			if (multipleColumnType != null) {
    				for (int ii = 0; ii < len; ii++) {
    					if (multipleColumnType.equals(columnTypes[ii])) {
    						multipleValueTypeSet.add(columnInfoArray[ii].getValueType());
    					}
    				}
    			}
    			
                // checks whether sampling rate is used for multiple values
    			if (multipleValueTypeSet.contains(VALUE_TYPE_SAMPLING_RATE)) {
    				return false;
    			}
    			
    			// checks whether the data column exists in multiple values
    			if (multipleValueTypeSet.contains(VALUE_TYPE_DATE)) {
    				return false;
    			}
            }
            
            boolean checkAppendedIndex = false;
            if (isArrayData(dataType)){
            	checkAppendedIndex = true;
            } else if (isNetCDFData(dataType)) {
                Boolean variable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
                if (variable == null) {
                    return false;
                }
                checkAppendedIndex = variable.booleanValue();
            } else {
            	return false;
            }

            if (checkAppendedIndex) {
            	// checks column indices appended to the error bars and tick labels
            	
                boolean isNetCDF = SGDataUtility.isNetCDFData(dataType);
                
                List<Integer> lList = new ArrayList<Integer>();
                List<Integer> uList = new ArrayList<Integer>();
                List<Integer> luList = new ArrayList<Integer>();
                for (int ii = 0; ii < columnTypes.length; ii++) {
                    String colType = columnTypes[ii];
                    if (colType.startsWith(LOWER_ERROR_VALUE)) {
                        Integer num = getColumnIndexOfAppendedColumnType(colType, isNetCDF, columnInfoArray);
                        if (num == null) {
                            return false;
                        }
                        lList.add(num);
                    } else if (colType.startsWith(UPPER_ERROR_VALUE)) {
                        Integer num = getColumnIndexOfAppendedColumnType(colType, isNetCDF, columnInfoArray);
                        if (num == null) {
                            return false;
                        }
                        uList.add(num);
                    } else if (colType.startsWith(LOWER_UPPER_ERROR_VALUE)) {
                        Integer num = getColumnIndexOfAppendedColumnType(colType, isNetCDF, columnInfoArray);
                        if (num == null) {
                            return false;
                        }
                        luList.add(num);
                    }
                }

                List<Integer> tList = new ArrayList<Integer>();
                for (int ii = 0; ii < columnTypes.length; ii++) {
                    String colType = columnTypes[ii];
                    if (colType.startsWith(TICK_LABEL)) {
                        Integer num = getColumnIndexOfAppendedColumnType(colType, isNetCDF, columnInfoArray);
                        if (num == null) {
                            return false;
                        }
                        tList.add(num);
                    }
                }

                if (!bothSingle) {
                    // checks whether error bars and tick labels are assigned to
                    // not multiple values
                    Integer singleIndex = null;
                    for (int ii = 0; ii < columnTypes.length; ii++) {
                        if (columnTypes[ii].equals(singleColumnType)) {
                            singleIndex = Integer.valueOf(ii);
                            break;
                    	}
                    }
                    if (singleIndex == null) {
                    	return false;
                    }
                    if (lList.contains(singleIndex)) {
                    	return false;
                    }
                    if (uList.contains(singleIndex)) {
                    	return false;
                    }
                    if (luList.contains(singleIndex)) {
                    	return false;
                    }
                    if (tList.contains(singleIndex)) {
                    	return false;
                    }
                }
                
                // checks whether error bars and tick labels for x values 
    			// and those for y values exist at the same time
    			Set<Integer> allSet = new TreeSet<Integer>();
    			allSet.addAll(lList);
    			allSet.addAll(uList);
    			allSet.addAll(luList);
    			allSet.addAll(tList);
    			List<Integer> xList = new ArrayList<Integer>();
    			List<Integer> yList = new ArrayList<Integer>();
    			Iterator<Integer> itr = allSet.iterator();
    			while (itr.hasNext()) {
    				Integer index = itr.next();
    				String colType = columnTypes[index.intValue()];
            		if (X_VALUE.equals(colType)) {
            			xList.add(index);
            		} else if (Y_VALUE.equals(colType)) {
            			yList.add(index);
            		}
    			}
    			if (xList.size() != 0 && yList.size() != 0) {
    				return false;
    			}

                // check duplication
                if (isDuplicated(lList)) {
                	return false;
                }
                if (isDuplicated(uList)) {
                	return false;
                }
                if (isDuplicated(luList)) {
                	return false;
                }
                if (isDuplicated(tList)) {
                	return false;
                }
                
                // check consistency
                if (!lList.equals(uList)) {
                	return false;
                }
                for (int ii = 0; ii < luList.size(); ii++) {
                	Object obj = luList.get(ii);
                	if (lList.contains(obj)) {
                		return false;
                	}
                	if (uList.contains(obj)) {
                		return false;
                	}
                }
            }
            
        } else if (dataType.equals(SGDataTypeConstants.SXYZ_DATA)
                || dataType.equals(SGDataTypeConstants.SXYZ_NETCDF_DATA)) {
            // duplication was already checked

            // only one x values must be selected
            final int xCnt = count(columnTypes, X_VALUE);
            if (xCnt != 1) {
                return false;
            }

            // only one y values must be selected
            final int yCnt = count(columnTypes, Y_VALUE);
            if (yCnt != 1) {
                return false;
            }
            
            // only one z values must be selected
            final int zCnt = count(columnTypes, Z_VALUE);
            if (zCnt != 1) {
                return false;
            }
            
//        } else if (dataType.equals(SGDataTypeConstants.SXY_SAMPLING_DATA)) {
//
////            // at least one y values are selected
////            final boolean y = exists(columnTypes, Y_VALUE);
////            if (!y) {
////                return false;
////            }
//
//            final int xCnt = count(columnTypes, X_VALUE);
//            final int yCnt = count(columnTypes, Y_VALUE);
//            if (xCnt > 0) {
//            	if (yCnt > 0) {
//            		return false;
//            	}
//            } else if (yCnt > 0) {
//            	if (xCnt > 0) {
//            		return false;
//            	}
//            } else if (xCnt == 0 || yCnt == 0) {
//            	return false;
//            }
//
//        } else if (dataType.equals(SGDataTypeConstants.SXY_DATE_DATA)) {
//            // duplication was already checked
//
//            // only one y values must be selected
//            final int yCnt = count(columnTypes, Y_VALUE);
//            if (yCnt != 1) {
//                return false;
//            }
//
//            // only one date values must be selected
//            final int dCnt = count(columnTypes, DATE);
//            if (dCnt != 1) {
//                return false;
//            }
//
//            // if one of lower or upper errors is selected, the other must be
//            // selected
//            final int lCnt = count(columnTypes, LOWER_ERROR_VALUE);
//            final int uCnt = count(columnTypes, UPPER_ERROR_VALUE);
//            if (!(lCnt == 0 && uCnt == 0) && !(lCnt == 1 && uCnt == 1)) {
//                return false;
//            }
//
//            // if lower and upper errors are selected, lower-upper must not 
//            // be selected
//            final int luCnt = count(columnTypes, LOWER_UPPER_ERROR_VALUE);
//            if (lCnt == 1 && uCnt == 1) {
//                if (luCnt != 0) {
//                    return false;
//                }
//            }
            
        } else if (dataType.equals(SGDataTypeConstants.VXY_DATA)
                || dataType.equals(SGDataTypeConstants.VXY_NETCDF_DATA)) {
            // duplication was already checked

            // only one x location values must be selected
            final int xCnt = count(columnTypes, X_COORDINATE);
            if (xCnt != 1) {
                return false;
            }

            // only one y location values must be selected
            final int yCnt = count(columnTypes, Y_COORDINATE);
            if (yCnt != 1) {
                return false;
            }

            // both of magnitude and angle must be selected in polar mode,
            // and both of x component and y component must be selected in the
            // other mode
            final int bCnt1, bCnt2;
            if (isPolar(infoMap)) {
                bCnt1 = count(columnTypes, MAGNITUDE);
                bCnt2 = count(columnTypes, ANGLE);
            } else {
                bCnt1 = count(columnTypes, X_COMPONENT);
                bCnt2 = count(columnTypes, Y_COMPONENT);
            }
            if (bCnt1 != 1) {
                return false;
            }
            if (bCnt2 != 1) {
                return false;
            }
        }
        
        if (isNetCDFData(dataType)) {
            // animation
            final int tCnt = count(columnTypes, TIME);
            if (tCnt > 1) {
                return false;
            }
        }

        return true;
    }

    /**
     * Check whether a given value exists in a given array.
     * 
     * @param items
     *            an array
     * @param value
     *            a value
     * @return true if it exists
     */
    private static boolean exists(Object[] items, Object value) {
        for (int ii = 0; ii < items.length; ii++) {
            if (value.equals(items[ii])) {
                return true;
            }
        }
        return false;
    }

    /**
     * Count the number of equal objects in a given array.
     * 
     * @param items
     *            an array
     * @param value
     *            a value
     * @return the number of items in a given array
     */
    private static int count(Object[] items, Object value) {
        int cnt = 0;
        for (int ii = 0; ii < items.length; ii++) {
            if (value.equals(items[ii])) {
                cnt++;
            }
        }
        return cnt;
    }
    
    private static boolean isDuplicated(List items) {
    	for (int ii = 0; ii < items.size() - 1; ii++) {
    		for (int jj = ii + 1; jj < items.size(); jj++) {
    			if (SGUtility.equals(items.get(ii), items.get(jj))) {
    				return true;
    			}
    		}
    	}
    	return false;
    }

    private static int indexOf(Object[] items, Object value) {
        for (int ii = 0; ii < items.length; ii++) {
            if (value.equals(items[ii])) {
            	return ii;
            }
        }
    	return -1;
    }

    private static int lastIndexOf(Object[] items, Object value) {
        for (int ii = items.length - 1; ii >= 0; ii--) {
            if (value.equals(items[ii])) {
            	return ii;
            }
        }
    	return -1;
    }

    /**
     * Returns whether the polar mode is selected.
     * 
     * @param infoMap
     *            information map for data
     * @return true if polar mode is selected
     * @throws Error throw if infoMap not have KEY_POLAR_SELECTED key
     */
    public static boolean isPolar(final Map<String, Object> infoMap) {
        Object value = infoMap.get(KEY_POLAR_SELECTED);
        if (value == null) {
            throw new Error("Mode for vector data is not selected.");
        }
        Boolean b = (Boolean) value;
        return b.booleanValue();
    }
    
    static boolean[] isEmptyOrRepeatedColumnTitle(SGDataColumnInfo[] columnInfo) {
        int len = columnInfo.length;
        boolean[] result = new boolean[len];
        Arrays.fill(result, false);
        
        for (int i = 0; i < len; i++) {
            SGDataColumnInfo cInfo = columnInfo[i];
            String title = cInfo.getTitle();
            if (title==null || title.trim().equals("")) {
                result[i] = true;
            }
            if (result[i]==false) {
                for (int j = i+1; j < len; j++) {
                    if (title.equalsIgnoreCase(columnInfo[j].getTitle())) {
                        result[i] = true;
                        result[j] = true;
                    }
                }
            }
        }
        
        return result;
    }

    /**
     * Determine the data-type from the first line.
     * 
     * @param tokenList
     *            the list of tokens of the first line
     * @param indexList
     *            the list of indices whether each token is of the number type or the text type
     * @param cList
     *            the list of candidates of data types
     * @return true if succeeded
     */
    public static boolean getDataTypeCandidateList(final List<Token> tokenList, 
            final List<Integer> indexList, final List<String> cList) {

    	// count the number of number type columns
        int cntNumber = 0;
        for (int ii = 0; ii < indexList.size(); ii++) {
            Integer obj = (Integer) indexList.get(ii);
            int num = obj.intValue();
            if (num == 1) {
                cntNumber++;
            }
        }
        
        String[] dataTypes = { SGDataTypeConstants.SXY_DATA,
				SGDataTypeConstants.SXY_MULTIPLE_DATA,
				SGDataTypeConstants.SXY_SAMPLING_DATA,
				SGDataTypeConstants.VXY_DATA, SGDataTypeConstants.SXYZ_DATA };
        for (int ii = 0; ii < dataTypes.length; ii++) {
        	final int num = getMinimumNumberColumns(dataTypes[ii]);
        	if (cntNumber >= num) {
        		cList.add(dataTypes[ii]);
        	}
        }
        if (cntNumber >= getMinimumNumberColumns(SGDataTypeConstants.SXY_DATE_DATA)) {
            for (int ii = 0; ii < tokenList.size(); ii++) {
                Token token = (Token) tokenList.get(ii);
                if (SGUtilityText.getDate(token.getString()) != null) {
                	cList.add(SGDataTypeConstants.SXY_DATE_DATA);
                    break;
                }
            }
        }
        
        return true;
    }
    
    /**
     * Returns the minimum number of number columns for given data type.
     * 
     * @param dataType
     *           the data type
     * @return the minimum number of number type columns
     */
    public static int getMinimumNumberColumns(final String dataType) {
    	int num = -1;
    	if (SGDataTypeConstants.SXY_DATA.equals(dataType)) {
    		num = 1;
    	} else if (SGDataTypeConstants.SXY_MULTIPLE_DATA.equals(dataType)) {
    		num = 1;
    	} else if (SGDataTypeConstants.SXY_SAMPLING_DATA.equals(dataType)) {
    		num = 1;
    	} else if (SGDataTypeConstants.VXY_DATA.equals(dataType)) {
    		num = 4;
    	} else if (SGDataTypeConstants.SXYZ_DATA.equals(dataType)) {
    		num = 3;
    	}
    	return num;
    }
    
    // returns a list of number convertable column index
    public static List<Integer> getColumnIndexListOfNumber(final List<Token> tokenlist) {
        List<Integer> list = new ArrayList<Integer>();
        for (int ii = 0; ii < tokenlist.size(); ii++) {
            Token token = (Token) tokenlist.get(ii);
            final String str = token.getString();
            if (token.isDoubleQuoted()) {
                list.add(Integer.valueOf(0));
            } else {
                Double d = SGUtilityText.getDouble(str);
                if (d!= null) {
                    list.add(Integer.valueOf(1));
                } else {
                    list.add(Integer.valueOf(0));
                }
            }
        }
        return list;
    }

    private static List<String> getColumnNameList(final SGNetCDFDataColumnInfo[] cols, 
            final String colType) {
        List<String> list = new ArrayList<String>();
        for (int ii = 0; ii < cols.length; ii++) {
        	String type = cols[ii].getColumnType();
            if (type.equals(colType)) {
                list.add(cols[ii].getVariableName());
            }
        }
        return list;
    }

    private static boolean getColumnNameAndAppendedNumberList(
            final SGNetCDFDataColumnInfo[] cols, 
            final String colType, List<String> nameList, List<Integer> indexList) {
        for (int ii = 0; ii < cols.length; ii++) {
            String type = cols[ii].getColumnType();
            if (type.startsWith(colType)) {
            	String name = cols[ii].getVariableName();
            	Integer index = getColumnIndexOfNetCDFAppendedColumnTitle(type, cols);
            	if (index == null) {
            		return false;
            	}
            	nameList.add(name);
            	indexList.add(index);
            }
        }
        return true;
    }

    /**
     * Checks the data columns for netCDF data.
     * 
     * @param cols
     *           the data columns
     * @param dataType
     *           the data type
     * @param ncfile
     *           the netCDF data
     * @param infoMap
     *           the information map
     * @return true if columns are valid
     */
    public static boolean checkNetCDFDataColumns(final SGNetCDFDataColumnInfo[] cols, 
            final String dataType, final SGNetCDFFile ncfile, 
            final Map<String, Object> infoMap) {
        
        // animation
        final List<String> timeNameList = getColumnNameList(cols, TIME);
        if (timeNameList.size() > 1) {
            return false;
        }
        SGVariable timeVar = null;
        Dimension timeDim = null;
        if (timeNameList.size() == 1) {
            String aName = timeNameList.get(0);
            timeVar = ncfile.findVariable(aName);
            if (timeVar == null) {
                return false;
            }
            if (!timeVar.isCoordinateVariable()) {
                return false;
            }
            timeDim = timeVar.getDimension(0);
        }
        
        // serial numbers
        final List<String> serialNumberNameList = getColumnNameList(cols, SERIAL_NUMBER);
        if (serialNumberNameList.size() > 1) {
            return false;
        }
        SGVariable serialNumberVar = null;
        Dimension serialNumberDim = null;
        if (serialNumberNameList.size() == 1) {
            String sName = serialNumberNameList.get(0);
            serialNumberVar = ncfile.findVariable(sName);
            if (serialNumberVar == null) {
                return false;
            }
            if (!serialNumberVar.isCoordinateVariable()) {
                return false;
            }
            serialNumberDim = serialNumberVar.getDimension(0);
        }
        
        if (isSXYTypeData(dataType)) {
            
            Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
            if (multiple == null) {
            	return false;
            }
            Boolean multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
        	if (multipleVariable == null) {
        		return false;
        	}
            
            // x and y values
            final List<String> xNameList = getColumnNameList(cols, X_VALUE);
            final List<String> yNameList = getColumnNameList(cols, Y_VALUE);
            if (xNameList.size() == 0 || yNameList.size() == 0) {
                return false;
            }
            if (multipleVariable.booleanValue()) {
                if (xNameList.size() != 1 && yNameList.size() != 1) {
                    return false;
                }
            } else {
                if (xNameList.size() != 1 || yNameList.size() != 1) {
                    return false;
                }
            }
            final String xName0 = xNameList.get(0);
            final String yName0 = yNameList.get(0);
            final SGVariable xVar0 = ncfile.findVariable(xName0);
            final SGVariable yVar0 = ncfile.findVariable(yName0);
            final SGVariable cVar;
            final List<String> oNameList;
        	if (serialNumberVar != null) {
                cVar = serialNumberVar;
                oNameList = new ArrayList<String>();
                oNameList.addAll(xNameList);
                oNameList.addAll(yNameList);
        	} else {
                if (xNameList.size() == 1 && yNameList.size() == 1) {
                    if (xVar0.isCoordinateVariable()) {
                    	cVar = xVar0;
                    	oNameList = yNameList;
                    } else if (yVar0.isCoordinateVariable()) {
                    	cVar = yVar0;
                    	oNameList = xNameList;
                    } else {
                    	return false;
                    }
                } else if (xNameList.size() == 1) {
                    if (!xVar0.isCoordinateVariable()) {
                    	return false;
                    }
                    cVar = xVar0;
                	oNameList = yNameList;
                } else {
                    if (!yVar0.isCoordinateVariable()) {
                    	return false;
                    }
                	cVar = yVar0;
                	oNameList = xNameList;
                }
        	}

            List<SGVariable> oVarList = new ArrayList<SGVariable>();
            for (int ii = 0; ii < oNameList.size(); ii++) {
            	String oName = oNameList.get(ii);
                SGVariable oVar = ncfile.findVariable(oName);
                if (oVar == null) {
                    return false;
                }
                if (oVar.isCoordinateVariable()) {
                	return false;
                }
                oVarList.add(oVar);
            }
            
            // variable must have a dimension equal to the coordinate variable
            Dimension cDim = cVar.getDimension(0);
            for (int ii = 0; ii < oVarList.size(); ii++) {
            	SGVariable oVar = oVarList.get(ii);
               	List<Dimension> oDims = oVar.getDimensions();
                if (!oDims.contains(cDim)) {
                    return false;
                }
            }
            
            // lower and upper error values
            if (!multipleVariable) {
            	// multiple dimension
            	
                final List<String> lNameList = getColumnNameList(cols, LOWER_ERROR_VALUE);
                final List<String> uNameList = getColumnNameList(cols, UPPER_ERROR_VALUE);
                if (!(lNameList.size() == 0 && uNameList.size() == 0) 
                        && !(lNameList.size() == 1 && uNameList.size() == 1)) {
                    return false;
                }
                
                // if lower and upper errors are selected, lower-upper must not 
                // be selected
                final List<String> luNameList = getColumnNameList(cols, LOWER_UPPER_ERROR_VALUE);
                if (lNameList.size() == 1 && uNameList.size() == 1) {
                    if (luNameList.size() != 0) {
                        return false;
                    }
                }
                
                if (lNameList.size() != 0) {
                    String lName = lNameList.get(0);
                    String uName = uNameList.get(0);
                    SGVariable lVar = ncfile.findVariable(lName);
                    if (lVar == null) {
                        return false;
                    }
                    SGVariable uVar = ncfile.findVariable(uName);
                    if (uVar == null) {
                        return false;
                    }
                    
                    // check validity of selected variables
                    final boolean bl = lVar.isCoordinateVariable();
                    if (bl) {
                        return false;
                    }
                    final boolean bu = uVar.isCoordinateVariable();
                    if (bu) {
                        return false;
                    }
                    
                    // variables must have a dimension equal to the coordinate variable
                    List<Dimension> lDims = lVar.getDimensions();
                    if (!lDims.contains(cDim)) {
                        return false;
                    }
                    List<Dimension> uDims = uVar.getDimensions();
                    if (!uDims.contains(cDim)) {
                        return false;
                    }
                    
                } else if (luNameList.size() != 0) {
                    
                    String luName = luNameList.get(0);
                    SGVariable luVar = ncfile.findVariable(luName);
                    if (luVar == null) {
                        return false;
                    }
                    
                    // check validity of selected variables
                    final boolean blu = luVar.isCoordinateVariable();
                    if (blu) {
                        return false;
                    }
                    
                    // variables must have a dimension equal to the coordinate variable
                    List<Dimension> luDims = luVar.getDimensions();
                    if (!luDims.contains(cDim)) {
                        return false;
                    }
                }
            	
            } else {
            	// single variable
            	
                final List<String> lNameList = new ArrayList<String>();
                final List<Integer> lIndexList = new ArrayList<Integer>();
                final List<String> uNameList = new ArrayList<String>();
                final List<Integer> uIndexList = new ArrayList<Integer>();
                if (getColumnNameAndAppendedNumberList(cols, LOWER_ERROR_VALUE, lNameList,
                		lIndexList) == false) {
                	return false;
                }
                if (getColumnNameAndAppendedNumberList(cols, UPPER_ERROR_VALUE, uNameList,
                		uIndexList) == false) {
                	return false;
                }
                if (lNameList.size() != uNameList.size()) {
                    return false;
                }
                for (int ii = 0; ii < lNameList.size(); ii++) {
                    String lName = lNameList.get(ii);
                    String uName = uNameList.get(ii);
                    SGVariable lVar = ncfile.findVariable(lName);
                    if (lVar == null) {
                        return false;
                    }
                    SGVariable uVar = ncfile.findVariable(uName);
                    if (uVar == null) {
                        return false;
                    }
                    
                    // check validity of selected variables
                    final boolean bl = lVar.isCoordinateVariable();
                    if (bl) {
                        return false;
                    }
                    final boolean bu = uVar.isCoordinateVariable();
                    if (bu) {
                        return false;
                    }
                    
                    // variables must have a dimension equal to the coordinate variable
                    List<Dimension> lDims = lVar.getDimensions();
                    if (!lDims.contains(cDim)) {
                        return false;
                    }
                    List<Dimension> uDims = uVar.getDimensions();
                    if (!uDims.contains(cDim)) {
                        return false;
                    }
                }
                
                final List<String> luNameList = new ArrayList<String>();
                final List<Integer> luIndexList = new ArrayList<Integer>();
                if (getColumnNameAndAppendedNumberList(cols, LOWER_UPPER_ERROR_VALUE, 
                		luNameList, luIndexList) == false) {
                	return false;
                }
                for (int ii = 0; ii < luNameList.size(); ii++) {

                    String luName = luNameList.get(ii);
                    SGVariable luVar = ncfile.findVariable(luName);
                    if (luVar == null) {
                        return false;
                    }
                    
                    // check validity of selected variables
                    final boolean blu = luVar.isCoordinateVariable();
                    if (blu) {
                        return false;
                    }
                    
                    // variables must have a dimension equal to the coordinate variable
                    List<Dimension> luDims = luVar.getDimensions();
                    if (!luDims.contains(cDim)) {
                        return false;
                    }
                    
                	// checks whether lower/upper indices is contained
                	// in lower indices or upper indices
                	Integer luIndex = luIndexList.get(ii);
                	if (lIndexList.contains(luIndex)) {
                		return false;
                	}
                	if (uIndexList.contains(luIndex)) {
                		return false;
                	}
                }
            }
            
            // time
            if (timeVar != null) {
                for (int ii = 0; ii < oVarList.size(); ii++) {
                	SGVariable oVar = oVarList.get(ii);
                   	List<Dimension> oDims = oVar.getDimensions();
                    if (!oDims.contains(timeDim)) {
                        return false;
                    }
                }
            }
            
            // serial numbers
            if (serialNumberVar != null) {
                for (int ii = 0; ii < oVarList.size(); ii++) {
                	SGVariable oVar = oVarList.get(ii);
                   	List<Dimension> oDims = oVar.getDimensions();
                    if (!oDims.contains(serialNumberDim)) {
                        return false;
                    }
                }
            }
            
            // multiple
            if (!multipleVariable.booleanValue()) {
                // multiple dimension
        		
                final List<String> dimList = getColumnNameList(cols, PICKUP);
                if (dimList.size() != 1) {
                    return false;
                }
                String multDimName = dimList.get(0);
                Dimension multDim = ncfile.findDimension(multDimName);
                if (multDim == null) {
                    return false;
                }
                if (multDim.equals(cDim) || multDim.equals(timeDim)) {
                    return false;
                }
                SGVariable oVar = oVarList.get(0);
                List<Dimension> oDims = oVar.getDimensions();
                if (!oDims.contains(multDim)) {
                    return false;
                }

                Integer start = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_START);
                if (start == null) {
                    return false;
                }
                if (start.intValue() < 0 || start.intValue() >= multDim.getLength()) {
                    return false;
                }
                Integer end = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_END);
                if (end == null) {
                    return false;
                }
                if (end.intValue() < 0 || end.intValue() >= multDim.getLength()) {
                    return false;
                }

                Integer step = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_STEP);
                if (step == null) {
                    return false;
                }
            }

        } else if (SGDataTypeConstants.VXY_NETCDF_DATA.equals(dataType)) {
            
            final boolean polar = isPolar(infoMap);
            String key1 = null;
            String key2 = null;
            if (polar) {
                key1 = MAGNITUDE;
                key2 = ANGLE;
            } else {
                key1 = X_COMPONENT;
                key2 = Y_COMPONENT;
            }
            
            final List<String> xNameList = getColumnNameList(cols, X_COORDINATE);
            if (xNameList.size() != 1) {
                return false;
            }
            final List<String> yNameList = getColumnNameList(cols, Y_COORDINATE);
            if (yNameList.size() != 1) {
                return false;
            }
            final List<String> com1NameList = getColumnNameList(cols, key1);
            if (com1NameList.size() != 1) {
                return false;
            }
            final List<String> com2NameList = getColumnNameList(cols, key2);
            if (com2NameList.size() != 1) {
                return false;
            }
            String xName = xNameList.get(0);
            String yName = yNameList.get(0);
            String com1Name = com1NameList.get(0);
            String com2Name = com2NameList.get(0);
            SGVariable xVar = ncfile.findVariable(xName);
            if (xVar == null) {
                return false;
            }
            SGVariable yVar = ncfile.findVariable(yName);
            if (yVar == null) {
                return false;
            }
            SGVariable com1Var = ncfile.findVariable(com1Name);
            if (com1Var == null) {
                return false;
            }
            SGVariable com2Var = ncfile.findVariable(com2Name);
            if (com2Var == null) {
                return false;
            }
            
            // check validity of selected variables
            if (serialNumberVar != null) {
                if (xVar.isCoordinateVariable()) {
                    return false;
                }
                if (yVar.isCoordinateVariable()) {
                    return false;
                }
            } else {
                if (!xVar.isCoordinateVariable()) {
                    return false;
                }
                if (!yVar.isCoordinateVariable()) {
                    return false;
                }
            }
            if (com1Var.isCoordinateVariable()) {
                return false;
            }
            if (com2Var.isCoordinateVariable()) {
                return false;
            }
            
            List<Dimension> xDims = xVar.getDimensions();
            List<Dimension> yDims = yVar.getDimensions();
            List<Dimension> com1Dims = com1Var.getDimensions();
            List<Dimension> com2Dims = com2Var.getDimensions();
            if (serialNumberVar != null) {
                if (xDims.contains(serialNumberDim) == false) {
                    return false;
                }
                if (yDims.contains(serialNumberDim) == false) {
                    return false;
                }
                if (com1Dims.contains(serialNumberDim) == false) {
                    return false;
                }
                if (com2Dims.contains(serialNumberDim) == false) {
                    return false;
                }
            } else {
                Dimension xDim = xDims.get(0);
                Dimension yDim = yDims.get(0);
                if (com1Dims.contains(xDim) == false) {
                    return false;
                }
                if (com1Dims.contains(yDim) == false) {
                    return false;
                }
                if (com2Dims.contains(xDim) == false) {
                    return false;
                }
                if (com2Dims.contains(yDim) == false) {
                    return false;
                }
            }
            
            // time
            if (timeVar != null) {
            	if (serialNumberDim != null) {
                    if (!xDims.contains(timeDim)) {
                        return false;
                    }
                    if (!yDims.contains(timeDim)) {
                        return false;
                    }
            	}
                if (!com1Dims.contains(timeDim)) {
                    return false;
                }
                if (!com2Dims.contains(timeDim)) {
                    return false;
                }
            }

        } else if (SGDataTypeConstants.SXYZ_NETCDF_DATA.equals(dataType)) {
            
            //
            // x, y and z values
            //
            
            final List<String> xNameList = getColumnNameList(cols, X_VALUE);
            if (xNameList.size() != 1) {
                return false;
            }
            final List<String> yNameList = getColumnNameList(cols, Y_VALUE);
            if (yNameList.size() != 1) {
                return false;
            }
            final List<String> zNameList = getColumnNameList(cols, Z_VALUE);
            if (zNameList.size() != 1) {
                return false;
            }
            String xName = xNameList.get(0);
            String yName = yNameList.get(0);
            String zName = zNameList.get(0);
            SGVariable xVar = ncfile.findVariable(xName);
            if (xVar == null) {
                return false;
            }
            SGVariable yVar = ncfile.findVariable(yName);
            if (yVar == null) {
                return false;
            }
            SGVariable zVar = ncfile.findVariable(zName);
            if (zVar == null) {
                return false;
            }
            
            // x and y variables must be coordinate variables
            // z variable must not be a coordinate variable
            if (serialNumberVar != null) {
                if (xVar.isCoordinateVariable()) {
                    return false;
                }
                if (yVar.isCoordinateVariable()) {
                    return false;
                }
            } else {
                if (!xVar.isCoordinateVariable()) {
                    return false;
                }
                if (!yVar.isCoordinateVariable()) {
                    return false;
                }
            }
            if (zVar.isCoordinateVariable()) {
                return false;
            }
            
            // variable must have a dimension equal to the coordinate variable
            List<Dimension> xDims = xVar.getDimensions();
            List<Dimension> yDims = yVar.getDimensions();
            List<Dimension> zDims = zVar.getDimensions();
            if (serialNumberVar != null) {
            	if (!xDims.contains(serialNumberDim)) {
            		return false;
            	}
            	if (!yDims.contains(serialNumberDim)) {
            		return false;
            	}
            	if (!zDims.contains(serialNumberDim)) {
            		return false;
            	}
            } else {
                Dimension xDim = xDims.get(0);
                Dimension yDim = yDims.get(0);
                if (!zDims.contains(xDim)) {
                    return false;
                }
                if (!zDims.contains(yDim)) {
                    return false;
                }
            }
            
            // time
            if (timeVar != null) {
            	if (serialNumberVar != null) {
                	if (!xDims.contains(timeDim)) {
                		return false;
                	}
                	if (!yDims.contains(timeDim)) {
                		return false;
                	}
            	}
                if (!zDims.contains(timeDim)) {
                    return false;
                }
            }
            
        } else {
            throw new Error("Invalid data type: " + dataType);
        }
        
        return true;
    }
    
    /**
     * Returns the maximum values in a given array excepting NaN and infinity.
     * 
     * @param array
     *           an array of values
     * @return
     *           the maximum value
     */
    public static double getMaxValue(double[] array) {
        boolean valid = false;
        double max = - Double.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
            if (Double.isNaN(array[ii]) || Double.isInfinite(array[ii])) {
                continue;
            }
            if (array[ii] > max) {
                max = array[ii];
            }
            valid = true;
        }
        if (!valid) {
        	return Double.NaN;
        }
        return max;
    }
    
    /**
     * Returns the minimum values in a given array excepting NaN and infinity.
     * 
     * @param array
     *           an array of values
     * @return
     *           the minimum value
     */
    public static double getMinValue(double[] array) {
        boolean valid = false;
        double min = Double.MAX_VALUE;
        for (int ii = 0; ii < array.length; ii++) {
            if (Double.isNaN(array[ii]) || Double.isInfinite(array[ii])) {
                continue;
            }
            if (array[ii] < min) {
                min = array[ii];
            }
            valid = true;
        }
        if (!valid) {
        	return Double.NaN;
        }
        return min;
    }
    
    /**
     * Returns the bounds from given array.
     * 
     * @param array
     *          an array of values to search the mininum and maximum value
     * @return the bounds of given array
     */
    public static SGValueRange getBounds(final double[] array) {
        return new SGValueRange(getMinValue(array), getMaxValue(array));
    }

    /**
     * Returns the bounds of x-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of x-values
     */
    public static SGValueRange getBoundsX(SGISXYTypeData data) {
        final double[] xValues = data.getXValueArray();
        if (xValues == null) {
            return null;
        }
        if (data.isErrorBarAvailable() && !data.isErrorBarVertical()) {
        	return getBounds(data, xValues);
        } else {
            return getBounds(xValues);
        }
    }

    /**
     * Returns the bounds of y-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of y-values
     */
    public static SGValueRange getBoundsY(SGISXYTypeData data) {
        final double[] yValues = data.getYValueArray();
        if (yValues == null) {
            return null;
        }
        if (data.isErrorBarAvailable() && data.isErrorBarVertical()) {
        	return getBounds(data, yValues);
        } else {
            return getBounds(yValues);
        }
    }
    
    private static SGValueRange getBounds(SGISXYTypeData data, double[] values) {
        final double[] lArray = data.getLowerErrorValueArray();
        final double[] uArray = data.getUpperErrorValueArray();
        final double[] lyValues = new double[values.length];
        final double[] uyValues = new double[values.length];
        for (int ii = 0; ii < values.length; ii++) {
            lyValues[ii] = values[ii] + lArray[ii];
            uyValues[ii] = values[ii] + uArray[ii];
        }
        SGValueRange vBounds = getBounds(values);
        SGValueRange lBounds = getBounds(lyValues);
        SGValueRange uBounds = getBounds(uyValues);
        List<SGValueRange> rangeList = new ArrayList<SGValueRange>();
        rangeList.add(vBounds);
        rangeList.add(lBounds);
        rangeList.add(uBounds);
        Double min = getMinValue(rangeList);
        Double max = getMaxValue(rangeList);
        final double minValue = (min != null) ? min.doubleValue() : Double.NaN;
        final double maxValue = (max != null) ? max.doubleValue() : Double.NaN;
        SGValueRange bounds = new SGValueRange(minValue, maxValue);
        return bounds;
    }
    
    public static Double getMinValue(List<SGValueRange> rangeList) {
    	Double min = null;
    	for (int ii = 0; ii < rangeList.size(); ii++) {
    		SGValueRange range = rangeList.get(ii);
    		if (range.isMinValid()) {
    			final double value = range.getMinValue();
    			if (min == null) {
    				min = Double.valueOf(value);
    			} else {
    				if (value < min.doubleValue()) {
    					min = Double.valueOf(value);
    				}
    			}
    		}
    	}
    	return min;
    }

    public static Double getMaxValue(List<SGValueRange> rangeList) {
    	Double max = null;
    	for (int ii = 0; ii < rangeList.size(); ii++) {
    		SGValueRange range = rangeList.get(ii);
    		if (range.isMaxValid()) {
    			final double value = range.getMaxValue();
    			if (max == null) {
    				max = Double.valueOf(value);
    			} else {
    				if (value > max.doubleValue()) {
    					max = Double.valueOf(value);
    				}
    			}
    		}
    	}
    	return max;
    }

    /**
     * Returns the bounds of x-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of x-values
     */
    public static SGValueRange getBoundsX(SGISXYTypeMultipleData data) {
        SGISXYTypeData[] sxyArray = data.getSXYDataArray();
        List<SGValueRange> rangeList = new ArrayList<SGValueRange>();
        for (int ii = 0; ii < sxyArray.length; ii++) {
        	SGValueRange range = getBoundsX(sxyArray[ii]);
        	rangeList.add(range);
        }
        
    	// disposes of data objects
    	for (int ii = 0; ii < sxyArray.length; ii++) {
    		sxyArray[ii].dispose();
    	}    	

    	return getBounds(rangeList);
    }

    /**
     * Returns the bounds of y-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of y-values
     */
    public static SGValueRange getBoundsY(SGISXYTypeMultipleData data) {
        SGISXYTypeData[] sxyArray = data.getSXYDataArray();
        List<SGValueRange> rangeList = new ArrayList<SGValueRange>();
        for (int ii = 0; ii < sxyArray.length; ii++) {
        	SGValueRange range = getBoundsY(sxyArray[ii]);
        	rangeList.add(range);
        }
        
    	// disposes of data objects
    	for (int ii = 0; ii < sxyArray.length; ii++) {
    		sxyArray[ii].dispose();
    	}    	
    	
    	return getBounds(rangeList);
    }
    
    private static SGValueRange getBounds(List<SGValueRange> rangeList) {
        Double min = getMinValue(rangeList);
        Double max = getMaxValue(rangeList);
    	final double minValue = (min != null) ? min.doubleValue() : Double.NaN;
    	final double maxValue = (max != null) ? max.doubleValue() : Double.NaN;
    	return new SGValueRange(minValue, maxValue);
    }

    /**
     * Returns the bounds of x-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of x-values
     */
    public static SGValueRange getBoundsX(SGIVXYTypeData data) {
        final SGTuple2d[] coordinates = data.getCoordinateArray();
        if (coordinates == null) {
            return null;
        }
        final int len = coordinates.length;
        double[] values = new double[len];
        for (int ii = 0; ii < len; ii++) {
            values[ii] = coordinates[ii].x;
        }
        return getBounds(values);
    }

    /**
     * Returns the bounds of y-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of y-values
     */
    public static SGValueRange getBoundsY(SGIVXYTypeData data) {
        final SGTuple2d[] coordinates = data.getCoordinateArray();
        if (coordinates == null) {
            return null;
        }
        final int len = coordinates.length;
        double[] values = new double[len];
        for (int ii = 0; ii < len; ii++) {
            values[ii] = coordinates[ii].y;
        }
        return getBounds(values);
    }

    /**
     * Returns whether the given type of data can have tick labels.
     * 
     * @param dataType
     *           the data type
     * @return
     *           true if the given type of data can have tick labels
     */
    public static boolean hasTickLabels(final String dataType) {
        return (SGDataTypeConstants.SXY_DATA.equals(dataType)
                || SGDataTypeConstants.SXY_DATE_DATA.equals(dataType));
    }
    
    /**
     * Returns the bounds of x-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of x-values
     */
    public static SGValueRange getBoundsX(SGISXYZTypeData data) {
        final double[] values = data.getXValueArray();
        if (values == null) {
            return null;
        }
        return getBounds(values);
    }

    /**
     * Returns the bounds of y-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of y-values
     */
    public static SGValueRange getBoundsY(SGISXYZTypeData data) {
        final double[] values = data.getYValueArray();
        if (values == null) {
            return null;
        }
        return getBounds(values);
    }

    /**
     * Returns the bounds of z-values of given data.
     * 
     * @param data
     *           a data
     * @return the bounds of z-values
     */
    public static SGValueRange getBoundsZ(SGISXYZTypeData data) {
        final double[] values = data.getZValueArray();
        if (values == null) {
            return null;
        }
        return getBounds(values);
    }

    /**
     * Returns whether the given data is of the array type.
     * 
     * @param data
     *           the data object
     * @return
     *           true if the given data is of the array type
     */
    public static boolean isArrayData(final SGData data) {
        return isArrayData(data.getDataType());
    }

    /**
     * Returns whether the given data is of the netCDF type.
     * 
     * @param data
     *           the data object
     * @return
     *           true if the given data is of the netCDF type
     */
    public static boolean isNetCDFData(final SGData data) {
        return isNetCDFData(data.getDataType());
    }

    /**
     * Returns whether the given type of data is of the array type.
     * 
     * @param dataType
     *           the data type
     * @return
     *           true if the given type of data is of the array type
     */
    public static boolean isArrayData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = { SGDataTypeConstants.SXY_DATA, 
                SGDataTypeConstants.SXY_MULTIPLE_DATA, 
                SGDataTypeConstants.SXY_SAMPLING_DATA,
                SGDataTypeConstants.VXY_DATA,
                SGDataTypeConstants.SXY_DATE_DATA,
                SGDataTypeConstants.SXYZ_DATA };
        return SGUtility.contains(types, dataType);
    }

    /**
     * Returns whether the given type of data is of the netCDF type.
     * 
     * @param dataType
     *           the data type
     * @return
     *           true if the given type of data is of the netCDF type
     */
    public static boolean isNetCDFData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = {
                SGDataTypeConstants.SXY_NETCDF_DATA, 
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA, 
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA,
                SGDataTypeConstants.VXY_NETCDF_DATA,
                SGDataTypeConstants.SXYZ_NETCDF_DATA
        };
        return SGUtility.contains(types, dataType);
    }
    
    /**
     * Returns whether the given type of data is of the scalar-XY type.
     * (Sampling XY data and date XY data are excepted.)
     * 
     * @param dataType
     *           the data type
     * @return true if the given type of data is of the scalar-XY type
     */
    public static boolean isSXYTypeData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = {
                SGDataTypeConstants.SXY_DATA,
                SGDataTypeConstants.SXY_MULTIPLE_DATA,
                SGDataTypeConstants.SXY_NETCDF_DATA,
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA,
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA
        };
        return SGUtility.contains(types, dataType);
    }

    /**
     * Returns whether the given type of data is of the vector-XY type.
     * 
     * @param dataType
     *           the data type
     * @return true if the given type of data is of the vector-XY type
     */
    public static boolean isVXYTypeData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = { SGDataTypeConstants.VXY_DATA,
				SGDataTypeConstants.VXY_NETCDF_DATA };
        return SGUtility.contains(types, dataType);
    }

    /**
     * Returns whether the given type of data is of the multiple type data.
     * This method is only for backward compatibility.
     * 
     * @param dataType
     *           the data type
     * @return true if the given type of data is of the multiple type
     */
    public static boolean isMultipleData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = {
                SGDataTypeConstants.SXY_MULTIPLE_DATA,
                SGDataTypeConstants.SXY_SAMPLING_DATA,
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DATA,
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA
        };
        return SGUtility.contains(types, dataType);
    }
    
    /**
     * Returns whether the given type of data is of the multiple type
     * netCDF dimension data.
     * 
     * @param dataType
     * @return true if the given type of data is of the multiple type of netCDF dimension
     * @throws IllegalArgumentException if dataType is null
     */
    public static boolean isNetCDFDimensionData(final String dataType) {
        if (dataType == null) {
            throw new IllegalArgumentException("Input data type is null.");
        }
        String[] types = {
                SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA
        };
        return SGUtility.contains(types, dataType);
    }

    /**
     * Returns whether the given data type is valid.
     * 
     * @param dataType
     *           the data type
     * @return
     *           true if the given type of data is valid
     */
    public static boolean isValidData(final String dataType) {
        if (isArrayData(dataType)) {
            return true;
        }
        if (isNetCDFData(dataType)) {
            return true;
        }
        return false;
    }
    
    /**
     * Updates the information map with given data type and column information.
     * 
     * @param dataType
     *           the data type
     * @param colInfo
     *           the column information
     * @param infoMap
     *           the information map
     * @return updated information map
     */
    public static Map<String, Object> updateInfoMap(String dataType, 
    		SGDataColumnInfo[] colInfo, Map<String, Object> infoMap) {
    	Map<String, Object> infoMapUpd = new HashMap<String, Object>(infoMap);
        if (isSXYTypeData(dataType)) {
        	// column information
        	infoMapUpd.put(KEY_COLUMN_INFO, colInfo.clone());
        }
    	return infoMapUpd;
    }
    
    /**
     * Updates the column types.
     * 
     * @param dataType
     *            data type
     * @param colInfo
     *            column information
     * @param columnTypes
     *            an array of columns types
     * @return new column types
     */
    public static String[] updateDataColumns(final String dataType,
            final SGDataColumnInfo[] colInfo, final String[] columnTypes) {
    	
    	String[] ret = (String[]) columnTypes.clone();
        if (isSXYTypeData(dataType)) {
        	
            // find the column for X or Y
            List<Integer> xIndexList = new ArrayList<Integer>();
            List<Integer> yIndexList = new ArrayList<Integer>();
            for (int ii = 0; ii < columnTypes.length; ii++) {
                final Integer index = Integer.valueOf(ii);
                if (X_VALUE.equals(columnTypes[ii])) {
                    xIndexList.add(index);
                } else if (Y_VALUE.equals(columnTypes[ii])) {
                    yIndexList.add(index);
                }
            }

            // get the list of column index those are assignable for the error bar
            List<Integer> indexList = getOptionalColumnsAssignableIndexList(
                    dataType, colInfo, xIndexList, yIndexList);
            if (indexList == null) {
                return null;
            }

            // clear the column type for error bars for non-existing X or Y
            for (int ii = 0; ii < columnTypes.length; ii++) {
                String colType = columnTypes[ii];
                if (colType.startsWith(LOWER_ERROR_VALUE)
                        || colType.startsWith(UPPER_ERROR_VALUE)
                        || colType.startsWith(LOWER_UPPER_ERROR_VALUE)
                        || colType.startsWith(TICK_LABEL)) {
                    // get the column number
                    Integer num = getColumnIndexOfAppendedColumnType(colType, isNetCDFData(dataType), colInfo);
                    if (num == null) {
                        return null;
                    }
                    if (!indexList.contains(num)) {
                        // clear the column type
                        ret[ii] = "";
                    }
                }
            }
        }

        return ret;
    }
    
    /**
     * Returns the list of column indices those are assignable optional columns
     * such as error bars tick labels.
     * 
     */
    private static List<Integer> getOptionalColumnsAssignableIndexList(
    		String dataType, SGDataColumnInfo[] colInfo,
    		List<Integer> xIndexList, List<Integer> yIndexList) {
    	if (!isSXYTypeData(dataType)) {
    		return null;
    	}
    	Set<Integer> indexSet = new TreeSet<Integer>();
//    	if (xIndexList.size() == 1 && yIndexList.size() == 1) {
//    		indexSet.addAll(xIndexList);
//    		indexSet.addAll(yIndexList);
//    	} else if (xIndexList.size() > 1 && yIndexList.size() == 1) {
//        	indexSet.addAll(xIndexList);
//    	} else if (yIndexList.size() > 1 && xIndexList.size() == 1){
//        	indexSet.addAll(yIndexList);
//    	} else {
//    		if (xIndexList.size() == 0 && yIndexList.size() != 0) {
//            	indexSet.addAll(yIndexList);
//    		} else if (yIndexList.size() == 0 && xIndexList.size() != 0) {
//            	indexSet.addAll(xIndexList);
//    		}
//    	}
		indexSet.addAll(xIndexList);
		indexSet.addAll(yIndexList);
    	List<Integer> indexList = new ArrayList<Integer>();
    	if (isArrayData(dataType)) {
    		indexList.addAll(indexSet);
    	} else if (isNetCDFData(dataType)) {
    		for (int ii = 0; ii < colInfo.length; ii++) {
    			SGNetCDFDataColumnInfo ncInfo = (SGNetCDFDataColumnInfo) colInfo[ii];
    			Integer index = Integer.valueOf(ii);
    			if (indexSet.contains(index)) {
        			SGVariable var = ncInfo.getVariable();
        			if (!var.isCoordinateVariable()) {
        				indexList.add(index);
        			}
    			}
    		}
    	} else {
    		return null;
    	}
    	return indexList;
    }
    
    /**
     * Creates a text string for the title with given sampling rate.
     * 
     * @return the title for this sampling data
     */
    public static String createSamplingRateTitle(final double samplingRate) {
    	StringBuffer sb = new StringBuffer();
    	sb.append("Sampling Rate ");
    	sb.append(samplingRate);
    	sb.append(" Hz");
    	return sb.toString();
    }

    /**
     * Get index list of column type on text data.
     * <p>
     * If column type in <b>colInfo</b> matches "X", add the index of <b>colInfo</b>
     * into <b>xIndexList</b>. Same, check matching of "Y", "Lower Error", "Upper Error",
     * "Lower / Upper Error" and "Tick Label".
     * 
     * @param colInfo
     * @param xIndexList
     * @param yIndexList
     * @param lIndexMap
     * @param uIndexMap
     * @param tIndexMap
     * @return false if column type name is a wrong name. true if succeeds.
     */
    public static boolean getSXYColumnType(SGDataColumnInfo[] colInfo,
    		List<Integer> xIndexList, List<Integer> yIndexList,
    		Map<Integer, Integer> lIndexMap, Map<Integer, Integer> uIndexMap,
    		Map<Integer, Integer> tIndexMap) {
        for (int ii = 0; ii < colInfo.length; ii++) {
            String cType = colInfo[ii].getColumnType();
            if (SGUtilityText.isEqualString(X_VALUE, cType)) {
            	xIndexList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.isEqualString(Y_VALUE, cType)) {
            	yIndexList.add(Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(cType, LOWER_ERROR_VALUE)) {
            	Integer num = SGDataUtility.getAppendedColumnIndex(cType, colInfo);
            	lIndexMap.put(num, Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(cType, UPPER_ERROR_VALUE)) {
            	Integer num = SGDataUtility.getAppendedColumnIndex(cType, colInfo);
            	uIndexMap.put(num, Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(cType, LOWER_UPPER_ERROR_VALUE)) {
            	Integer num = SGDataUtility.getAppendedColumnIndex(cType, colInfo);
            	lIndexMap.put(num, Integer.valueOf(ii));
            	uIndexMap.put(num, Integer.valueOf(ii));
            } else if (SGUtilityText.startsWith(cType, TICK_LABEL)) {
            	Integer num = SGDataUtility.getAppendedColumnIndex(cType, colInfo);
            	tIndexMap.put(num, Integer.valueOf(ii));
            } else if ("".equals(cType)) {
                continue;
            } else {
            	return false;
            }
        }
        return true;
    }
    
    /**
     * Get name list of column type on netCDF data.
     * <p>
     * If column type in <b>colInfo</b> matches "X", add the variable name of <b>colInfo</b>
     * into <b>xNameList</b>. Same, check matching of "Y", "Lower Error", "Upper Error",
     * "Lower / Upper Error", "Tick Label", "Time" and "Serial Number".
     * 
     * @param colInfo
     * @param xNameList contains variable names which have "X" column type.
     * @param yNameList
     * @param lNameMap
     * @param uNameMap
     * @param tickNameMap
     * @param timeNameList
     * @param serialNumberNameList
     * @return
     */
    public static boolean getSXYNetCDFColumnType(SGDataColumnInfo[] colInfo,
            List<String> xNameList, List<String> yNameList,
            Map<String, String> lNameMap, Map<String, String> uNameMap,
            Map<String, String> tickNameMap, List<String> timeNameList,
            List<String> serialNumberNameList) {
        for (int ii = 0; ii < colInfo.length; ii++) {
            SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[ii];
            String cType = col.getColumnType();
            String name = col.getVariableName();
            if (SGUtilityText.isEqualString(X_VALUE, cType)) {
                xNameList.add(name);
            } else if (SGUtilityText.isEqualString(Y_VALUE, cType)) {
                yNameList.add(name);
            } else if (SGUtilityText.startsWith(cType, LOWER_ERROR_VALUE)) {
                Integer num = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(cType, colInfo);
                SGNetCDFDataColumnInfo colAppend = (SGNetCDFDataColumnInfo) colInfo[num.intValue()];
                String varName = colAppend.getVariableName();
                lNameMap.put(varName, name);
            } else if (SGUtilityText.startsWith(cType, UPPER_ERROR_VALUE)) {
                Integer num = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(cType, colInfo);
                SGNetCDFDataColumnInfo colAppend = (SGNetCDFDataColumnInfo) colInfo[num.intValue()];
                String varName = colAppend.getVariableName();
                uNameMap.put(varName, name);
            } else if (SGUtilityText.startsWith(cType, LOWER_UPPER_ERROR_VALUE)) {
                Integer num = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(cType, colInfo);
                SGNetCDFDataColumnInfo colAppend = (SGNetCDFDataColumnInfo) colInfo[num.intValue()];
                String varName = colAppend.getVariableName();
                lNameMap.put(varName, name);
                uNameMap.put(varName, name);
            } else if (SGUtilityText.startsWith(cType, TICK_LABEL)) {
                Integer num = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(cType, colInfo);
                SGNetCDFDataColumnInfo colAppend = (SGNetCDFDataColumnInfo) colInfo[num.intValue()];
                String varName = colAppend.getVariableName();
                tickNameMap.put(varName, name);
            } else if (SGUtilityText.isEqualString(TIME, cType)) {
                timeNameList.add(name);
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, cType)) {
                serialNumberNameList.add(name);
            } else if ("".equals(cType)) {
                continue;
            } else {
                return false;
            }
        }
        return true;
    }

    /**
     * Get name list of column type on netCDF data.
     * <p>
     * If column type in <b>colInfo</b> matches "X", add the variable name of <b>colInfo</b>
     * into <b>xNameList</b>. Same, check matching of "Y", "Lower Error", "Upper Error",
     * "Lower / Upper Error", "Tick Label", "Dimension", "Time" and "Index".
     * 
     * @param colInfo
     * @param xNameList contains variable names which have "X" column type.
     * @param yNameList
     * @param lNameList
     * @param uNameList
     * @param tNameList
     * @param mNameList
     * @param timeNameList
     * @param serialNumberNameList
     * @return
     */
    public static boolean getSXYNetCDFColumnType(SGDataColumnInfo[] colInfo,
            List<String> xNameList, List<String> yNameList,
            List<String> lNameList, List<String> uNameList,
            List<String> tNameList, List<String> mNameList,
            List<String> timeNameList, List<String> serialNumberNameList) {
        for (int ii = 0; ii < colInfo.length; ii++) {
            SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[ii];
            String cType = col.getColumnType();
            String name = col.getVariableName();
            if (SGUtilityText.isEqualString(X_VALUE, cType)) {
                xNameList.add(name);
            } else if (SGUtilityText.isEqualString(Y_VALUE, cType)) {
                yNameList.add(name);
            } else if (SGUtilityText.isEqualString(LOWER_ERROR_VALUE, cType)) {
                lNameList.add(name);
            } else if (SGUtilityText.isEqualString(UPPER_ERROR_VALUE, cType)) {
                uNameList.add(name);
            } else if (SGUtilityText.isEqualString(LOWER_UPPER_ERROR_VALUE, cType)) {
                lNameList.add(name);
                uNameList.add(name);
            } else if (SGUtilityText.isEqualString(TICK_LABEL, cType)) {
                tNameList.add(name);
            } else if (SGUtilityText.isEqualString(PICKUP, cType)) {
                mNameList.add(name);
            } else if (SGUtilityText.isEqualString(TIME, cType)) {
                timeNameList.add(name);
            } else if (SGUtilityText.isEqualString(SERIAL_NUMBER, cType)) {
                serialNumberNameList.add(name);
            } else if ("".equals(cType)) {
                continue;
            } else {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns whether the complement button is to be visible.
     * <p>
     * If data type is Scalar-XY, and multiple graphs drawing and
     * multiple variables are selected, complement button is visible.
     * 
     * @param infoMap
     *           the information map
     * @return true to set visible
     */
    public static boolean isComplementButtonVisible(final Map<String, Object> infoMap) {
        String dataType = (String) infoMap.get(KEY_DATA_TYPE);
        if (!isSXYTypeData(dataType)) {
            return false;
        }
        Boolean multiple = (Boolean) infoMap.get(KEY_MULTIPLE);
        if (multiple == null) {
            return false;
        }
        final boolean compVisible;
        if (multiple.booleanValue()) {
            Boolean multipleVariable = (Boolean) infoMap.get(KEY_MULTIPLE_VARIABLE);
            if (multipleVariable == null) {
                return false;
            }
            compVisible = multipleVariable.booleanValue();
        } else {
            compVisible = false;
        }
        return compVisible;
    }
    
    /**
     * Returns whether the complement button is to be enabled.
     * <p>
     * If data type is scalar-XY, and if number of "X" column type is 1 or
     * number of "Y" is 1, the complement button is enabled.
     * 
     * @param infoMap the information map
     * @param curColType current column type
     * @param colInfoList list of column information
     * @return true to set visible
     */
    public static boolean isComplementedButtonEnabled(
            final Map<String, Object> infoMap,
            final String[] curColType, final List<SGDataColumnInfo> colInfoList) {
        String dataType = (String) infoMap.get(KEY_DATA_TYPE);
        if (isSXYTypeData(dataType)) {
            List<Integer> xIndexList = new ArrayList<Integer>();
            List<Integer> yIndexList = new ArrayList<Integer>();
            for (int ii = 0; ii < curColType.length; ii++) {
                if (X_VALUE.equals(curColType[ii])) {
                    xIndexList.add(Integer.valueOf(ii));
                } else if (Y_VALUE.equals(curColType[ii])) {
                    yIndexList.add(Integer.valueOf(ii));
                }
            }
            if (isArrayData(dataType)) {
                if (xIndexList.size() == 1 || yIndexList.size() == 1) {
                    return true;
                } else {
                    return false;
                }
            } else if (isNetCDFData(dataType)) {
                if (xIndexList.size() == 1) {
                    return isComplementedButtonEnabled(infoMap, curColType, colInfoList, xIndexList.get(0));
                } else if (yIndexList.size() == 1) {
                    return isComplementedButtonEnabled(infoMap, curColType, colInfoList, yIndexList.get(0));
                } else {
                    return false;
                }
            } else {
                return false;
            }
        } else {
            return false;
        }
    }
    
    /**
     * If data is netCDF, single data selected must be a coordinate variable.
     */
    private static boolean isComplementedButtonEnabled(
            final Map<String, Object> infoMap, final String[] curColType,
            final List<SGDataColumnInfo> colInfoList, final Integer singleIndex) {
        SGNetCDFDataColumnInfo singleColInfo = (SGNetCDFDataColumnInfo) colInfoList.get(singleIndex);
        SGVariable singleVar = singleColInfo.getVariable();
        if (!singleVar.isCoordinateVariable()) {
            return false;
        }
        return true;
    }
    
    /**
     * Complements the column types.
     * <p>
     * If data type is not scalar-XY, return current column type.
     * If data type is scalar-XY, and if number of "X" column type is 1 and "Y" is 0,
     * set "Y" column type to columns which value types are Number (array data) or
     * which dimensions contain the dimension of "X" (netCDF data).
     * Same for number of "X" is 0 and "Y" is 1.
     * Otherwise return current column type.
     * 
     * @param infoMap
     *            the information map
     * @param curColType
     *            the current column types
     * @param colInfoList
     *            the list of column information
     * @return complemented column types
     */
    public static String[] getComplementedColumnType(
            final Map<String, Object> infoMap,
            final String[] curColType, final List<SGDataColumnInfo> colInfoList) {
        if (curColType.length != colInfoList.size()) {
            return null;
        }
        String dataType = (String) infoMap.get(KEY_DATA_TYPE);
        String[] compColType = null;
        if (isSXYTypeData(dataType)) {
            List<Integer> xIndexList = new ArrayList<Integer>();
            List<Integer> yIndexList = new ArrayList<Integer>();
            for (int ii = 0; ii < curColType.length; ii++) {
                if (X_VALUE.equals(curColType[ii])) {
                    xIndexList.add(Integer.valueOf(ii));
                } else if (Y_VALUE.equals(curColType[ii])) {
                    yIndexList.add(Integer.valueOf(ii));
                }
            }
            if (xIndexList.size() == 1 && yIndexList.size() == 0) {
                compColType = getComplementedColumnType(infoMap, curColType,
                        colInfoList, xIndexList.get(0), X_VALUE, Y_VALUE);
            } else if (xIndexList.size() == 0 && yIndexList.size() == 1) {
                compColType = getComplementedColumnType(infoMap, curColType,
                        colInfoList, yIndexList.get(0), Y_VALUE, X_VALUE);
            } else {
                compColType = curColType.clone();
            }
        } else {
            compColType = curColType.clone();
        }
        return compColType;
    }
    
    private static String[] getComplementedColumnType(
            final Map<String, Object> infoMap,
            final String[] curColType, final List<SGDataColumnInfo> colInfoList,
            final Integer singleIndex, final String singleColumnType,
            final String multColumnType) {
    	
        String dataType = (String) infoMap.get(KEY_DATA_TYPE);
        String[] compColType = new String[curColType.length];
        compColType[singleIndex.intValue()] = singleColumnType;
        if (isArrayData(dataType)) {
            for (int ii = 0; ii < colInfoList.size(); ii++) {
                if (ii == singleIndex.intValue()) {
                    continue;
                }
                SGDataColumnInfo colInfo = colInfoList.get(ii);
                String valueType = colInfo.getValueType();
                if (VALUE_TYPE_NUMBER.equals(valueType)) {
                    compColType[ii] = multColumnType;
                } else {
                    compColType[ii] = curColType[ii];
                }
            }			
        } else if (isNetCDFData(dataType)) {
            SGNetCDFDataColumnInfo singleColInfo = (SGNetCDFDataColumnInfo) colInfoList.get(singleIndex);
            SGVariable singleVar = singleColInfo.getVariable();
            if (singleVar.isCoordinateVariable()) {
                Dimension cDim = singleVar.getDimension(0);
                for (int ii = 0; ii < colInfoList.size(); ii++) {
                    if (ii == singleIndex.intValue()) {
                        continue;
                    }
                    SGNetCDFDataColumnInfo nColInfo = (SGNetCDFDataColumnInfo) colInfoList.get(ii);
                    SGVariable var = nColInfo.getVariable();
                    if (var.isCoordinateVariable()) {
                        compColType[ii] = curColType[ii];
                    }
                    List<Dimension> dimList = var.getDimensions();
                    if (dimList.contains(cDim)) {
                        compColType[ii] = multColumnType;
                    } else {
                        compColType[ii] = "";
                    }
                }
            } else {
                compColType = curColType.clone();
            }
        }

    	return compColType;
    }
    
    /**
     * Calculates the dimension indices.
     * 
     * @param dim
     *            the dimension
     * @param dimStart
     *            the index start
     * @param dimEnd
     *            the index end
     * @param dimStep
     *            the index step
     * @return an array of dimension indices
     */
    public static Integer[] calcDimIndices(Dimension dim,
            Integer dimStart, Integer dimEnd, Integer dimStep) {
    	if (dim == null || dimStart == null || dimEnd == null || dimStep == null) {
    		throw new IllegalArgumentException(
    				"dim == null || dimStart == null || dimEnd == null || dimStep == null");
    	}
    	final int dimLen = dim.getLength();
    	int start = dimStart.intValue();
    	int end = dimEnd.intValue();
    	final int step = Math.abs(dimStep.intValue());
    	if (start < 0 || start > dimLen) {
    	    throw new IllegalArgumentException("Dimension start is out of bounds: " + start);
    	}
        if (end < 0) {
            throw new IllegalArgumentException("Dimension end is out of bounds: " + end);
        }
        if (end > dimLen) {
            end = dimLen-1;
        }
        Set<Integer> dimIndexSet = new TreeSet<Integer>();
        dimIndexSet.add(dimStart);
        if (step != 0) {
            if (start <= end) {
                int cnt = start;
                while (cnt >= 0 && cnt <= end) {
                    dimIndexSet.add(Integer.valueOf(cnt));
                    cnt += step;
                }
            } else {
                int cnt = start;
                while (cnt >= 0 && cnt >= end) {
                    dimIndexSet.add(Integer.valueOf(cnt));
                    cnt -= step;
                }
            }
        }
        Integer[] dimIndices = dimIndexSet.toArray(new Integer[dimIndexSet.size()]);
        return dimIndices;
    }
    
    /**
     * 
     * @param colInfo
     * @return true if colInfo has PICKUP column type.
     */
    public static boolean isPickupColumnContained(final SGDataColumnInfo[] colInfo) {
        for (int i = 0; i < colInfo.length; i++) {
            if (PICKUP.equals(colInfo[i].getColumnType())) {
                return true;
            }
        }
        return false;
    }
    
    /**
     * Convert SGSXYMultipleVariableNetCDFData data to SGSXYMultipleDimensionNetCDFData.
     * 
     * @param data
     * @param colInfo
     * @return
     */
    public static SGSXYMultipleNetCDFData toDimensionData(
            final SGSXYMultipleVariableNetCDFData data, final SGDataColumnInfo[] colInfo) {
        SGNetCDFFile ncFile = (SGNetCDFFile)data.mNCFile.clone();
        SGDataSourceObserver obs = data.mObserver;

        int startIndex = data.getDimensionStart();
        int endIndex = data.getDimensionEnd();
        int stepIndex = data.getDimensionStep();
        
        SGNetCDFDataColumnInfo[] ncols = new SGNetCDFDataColumnInfo[colInfo.length];
        for (int i = 0; i < colInfo.length; i++) {
            ncols[i] = (SGNetCDFDataColumnInfo)colInfo[i];
        }
        
        Map<String, Object> infoMap = new HashMap<String, Object>();
        SGDataUtility.updatePickupParameters(infoMap, ncols);
        if (startIndex<0) {
            startIndex = 0;
            Object start = infoMap.get(KEY_MULTIPLE_DIMENSION_START);
            if (start instanceof Integer) {
                startIndex = ((Integer)start).intValue();
            }
        }
        if (endIndex<0) {
            endIndex = 0;
            Object end = infoMap.get(KEY_MULTIPLE_DIMENSION_END);
            if (end instanceof Integer) {
                endIndex = ((Integer)end).intValue();
            }
        }
        if (stepIndex<0) {
            stepIndex = 0;
            Object step = infoMap.get(KEY_MULTIPLE_DIMENSION_STEP);
            if (step instanceof Integer) {
                stepIndex = ((Integer)step).intValue();
            }
        }

        SGSXYMultipleNetCDFData newData =
            createDimensionNetCDFData(ncFile, obs, colInfo, startIndex, endIndex, stepIndex);
        if (null==newData) {
            return null;
        }

        for (int i = 0; i < colInfo.length; i++) {
            SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[i];
            final String name = col.getVariableName();
            final int origin = col.getOrigin();
            if (ncFile.findDimension(name) != null) {
                newData.setOrigin(name, origin);
            }
        }

        return newData;
    }
    
    /**
     * Convert SGSXYMultipleDimensionNetCDFData data to SGSXYMultipleVariableNetCDFData.
     * 
     * @param data
     * @param colInfo
     * @return
     */
    public static SGSXYMultipleNetCDFData toVariableData(
            final SGSXYMultipleDimensionNetCDFData data, final SGDataColumnInfo[] colInfo) {
        SGNetCDFFile ncFile = (SGNetCDFFile)data.mNCFile.clone();
        SGDataSourceObserver obs = data.mObserver;

        SGSXYMultipleNetCDFData newData = createVariableNetCDFData(ncFile, obs, colInfo);
        if (null==newData) {
            return null;
        }
        
        newData.setDimensionStart(data.getDimensionStart());
        newData.setDimensionEnd(data.getDimensionEnd());
        newData.setDimensionStep(data.getDimensionStep());

        for (int i = 0; i < colInfo.length; i++) {
            SGNetCDFDataColumnInfo col = (SGNetCDFDataColumnInfo) colInfo[i];
            final String name = col.getVariableName();
            final int origin = col.getOrigin();
            if (ncFile.findDimension(name) != null) {
                newData.setOrigin(name, origin);
            }
        }

        return newData;
    }
    
    public static SGSXYMultipleDimensionNetCDFData createDimensionNetCDFData(
            final SGNetCDFFile ncFile,
            final SGDataSourceObserver obs,
            final SGDataColumnInfo[] colInfo,
            final int startIndex, final int endIndex, final int stepIndex) {
        
        List<String> xNameList = new ArrayList<String>();
        List<String> yNameList = new ArrayList<String>();
        List<String> lNameList = new ArrayList<String>();
        List<String> uNameList = new ArrayList<String>();
        List<String> tNameList = new ArrayList<String>();
        List<String> mNameList = new ArrayList<String>();
        List<String> timeNameList = new ArrayList<String>();
        List<String> indexNameList = new ArrayList<String>();
        if (SGDataUtility.getSXYNetCDFColumnType(
                colInfo, xNameList, yNameList,
                lNameList, uNameList, tNameList, mNameList, timeNameList,
                indexNameList) == false) {
            return null;
        }
        if (xNameList.size() != 1) {
            return null;
        }
        if (yNameList.size() != 1) {
            return null;
        }
        if (mNameList.size() != 1) {
            return null;
        }
        if (!(lNameList.size() == 0 && uNameList.size() == 0) 
                && !(lNameList.size() == 1 && uNameList.size() == 1)) {
            return null;
        }
        String xName = xNameList.get(0);
        String yName = yNameList.get(0);
        String lName = null;
        if (lNameList.size() > 0) {
            lName = lNameList.get(0);
        }
        String uName = null;
        if (uNameList.size() > 0) {
            uName = uNameList.get(0);
        }
        String tName = null;
        if (tNameList.size() > 0) {
            tName = tNameList.get(0);
        }
        String dimName = mNameList.get(0);
        String timeName = null;
        if (timeNameList.size() != 0) {
            timeName = timeNameList.get(0);
        }
        String indexName = null;
        if (indexNameList.size() != 0) {
            indexName = indexNameList.get(0);
        }
        
        SGSXYMultipleDimensionNetCDFData data = new SGSXYMultipleDimensionNetCDFData(
                ncFile, obs,
                xName, yName, lName, uName, 
                tName, dimName,
                startIndex, endIndex, stepIndex,
                timeName, indexName);
        
        return data;
    }
    
    public static SGSXYMultipleVariableNetCDFData createVariableNetCDFData(
            final SGNetCDFFile ncFile,
            final SGDataSourceObserver obs,
            final SGDataColumnInfo[] colInfo) {
        
        List<String> xNameList = new ArrayList<String>();
        List<String> yNameList = new ArrayList<String>();
        Map<String, String> lNameMap = new HashMap<String, String>();
        Map<String, String> uNameMap = new HashMap<String, String>();
        Map<String, String> tNameMap = new HashMap<String, String>();
        List<String> timeNameList = new ArrayList<String>();
        List<String> serialNumberNameList = new ArrayList<String>();
        if (!SGDataUtility.getSXYNetCDFColumnType(colInfo, xNameList,
                yNameList, lNameMap, uNameMap, tNameMap, timeNameList, serialNumberNameList)) {
            return null;
        }
        if (lNameMap.size() != uNameMap.size()) {
            return null;
        }
        String[] xNames = xNameList.toArray(new String[xNameList.size()]);
        String[] yNames = yNameList.toArray(new String[yNameList.size()]);
        
        Iterator<String> eItr = lNameMap.keySet().iterator();
        List<String> lNameList = new ArrayList<String>();
        List<String> uNameList = new ArrayList<String>();
        List<String> ehNameList = new ArrayList<String>();
        while (eItr.hasNext()) {
            String eh = eItr.next();
            ehNameList.add(eh);
            String lName = lNameMap.get(eh);
            if (lName != null) {
                lNameList.add(lName);
            }
            String uName = uNameMap.get(eh);
            if (uName != null) {
                uNameList.add(uName);
            }
        }
        if (lNameList.size() != uNameList.size()) {
            return null;
        }
        if (lNameList.size() != ehNameList.size()) {
            return null;
        }
        String[] lNames = lNameList.toArray(new String[lNameList.size()]);
        String[] uNames = uNameList.toArray(new String[uNameList.size()]);
        String[] ehNames = ehNameList.toArray(new String[ehNameList.size()]);
        
        Iterator<String> tItr = tNameMap.keySet().iterator();
        List<String> tNameList = new ArrayList<String>();
        List<String> thNameList = new ArrayList<String>();
        while (tItr.hasNext()) {
            String th = tItr.next();
            thNameList.add(th);
            String tName = tNameMap.get(th);
            if (tName != null) {
                tNameList.add(tName);
            }
        }
        if (tNameList.size() != thNameList.size()) {
            return null;
        }
        String[] tNames = tNameList.toArray(new String[tNameList.size()]);
        String[] thNames = thNameList.toArray(new String[thNameList.size()]);
        
        String timeName = null;
        if (timeNameList.size() != 0) {
            timeName = timeNameList.get(0);
        }
        String serialNumberName = null;
        if (serialNumberNameList.size() != 0) {
            serialNumberName = serialNumberNameList.get(0);
        }
        
        SGSXYMultipleVariableNetCDFData data = new SGSXYMultipleVariableNetCDFData(
                ncFile, obs,
                xNames, yNames, lNames, uNames, ehNames, 
                tNames, thNames,
                timeName, serialNumberName);
        
        return data;
    }
    
    /**
     * 
     * @param nc
     * @param columns
     * @param xValueNames
     * @param yValueNames
     * @param leValueNames
     * @param ueValueNames
     * @param ehValueNames
     * @param tlValueNames
     * @param thValueNames
     */
    public static void updateColumnTypeOfSXYMultipleVariableNetCDFDataFromVariableNames(
            final SGNetCDFFile nc, final SGDataColumnInfo[] columns,
            final String[] xValueNames, final String[] yValueNames,
            final String[] leValueNames, final String[] ueValueNames, final String[] ehValueNames,
            final String[] tlValueNames, final String[] thValueNames) {
        for (int i = 0; i < xValueNames.length; i++) {
            int xVarIndex = nc.getVariableIndex(xValueNames[i]);
            columns[xVarIndex].setColumnType(SGIDataColumnTypeConstants.X_VALUE);
        }
        for (int i = 0; i < yValueNames.length; i++) {
            int yVarIndex = nc.getVariableIndex(yValueNames[i]);
            columns[yVarIndex].setColumnType(SGIDataColumnTypeConstants.Y_VALUE);
        }

        boolean errorbarFlag = false;
        boolean ticklabelFlag = false;
        if (leValueNames!=null && leValueNames.length>0 && 
                ueValueNames!=null && ueValueNames.length>0 &&
                ehValueNames!=null && ehValueNames.length>0 &&
                leValueNames.length==ueValueNames.length &&
                leValueNames.length==ehValueNames.length) {
            errorbarFlag = true;
        }
        if (tlValueNames!=null && tlValueNames.length>0 &&
                thValueNames!=null && thValueNames.length>0 &&
                tlValueNames.length==thValueNames.length) {
            ticklabelFlag = true;
        }
        if (errorbarFlag) {
            for (int i = 0; i < leValueNames.length; i++) {
                int leVarIndex = nc.getVariableIndex(leValueNames[i]);
                int ueVarIndex = nc.getVariableIndex(ueValueNames[i]);
                if (leVarIndex!=ueVarIndex) {
                    columns[leVarIndex].setColumnType(SGIDataColumnTypeConstants.LOWER_ERROR_VALUE+" for "+ehValueNames[i]);
                    columns[ueVarIndex].setColumnType(SGIDataColumnTypeConstants.UPPER_ERROR_VALUE+" for "+ehValueNames[i]);
                } else {
                    columns[leVarIndex].setColumnType(SGIDataColumnTypeConstants.LOWER_UPPER_ERROR_VALUE+" for "+ehValueNames[i]);
                }
            }
        }
        if (ticklabelFlag) {
            for (int i = 0; i < tlValueNames.length; i++) {
                int tlVarIndex = nc.getVariableIndex(tlValueNames[i]);
                columns[tlVarIndex].setColumnType(SGIDataColumnTypeConstants.TICK_LABEL+" for "+thValueNames[i]);
            }
        }
    }
    
    /**
     * Return true if name1 and name2 belongs to a same netCDF group.
     * @param name1
     * @param name2
     * @return
     */
    public static boolean isSameNetCDFGroup(final String name1, final String name2) {
        String[] str1 = name1.split("/");
        String[] str2 = name2.split("/");
        if (str1.length==str2.length) {
            for (int i = 0; i < str1.length-1; i++) {
                if (str1[i].equals(str2[i])==false) {
                    return false;
                }
            }
            return true;
        }
        return false;
    }
    
    /**
     * 
     * @param nCols
     * @return
     */
    private static SGVariable getDimensionVariable(final SGNetCDFDataColumnInfo[] nCols) {
        // search dimension
        SGVariable var = null;
        
        for (int i = 0; i < nCols.length; i++) {
            String colType = nCols[i].getColumnType();
            if (PICKUP.equals(colType)) {
                var = nCols[i].getVariable();
                break;
            }
        }
        
        if (var != null && var.isCoordinateVariable()) {
            return var;
        } else {
            return null;
        }
    }
    
    /**
     * Update pick start, end, step parameters for multiple dimension netCDF data.
     * <p>
     * If infoMap does not have these values, give default.
     * 
     * @param infoMap
     * @param nCols
     * @return true if succeeds. false if a dimension variable is found but
     * it is not a coordinate variable.
     */
    public static boolean updatePickupParameters(
            final Map<String, Object> infoMap, SGNetCDFDataColumnInfo[] nCols) {
        Integer start = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_START);
        if (start == null) {
            // set zero by default
            start = Integer.valueOf(0);
        }
        infoMap.put(KEY_MULTIPLE_DIMENSION_START, start);
        
        Integer end = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_END);
        SGVariable var = getDimensionVariable(nCols);
        if (var != null) {
            Dimension dim = var.getDimension(0);
            final int len = dim.getLength();
            if (end != null) {
                if (end.intValue()>=len) {
                    end = Integer.valueOf(len-1);
                }
            } else {
                end = Integer.valueOf(len-1);
            }
        } else {
            if (end == null) {
                end = Integer.valueOf(0);
            }
        }
        infoMap.put(KEY_MULTIPLE_DIMENSION_END, end);

        Integer step = (Integer) infoMap.get(KEY_MULTIPLE_DIMENSION_STEP);
        if (step == null) {
            if (var != null) {
                if (!var.isCoordinateVariable()) {
                    return false;
                }
                Dimension dim = var.getDimension(0);
                final int len = dim.getLength();
                step = Integer.valueOf(len / DEFAULT_MULTIPLE_DIMENSION_NUM);
            } else {
                // set zero by default
                step = Integer.valueOf(0);
            }
        }
        infoMap.put(KEY_MULTIPLE_DIMENSION_STEP, step);
        
        return true;
    }
    
    /**
     * Return whether given netCDF variable is SGDateVariable.
     * <p>
     * True if data type of the variable is CHAR and it has an attribute "value_type=Date".
     * 
     * @param var
     * @return true if a given variable is SGDateVariable.
     */
    public static boolean isSGDateVariable(final Variable var) {
        if (DataType.CHAR.equals(var.getDataType())) {
            List<Attribute> attrList = var.getAttributes();
            for (Attribute attr : attrList) {
                if (attr.isString()) {
                    String name = attr.getName();
                    String value = attr.getStringValue();
                    if (ATTRIBUTE_KEY_VALUE_TYPE.equals(name.trim()) &&
                        SGIDataColumnTypeConstants.VALUE_TYPE_DATE.equals(value.trim())) {
                            return true;
                    }
                }
            }
        }
        return false;
    }
    
    /**
     * Return whether given netCDF variable is SGTextVariable.
     * <p>
     * True if data type of the variable is CHAR and it has an attribute "value_type=Text".
     * 
     * @param var
     * @return true if var is SGTextVariable.
     */
    public static boolean isSGTextVariable(final Variable var) {
        if (DataType.CHAR.equals(var.getDataType())) {
            List<Attribute> attrList = var.getAttributes();
            for (Attribute attr : attrList) {
                if (attr.isString()) {
                    String name = attr.getName();
                    String value = attr.getStringValue();
                    if (name!=null && value!=null &&
                            ATTRIBUTE_KEY_VALUE_TYPE.equals(name.trim()) &&
                            SGIDataColumnTypeConstants.VALUE_TYPE_TEXT.equals(value.trim())) {
                            return true;
                    }
                }
            }
        }
        return false;
    }
    
    /**
     * Return netCDF attribute for given valueType.
     * 
     * @param valueType
     * @return
     */
    public static Attribute getValueTypeAttribute(final String valueType) {
        Attribute attr = new Attribute(ATTRIBUTE_KEY_VALUE_TYPE, valueType);
        return attr;
    }
    
    /**
     * Return whether given netCDF data can connect with its file.
     * 
     * @param data
     * @return
     */
    public static boolean canOpenNetCDF(SGNetCDFData data) {
        try {
            boolean isNetCDFFile = NetcdfFile.canOpen(data.getNetcdfFile().getNetcdfFile().getLocation());
            return isNetCDFFile;
        } catch (Exception e) {
            return false;
        }
    }
    
    /**
     * Dump map for debug.
     * 
     * @param infoMap
     */
    public static void dumpInfoMap(final Map<String, Object> infoMap) {
        for (String key : infoMap.keySet()) {
            if (null==infoMap.get(key)) {
                System.out.println("   key="+key+" : null");
            } else {
                if (key.equals(SGIDataColumnTypeConstants.KEY_DATA_SOURCE)) {
                    System.out.println("   key="+key+" : "+infoMap.get(key).getClass().getSimpleName());
                } else {
                    System.out.println("   key="+key+" : "+infoMap.get(key).toString());
                }
            }
        }
    }

    /**
     * Returns the values of date objects.
     * 
     * @param dArray
     *            an array of date objects
     * @return the values of date objects
     */
    public static double[] getDateValueArray(SGDate[] dArray) {
        final int num = dArray.length;
        final double[] values = new double[num];
        final Calendar cal = Calendar.getInstance();
        final int ratio = 1000 * 3600 * 24;
        for (int ii = 0; ii < num; ii++) {
            final SGDate d = dArray[ii];
            cal.set(d.year, d.month, d.date);
            values[ii] = cal.getTimeInMillis() / ratio;
        }
        return values;
    }

	public static SGNetCDFDataColumnInfo[] getNetCDFDataColumnInfo(
			List<SGVariable> varList, Map<String, Object> infoMap) {
	
		// get the origin map
		NamedNodeMap nodeMap = (NamedNodeMap) infoMap
				.get(SGIFigureElementGraph.KEY_NODE_MAP);
		Map<String, Integer> originMap = null;
		if (nodeMap != null) {
			originMap = SGDefaultColumnTypeUtility.getOriginMap(nodeMap);
		}

		// get the group name
		String groupName = (String) infoMap.get(SGIFigureElementGraph.KEY_GROUP_NAME);
		if (groupName != null) {
			// replaces the origin map
			Map<String, Integer> tempMap = new HashMap<String, Integer>();
			Set<Entry<String, Integer>> entrySet = originMap.entrySet();
			for (Entry<String, Integer> e : entrySet) {
				String varName = e.getKey();
				Integer value = e.getValue();
				String key = SGDataUtility.appendGroupName(varName, groupName);
				tempMap.put(key, value);
			}
			originMap = tempMap;
		}
		
		// get column information: title and value type
		SGNetCDFDataColumnInfo[] colArray = new SGNetCDFDataColumnInfo[varList.size()];
		for (int ii = 0; ii < colArray.length; ii++) {
			SGVariable var = varList.get(ii);
	
			// get the origin
			int origin = 0;
			if (originMap != null) {
				if (var.isCoordinateVariable()) {
					String varName = var.getName();
					Integer num = originMap.get(varName);
					if (num != null) {
						Dimension dim = var.getDimension(0);
						origin = num.intValue();
						if (origin < 0 || origin >= dim.getLength()) {
							origin = 0;
						}
					}
				}
			}
	
			colArray[ii] = new SGNetCDFDataColumnInfo(var,
					var.getNameInPriorityOrder(), var.getValueType(), origin);
		}
	
		return colArray;
	}

	static String appendGroupName(final String name, final String groupName) {
		StringBuffer sb = new StringBuffer();
		sb.append(groupName);
		sb.append('/');
		sb.append(name);
		return sb.toString();
	}
}
