package jp.riken.brain.ni.samuraigraph.data;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGIDataSource;
import ucar.ma2.Array;
import ucar.ma2.ArrayChar;
import ucar.ma2.DataType;
import ucar.ma2.Index;
import ucar.nc2.Dimension;
import ucar.nc2.NetcdfFile;
import ucar.nc2.Variable;

/**
 * The wrapper class for netCDF file.
 * @author kuromaru
 *
 */
public class SGNetCDFFile implements SGIDataSource, SGINetCDFConstants {

	/**
	 * The netCDF file.
	 * 
	 */
	private NetcdfFile mNetcdfFile = null;
	
	/**
	 * The list of variables.
	 * 
	 */
	private List<SGVariable> mVariableList = new ArrayList<SGVariable>();

	/**
	 * Builds this object with given netCDF file.
	 * 
	 * @param ncfile
	 *           a netCDF file
	 */
	public SGNetCDFFile(NetcdfFile ncfile) {
		super();
		if (ncfile == null) {
			throw new IllegalArgumentException("ncfile == null");
		}
		this.mNetcdfFile = ncfile;
		
        List<Variable> varList = ncfile.getVariables();
		
		// get dimension which is used as max length of string.
		List<Dimension> textLengthDimension = new ArrayList<Dimension>();
		for (Variable var : varList) {
		    if (SGDataUtility.isSGTextVariable(var) || SGDataUtility.isSGDateVariable(var)) {
		        textLengthDimension.add(var.getDimension(1));
		    }
		}
		
		// get dimension which is used as the length of byte data
		List<Dimension> byteLengthDimension = new ArrayList<Dimension>();
		for (Variable var : varList) {
		    if (DataType.BYTE.equals(var.getDataType())) {
		    	byteLengthDimension.add(var.getDimension(0));
		    }
		}

		// add coordinate variables
		for (Variable var : varList) {
			if (var.isCoordinateVariable()) {
                this.mVariableList.add(new SGVariable(var));
			}
		}
		
		// add coordinate variables that do not exist in the file
		List<Dimension> dimList = ncfile.getDimensions();
		for (Dimension dim : dimList) {
			boolean exists = false;
			for (Variable var : varList) {
				if (var.isCoordinateVariable()) {
					Dimension varDim = var.getDimension(0);
					if (varDim.equals(dim)) {
						// the coordinate variable exists
						exists = true;
						break;
					}
				}
			}
			if (!exists) {
			    if (textLengthDimension.contains(dim)) {
			        exists = true;
			    }
			}
			if (!exists) {
			    if (byteLengthDimension.contains(dim)) {
			        exists = true;
			    }
			}
			if (!exists) {
				this.mVariableList.add(new SGCoordinateVariable(dim));
			}
		}
		
		// add remaining variables
		for (Variable var : varList) {
			if (!var.isCoordinateVariable()) {
			    if (DataType.CHAR.equals(var.getDataType())) {
			    	// text or date
			        List<Dimension> dimensionList = var.getDimensions();
			        if (null != dimensionList && dimensionList.size() == 2) {
	                    int[] shape = var.getShape();
	                    String[] str = new String[shape[0]];
	                    try {
	                        Array array = var.read();
	                        if (array instanceof ArrayChar) {
	                        	ArrayChar cArray = (ArrayChar)array;
	                        	Index index = cArray.getIndex();
	                            for (int i = 0; i < shape[0]; i++) {
	                            	byte[] byteArray = new byte[shape[1]];
	                            	for (int j = 0; j < shape[1]; j++) {
	                            		byteArray[j] = cArray.getByte(index.set(i, j));
	                            	}
	                            	str[i] = new String(byteArray, "UTF-8").trim();
	                            }
	                        }
	                    } catch (IOException e) {
	                        continue;
	                    }

	                    SGVariable textVar = null;
	                    if (SGDataUtility.isSGDateVariable(var)) {
		                    textVar = new SGDateVariable(var, str);
	                    } else {
		                    textVar = new SGTextVariable(var, str);
	                    }
	                    this.mVariableList.add(textVar);
			        }
			    } else if (DataType.BYTE.equals(var.getDataType())) {
			    	// byte data
                    this.mVariableList.add(new SGByteDataVariable(var));
			    } else {
			    	// number
			        this.mVariableList.add(new SGVariable(var));
			    }
			}
		}
	}
	
	/**
	 * Returns the netCDF file.
	 * 
	 * @return the netCDF file
	 */
	public NetcdfFile getNetcdfFile() {
		return this.mNetcdfFile;
	}

	/**
	 * Returns a list of variables.
	 * 
	 * @return a list of variables
	 */
	public List<SGVariable> getVariables() {
		return new ArrayList<SGVariable>(this.mVariableList);
	}
	
	/**
	 * Finds and returns a variable of a given name if it exists.
	 * 
	 * @param name
	 *           the name of variable
	 * @return found variable
	 */
	public SGVariable findVariable(String name) {
		for (SGVariable var : this.mVariableList) {
			if (var.getName().equals(name)) {
				return var;
			}
		}
		return null;
	}

	/**
	 * Finds and returns a variable of a given name if it exists.
	 * 
	 * @param name
	 *           the name of variable
	 * @return found variable
	 */
	public SGVariable findVariableEx(final String name) {
		for (int ii = 0; ii < this.mVariableList.size(); ii++) {
			SGVariable var = this.mVariableList.get(ii);
			String vName;
			if ((vName = var.getLongName()) != null) {
				if (name.equals(vName)) {
					return var;
				}
			}
			if ((vName = var.getStandardName()) != null) {
				if (name.equals(vName)) {
					return var;
				}
			}
			if (name.equals(var.getName())) {
				return var;
			}
		}
		return null;
	}
	
	/**
	 * Finds a variable of a given name if it exists and returns its array index.
	 * 
	 * @param name
	 *           the name of variable
	 * @return the array index of found variable or -1 if it is not found
	 */
	public int getVariableIndex(final String name) {
		for (int ii = 0; ii < this.mVariableList.size(); ii++) {
			SGVariable var = this.mVariableList.get(ii);
			if (name.equals(var.getName())) {
				return ii;
			}
			if (var instanceof SGTextVariable) {
			    if (name.equals(var.getOriginalName())) {
			        return ii;
			    }
			}
		}
		return -1;
	}
	
	public SGVariable findTopVariable(String name) {
		Variable var = this.mNetcdfFile.findTopVariable(name);
		if (var != null) {
			return new SGVariable(var);
		} else {
			return null;
		}
	}
	
	public List<Dimension> getDimensions() {
		return this.mNetcdfFile.getDimensions();
	}
	
	public Dimension findDimension(String name) {
		return this.mNetcdfFile.findDimension(name);
	}

	public String toString() {
		if (this.mNetcdfFile == null) {
			return "null";
		} else {
			return this.mNetcdfFile.toString();
		}
	}

	/**
	 * Finds and returns a text variable.
	 * 
	 * @param dimName
	 *            the name of dimension
	 * @return a text variable for a given dimension
	 */
	public SGTextVariable findTextVariable(String dimName) {
		SGTextVariable ret = null;
		for (int ii = 0; ii < this.mVariableList.size(); ii++) {
			SGVariable var = this.mVariableList.get(ii);
			if (var instanceof SGTextVariable) {
				SGTextVariable tVar = (SGTextVariable) var;
				if (tVar.getName().endsWith(dimName)) {
					ret = tVar;
					break;
				}
			}
		}
		return ret;
	}

    /**
     * Clones this data file.
     * 
     * @return copy of this data file
     */
    public Object clone() {
    	SGNetCDFFile dataFile = null;
        try {
            dataFile = (SGNetCDFFile) super.clone();
        } catch (CloneNotSupportedException ex) {
            // this shouldn't happen, since we are Cloneable
            throw new InternalError();
        }
        dataFile.mVariableList = new ArrayList<SGVariable>();
        dataFile.mVariableList.addAll(this.mVariableList);
        return dataFile;
    }

	/**
	 * Disposes of this object.
	 */
	public void dispose() {
		this.mDisposed = true;
		this.mVariableList.clear();
        this.mVariableCache.clear();
        try {
            this.mNetcdfFile.close();
        } catch (IOException ioe) {
        }
        this.mNetcdfFile = null;
        this.mVariableList = null;
        this.mVariableCache = null;
	}

    // The flag whether this object is already disposed of.
    private boolean mDisposed = false;

    /**
     * Returns whether this object is already disposed of.
     * 
     * @return true if this object is already disposed of
     */
    public boolean isDisposed() {
    	return this.mDisposed;
    }
	
	/**
	 * The cache for variables.
	 */
	protected Map<String, Object> mVariableCache = new HashMap<String, Object>();

	/**
	 * Returns the cache of a variable of given name.
	 * 
	 * @param varName
	 *           the name of a variable
	 * 
	 * @return the cache of a variable of given name
	 */
	public Object getVariableCache(String varName) {
		return this.mVariableCache.get(varName);
	}

	/**
	 * Sets the cache of a variable of given name.
	 * 
	 * @param varName
	 *           the name of a variable
	 * @param cache
	 *           the cache of a variable of given name
	 */
	public void setVariableCache(String varName, Object cache) {
		this.mVariableCache.put(varName, cache);
	}
	
}
