package jp.riken.brain.ni.samuraigraph.data;

import java.io.IOException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;

import org.w3c.dom.Element;

import ucar.nc2.Dimension;

/**
 * Scalar type XY data with multiple indices of specified dimension.
 * @author kuromaru
 *
 */
public class SGSXYMultipleDimensionNetCDFData extends SGSXYMultipleNetCDFData {

    /**
     * The variable for x-values.
     */
    protected SGVariable mXVariable = null;

    /**
     * The variable for y-values.
     */
    protected SGVariable mYVariable = null;

    /**
     * The variable for lower error values.
     */
    protected SGVariable mLowerErrorVariable = null;

    /**
     * The variable for upper error values.
     */
    protected SGVariable mUpperErrorVariable = null;

    /**
     * The variable for tick labels.
     */
    protected SGVariable mTickLabelVariable = null;

    /**
     * The dimension.
     */
    protected Dimension mDimension = null;

    /**
     * Indices of the dimension.
     */
    protected Integer[] mDimensionIndices = null;

    /**
     * The default constructor.
     *
     */
    public SGSXYMultipleDimensionNetCDFData() {
        super();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xName
     *            the name of x variable
     * @param yName
     *            the name of y variable
     * @param lName
     *            the name of lower error variable
     * @param uName
     *            the name of upper error variable
     * @param tName
     *            the name of tick label variable
     * @param dimName
     *            the name of dimension
     * @param dimStart
     *            start index for given dimension
     * @param dimEnd
     *            end index for given dimension
     * @param dimStep
     *            step of index for given dimension
     * @param timeName
     *            the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGSXYMultipleDimensionNetCDFData(final SGNetCDFFile ncfile,
            final SGDataSourceObserver obs, final String xName,
            final String yName, final String lName, final String uName,
            final String tName, final String dimName,
            final int dimStart, final int dimEnd, final int dimStep,
            final String timeName, final String serialNumberName) {

        super(ncfile, obs, timeName, serialNumberName);

        if (xName == null || yName == null || dimName == null) {
            throw new IllegalArgumentException(
            		"xName == null || yName == null || dimName == null");
        }
        if (!(lName == null && uName == null) 
        		&& !(lName != null && uName != null)) {
            throw new IllegalArgumentException("lName = " + lName + ", uName = " + uName);
        }

        // check the dimension
        Dimension dim = ncfile.findDimension(dimName);
        if (dim == null) {
            throw new IllegalArgumentException("Dimension is not found: " + dimName);
        }

        // get dimension indices
        final int dimLen = dim.getLength();
        if (dimStart < 0 || dimStart >= dimLen) {
            throw new IllegalArgumentException("Dimension start is out of bounds: " + dimStart);
        }
        if (dimEnd < 0 || dimEnd >= dimLen) {
            throw new IllegalArgumentException("Dimension end is out of bounds: " + dimEnd);
        }

        // get variables
        SGVariable xVar = this.getVariable(ncfile, xName);
        SGVariable yVar = this.getVariable(ncfile, yName);
        SGVariable lVar = this.getVariable(ncfile, lName);
        SGVariable uVar = this.getVariable(ncfile, uName);
        SGVariable tVar = this.getVariable(ncfile, tName);

        //
        // check variables
        //
        List<Dimension> cDimList = new ArrayList<Dimension>();
        cDimList.add(dim);
        
        // time variable
    	if (this.mTimeVariable != null) {
    		Dimension tDim = this.mTimeVariable.getDimension(0);
            if (tDim.equals(dim)) {
                throw new IllegalArgumentException(
        			"The dimension of a coordinate variable is equal to the given dimension.");
            }
    		cDimList.add(tDim);
    	}
    	
        if (this.mSerialNumberVariable != null) {
            if (!xVar.isCoordinateVariable() || !yVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Both of x and y variables must not be coordinate variables.");
            }
        	
        	// index variable
            Dimension iDim = this.mSerialNumberVariable.getDimension(0);
            cDimList.add(iDim);
            
            this.checkNotCoordinateVariable(xVar, cDimList);
            this.checkNotCoordinateVariable(yVar, cDimList);
        } else {
        	// x and y variables
            if ((xVar.isCoordinateVariable() && yVar.isCoordinateVariable())
            		|| (!xVar.isCoordinateVariable() && !yVar.isCoordinateVariable())) {
                throw new IllegalArgumentException(
                		"Either of x and y variables must be a coordinate variable.");
            }
            SGVariable cVar = null;
            SGVariable oVar = null;
            if (xVar.isCoordinateVariable()) {
            	cVar = xVar;
            	oVar = yVar;
            } else {
            	cVar = yVar;
            	oVar = xVar;
            }
            Dimension cDim = cVar.getDimension(0);
            if (cDim.equals(dim)) {
                throw new IllegalArgumentException(
        			"The dimension of a coordinate variable is equal to the given dimension.");
            }
            cDimList.add(cDim);
            
            // time variable
        	if (this.mTimeVariable != null) {
        		Dimension tDim = this.mTimeVariable.getDimension(0);
                if (tDim.equals(cDim)) {
                    throw new IllegalArgumentException(
            			"The dimension of a coordinate variable is equal to the dimension for time.");
                }
        	}
        	
            this.checkNotCoordinateVariable(oVar, cDimList);
        }
    	this.checkNotCoordinateVariable(lVar, cDimList);
    	this.checkNotCoordinateVariable(uVar, cDimList);
    	this.checkNotCoordinateVariable(tVar, cDimList);

        // set to attributes
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mLowerErrorVariable = lVar;
        this.mUpperErrorVariable = uVar;
        this.mTickLabelVariable = tVar;
        this.mDimension = dim;
        this.mDimensionStart = dimStart;
        this.mDimensionEnd = dimEnd;
        this.mDimensionStep = dimStep;
        
        this.updateDimensionIndices();
    }

    @Override
    public boolean setDimensionStart(final int start) {
        final int dimLen = this.mDimension.getLength();
        if (start < 0 || start >= dimLen) {
            throw new IllegalArgumentException("Dimension start is out of bounds: " + start);
        }
        if (super.setDimensionStart(start)==false) {
            return false;
        }
        this.updateDimensionIndices();
        return true;
    }

    @Override
    public boolean setDimensionEnd(final int end) {
        final int dimLen = this.mDimension.getLength();
        if (end < 0 || end >= dimLen) {
            throw new IllegalArgumentException("Dimension end is out of bounds: " + end);
        }
        if (super.setDimensionEnd(end)==false) {
            return false;
        }
        this.updateDimensionIndices();
        return true;
    }

    @Override
    public boolean setDimensionStep(final int step) {
        if (super.setDimensionStep(step)==false) {
            return false;
        }
        this.updateDimensionIndices();
        return true;
    }
    
    /**
     * Returns name of the dimension.
     * 
     * @return name of the dimension
     */
    public String getDimensionName() {
        if (null==this.mDimension) {
            return "";
        }
        return this.mDimension.getName();
    }

    /**
     * Returns an index array of dimension.
     * 
     * @return an index array of dimension
     */
    public Integer[] getDimensionIndices() {
        return this.mDimensionIndices.clone();
    }

    // Updates dimension indices.
    private void updateDimensionIndices() {
        this.mDimensionIndices = SGDataUtility.calcDimIndices(
                this.mDimension, 
                this.mDimensionStart, this.mDimensionEnd, this.mDimensionStep);
    }

    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
    	
        this.mXVariable = null;
        this.mYVariable = null;
        this.mLowerErrorVariable = null;
        this.mUpperErrorVariable = null;
        this.mTickLabelVariable = null;
        this.mDimension = null;
        this.mDimensionStart = null;
        this.mDimensionEnd = null;
        this.mDimensionStep = null;
        this.mDimensionIndices = null;
        
        // clear the cache
        this.mVariableCache.clear();
    }

    /**
     * Returns whether this multiple data has multiple arrays for y-values.
     * 
     * @return true if this multiple data has multiple arrays for y-values
     */
	protected boolean hasMultipleYValues() {
    	// if the x variable is a coordinate variable,
    	// this data has multiple y values
    	return this.mXVariable.isCoordinateVariable();
	}

    /**
     * Returns a variable for multiple values.
     * 
     * @return a variable for multiple values
     */
	public SGVariable getMultipleVariable() {
		if (this.hasMultipleYValues()) {
			return this.mYVariable;
		} else {
			return this.mXVariable;
		}
	}

    /**
     * Returns a variable for "not" multiple values.
     * 
     * @return a variable for "not" multiple values
     */
	public SGVariable getSingleVariable() {
		if (this.hasMultipleYValues()) {
			return this.mXVariable;
		} else {
			return this.mYVariable;
		}
	}
	
    public SGVariable getXVariable() {
        return this.mXVariable;
    }

    public SGVariable getYVariable() {
        return this.mYVariable;
    }
    
    public SGVariable getLowerErrorVariable() {
        return this.mLowerErrorVariable;
    }

    public SGVariable getUpperErrorVariable() {
        return this.mUpperErrorVariable;
    }
    
    public SGVariable getTickLabelVariable() {
        return this.mTickLabelVariable;
    }

    /**
     * Returns a text string of data type.
     * 
     * @return a text string of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.SXY_MULTIPLE_NETCDF_DIMENSION_DATA;
    }

    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected String getDataTypeForDataSet() {
        return SGDataTypeConstants.SXY_MULTIPLE_DATA;
    }

    /**
     * Returns an array of all variables.
     * 
     * @return an array of all variables
     */
	public SGVariable[] getVariables() {
        List<SGVariable> varList = new ArrayList<SGVariable>();
        varList.add(this.mXVariable);
        varList.add(this.mYVariable);
        if (this.mLowerErrorVariable != null) {
        	varList.add(this.mLowerErrorVariable);
        }
        if (this.mUpperErrorVariable != null) {
        	varList.add(this.mUpperErrorVariable);
        }
        if (this.mTickLabelVariable != null) {
        	varList.add(this.mTickLabelVariable);
        }
        if (this.mTimeVariable != null) {
        	varList.add(this.mTimeVariable);
        }
        if (this.mSerialNumberVariable != null) {
        	varList.add(this.mSerialNumberVariable);
        }
        SGVariable[] varArray = new SGVariable[varList.size()];
        return (SGVariable[]) varList.toArray(varArray);
	}

    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
	protected double[][] getDataSetValues() {
//      TODO fix after the specification of data set is determined 
		return null;
	}

    /**
	 * Returns the m-th x value at the n-th index.
	 * 
	 * @param m
	 *            the index of data
	 * @param n
	 *            the index of the data point
	 * @return the n-th x-value of the m-th data
	 */
	public Double getXValue(final int m, final int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (m < 0 || m >= this.mDimension.getLength()) {
            throw new IllegalArgumentException("given index is out of bounds: " + m);
        }
        return super.getXValue(m, n);
    }

    /**
     * Returns the m-th y value at the n-th index.
     * 
     * @param m
     *          the index of data
     * @param n
     *          the index of the data point
     * @return
     *          the n-th y-value of the m-th data
     */
    public Double getYValue(final int m, final int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (m < 0 || m >= this.mDimension.getLength()) {
            throw new IllegalArgumentException("given index is out of bounds: " + m);
        }
        return super.getYValue(m, n);
    }

    /**
     * Returns the maximum number of SXY data objects that can be obtained from
     * this data object.
     * 
     * @return the maximum number of SXY data objects that can be obtained from
     *         this data object
     */
    @Override
    public int getSXYDataNumMax() {
        return this.mDimension.getLength();
    }
    
    /**
     * Returns the value of coordinate variable at a given index which is specified
     *  as dimension.
     * @param index
     *         the array index
     * @return the value of coordinate variable at a given index.
     * null if file reading failure.
     * @throws IllegalArgumentException
     *  if the dimension variable is not a coordinate variable or index is illegal.
     */
    public Double getDimensionValue(final int index) {
        try {
            double value = this.getCoordinateValue(this.mDimension.getName(), index);
            return Double.valueOf(value);
        } catch (IOException e) {
            return null;
        }
    }

    /**
     * Returns whether error bars are available.
     * 
     * @return true if error bars are available
     */
    public boolean isErrorBarAvailable() {
		return (this.mLowerErrorVariable != null && this.mUpperErrorVariable != null);
    }

    /**
     * Returns whether tick labels are available.
     * 
     * @return true if tick labels are available
     */
    public boolean isTickLabelAvailable() {
		return (this.mTickLabelVariable != null);
    }

    /**
	 * Returns whether the error bars are vertical. If this data does not have
	 * error values, returns null.
	 * 
	 * @return true if error bars are vertical, false if they are horizontal and
	 *         null if this data do not have error values
	 */
	public Boolean isErrorBarVertical() {
    	if (this.isErrorBarAvailable()) {
    		return this.hasMultipleYValues();
    	} else {
    		return null;
    	}
	}

    /**
	 * Returns whether the tick labels align horizontally. If this data does not have
	 * tick labels, returns null.
	 * 
	 * @return true if tick labels align horizontally, false if they do not so and
	 *         null if this data do not have tick labels
	 */
	public Boolean isTickLabelHorizontal() {
    	if (this.isTickLabelAvailable()) {
    		return this.hasMultipleYValues();
    	} else {
    		return null;
    	}
	}

    /**
     * Returns the title for the X-axis.
     * 
     * @return the title for the X-axis
     */
    public String getTitleX() {
    	if (this.hasMultipleYValues()) {
            return this.getNameWithUnit(this.mXVariable);
    	} else {
    	    return "";
    	}
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleY() {
    	if (this.hasMultipleYValues()) {
    	    return "";
    	} else {
            return this.getNameWithUnit(this.mYVariable);
    	}
    }

    /**
     * Returns an array of current column types.
     * 
     * @return an array of current column types
     */
    public String[] getCurrentColumnType() {
        List<SGVariable> varList = this.mNCFile.getVariables();
        final int varNum = varList.size();
        String[] array = new String[varNum];
        for (int ii = 0; ii < varNum; ii++) {
            SGVariable var = varList.get(ii);
            if (var.equals(this.mXVariable)) {
                array[ii] = X_VALUE;
            } else if (var.equals(this.mYVariable)) {
                array[ii] = Y_VALUE;
            } else if (var.equals(this.mLowerErrorVariable) 
            		|| var.equals(this.mUpperErrorVariable)) {
            	String name = null;
                if (this.mLowerErrorVariable.equals(this.mUpperErrorVariable)) {
                    name = LOWER_UPPER_ERROR_VALUE;
                } else {
                    if (var.equals(this.mLowerErrorVariable)) {
                        name = LOWER_ERROR_VALUE;
                    } else if (var.equals(this.mUpperErrorVariable)) {
                        name = UPPER_ERROR_VALUE;
                    }
                }
                array[ii] = name;
            } else if (var.equals(this.mTickLabelVariable)) {
                array[ii] = TICK_LABEL;
			} else if (var.equals(this.mTimeVariable)) {
				array[ii] = TIME;
			} else if (var.equals(this.mSerialNumberVariable)) {
				array[ii] = SERIAL_NUMBER;
			} else {
				if (var.isCoordinateVariable()) {
					Dimension dim = var.getDimension(0);
					if (this.mDimension.equals(dim)) {
						array[ii] = PICKUP;
					} else {
						array [ii] = "";
					}
				} else {
					array[ii] = "";
				}
			}
		}        
        return array;
    }

    /**
     * Sets the type of data columns.
     * 
     * @param column
     *            an array of column types
     * @return true if succeeded
     */
    public boolean setColumnType(String[] columns) {
    	SGVariable xVar = null;
    	SGVariable yVar = null;
    	SGVariable lVar = null;
    	SGVariable uVar = null;
    	SGVariable tVar = null;
    	SGVariable timeVar = null;
    	SGVariable dVar = null;
    	SGVariable serialNumberVar = null;
        
        // set variables
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
            SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                xVar = var;
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                yVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], LOWER_ERROR_VALUE)) {
                lVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], UPPER_ERROR_VALUE)) {
                uVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], LOWER_UPPER_ERROR_VALUE)) {
                lVar = var;
                uVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], TICK_LABEL)) {
                tVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], PICKUP)) {
                dVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], TIME)) {
                timeVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], SERIAL_NUMBER)) {
                serialNumberVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }
        if (xVar == null || yVar == null || dVar == null) {
        	return false;
        }
        if (!((lVar != null) && (uVar != null)) && !((lVar == null) && (uVar == null))) {
        	return false;
        }
        
        // set variables
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mLowerErrorVariable = lVar;
        this.mUpperErrorVariable = uVar;
        this.mTickLabelVariable = tVar;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;
        this.mDimension = dVar.getDimension(0);
        
        return true;
    }

    /**
     * Returns the properties of this data.
     * 
     * @return the properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYMultipleDimensionNetCDFDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Gets the properties of this data.
     * 
     * @param p
     *      the properties of this data
     * @return true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof SXYMultipleDimensionNetCDFDataProperties)) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYMultipleDimensionNetCDFDataProperties sp = (SXYMultipleDimensionNetCDFDataProperties) p;
        sp.xName = this.getName(this.mXVariable);
        sp.yName = this.getName(this.mYVariable);
        sp.lName = this.getName(this.mLowerErrorVariable);
        sp.uName = this.getName(this.mUpperErrorVariable);
        sp.tName = this.getName(this.mTickLabelVariable);
        sp.dimName = this.getName(this.mDimension);
        sp.dimStart = this.mDimensionStart;
        sp.dimEnd = this.mDimensionEnd;
        sp.dimStep = this.mDimensionStep;
        return true;
    }

    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof SXYMultipleDimensionNetCDFDataProperties)) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYMultipleDimensionNetCDFDataProperties sp = (SXYMultipleDimensionNetCDFDataProperties) p;
        this.mXVariable = this.findVariable(sp.xName);
        this.mYVariable = this.findVariable(sp.yName);
        this.mLowerErrorVariable = this.findVariable(sp.lName);
        this.mUpperErrorVariable = this.findVariable(sp.uName);
        this.mTickLabelVariable = this.findVariable(sp.tName);
        this.mDimension = this.findDimension(sp.dimName);
        this.mDimensionStart = sp.dimStart;
        this.mDimensionEnd = sp.dimEnd;
        this.mDimensionStep = sp.dimStep;
        
        this.updateDimensionIndices();

        return true;
    }

    /**
     * Read properties.
     * 
     * @param el
     *           an Element object
     * @return true if succeeded
     */
    public boolean readProperty(Element el) {
        
        if (super.readProperty(el) == false) {
            return false;
        }
        
        String str = null;
        SGVariable xVar = null;
        SGVariable yVar = null;
        SGVariable lVar = null;
        SGVariable uVar = null;
        SGVariable tVar = null;
        Dimension dim = null;
        Integer dimStart = null;
        Integer dimEnd = null;
        Integer dimStep = null;
        
        // x variable
        str = el.getAttribute(KEY_X_VALUE_NAME);
        if (str.length() != 0) {
            xVar = this.mNCFile.findVariable(str);
            if (xVar == null) {
                return false;
            }
        } else {
            return false;
        }
        
        // y variable
        str = el.getAttribute(KEY_Y_VALUE_NAME);
        if (str.length() != 0) {
            yVar = this.mNCFile.findVariable(str);
            if (yVar == null) {
                return false;
            }
        } else {
            return false;
        }

        // lower error variable
        str = el.getAttribute(KEY_LOWER_ERROR_VALUE_NAME);
        if (str.length() != 0) {
            lVar = this.mNCFile.findVariable(str);
            if (lVar == null) {
                return false;
            }
        }

        // upper error variable
        str = el.getAttribute(KEY_UPPER_ERROR_VALUE_NAME);
        if (str.length() != 0) {
            uVar = this.mNCFile.findVariable(str);
            if (uVar == null) {
                return false;
            }
        }

        if (!(lVar == null && uVar == null) && !(lVar != null && uVar != null)) {
            return false;
        }
        
        // tick label variable
        str = el.getAttribute(KEY_TICK_LABEL_NAME);
        if (str.length() != 0) {
            tVar = this.mNCFile.findVariable(str);
            if (tVar == null) {
                return false;
            }
        }
        
        // dimension
        str = el.getAttribute(KEY_PICKUP_DIMENSION_NAME);
        if (str.length() != 0) {
            dim = this.mNCFile.findDimension(str);
            if (dim == null) {
                return false;
            }
        }
        
        // dimension start
        str = el.getAttribute(KEY_PICKUP_START);
        if (str.length() != 0) {
            dimStart = SGUtilityText.getInteger(str);
            if (dimStart == null) {
                return false;
            }
        }
        
        // dimension end
        str = el.getAttribute(KEY_PICKUP_END);
        if (str.length() != 0) {
            dimEnd = SGUtilityText.getInteger(str);
            if (dimEnd == null) {
                return false;
            }
        }

        // dimension step
        str = el.getAttribute(KEY_PICKUP_STEP);
        if (str.length() != 0) {
            dimStep = SGUtilityText.getInteger(str);
            if (dimStep == null) {
                return false;
            }
        }

        // set to attribute
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mLowerErrorVariable = lVar;
        this.mUpperErrorVariable = uVar;
        this.mTickLabelVariable = tVar;
        this.mDimension = dim;
        this.mDimensionStart = dimStart;
        this.mDimensionEnd = dimEnd;
        this.mDimensionStep = dimStep;
        
        this.updateDimensionIndices();

        return true;
    }
    
    @Override
    public boolean writeProperty(Element el, final int type) {
        if (super.writeProperty(el, type) == false) {
            return false;
        }
        
        el.setAttribute(KEY_X_VALUE_NAME, this.mXVariable.getName());
        el.setAttribute(KEY_Y_VALUE_NAME, this.mYVariable.getName());
        if (this.isErrorBarAvailable()) {
            el.setAttribute(KEY_LOWER_ERROR_VALUE_NAME, this.mLowerErrorVariable.getName());
            el.setAttribute(KEY_UPPER_ERROR_VALUE_NAME, this.mUpperErrorVariable.getName());
        }
        if (this.isTickLabelAvailable()) {
            el.setAttribute(KEY_TICK_LABEL_NAME, this.mTickLabelVariable.getOriginalName());
        }
        
        if (this.mDimension!=null) {
            el.setAttribute(KEY_PICKUP_DIMENSION_NAME, this.mDimension.getName());
        }
        if (this.mDimensionStart!=null) {
            el.setAttribute(KEY_PICKUP_START, String.valueOf(this.mDimensionStart.intValue()));
        }
        if (this.mDimensionEnd!=null) {
            el.setAttribute(KEY_PICKUP_END, String.valueOf(this.mDimensionEnd.intValue()));
        }
        if (this.mDimensionStep!=null) {
            el.setAttribute(KEY_PICKUP_STEP, String.valueOf(this.mDimensionStep.intValue()));
        }
        
        return true;
    }
    
    @Override
    public boolean setData(SGData data) {
        if (!(data instanceof SGSXYMultipleDimensionNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGSXYMultipleDimensionNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYMultipleDimensionNetCDFData nData = (SGSXYMultipleDimensionNetCDFData) data;
        this.mXVariable = nData.mXVariable;
        this.mYVariable = nData.mYVariable;
        if (nData.isErrorBarAvailable()) {
            this.mLowerErrorVariable = nData.mLowerErrorVariable;
            this.mUpperErrorVariable = nData.mUpperErrorVariable;
        }
        if (nData.isTickLabelAvailable()) {
            this.mTickLabelVariable = nData.mTickLabelVariable;
        }
        this.mDimension = nData.mDimension;
        this.updateDimensionIndices();
        return true;
    }

    /**
     * Returns an array of values of multiple variables.
     * 
     * @return an array of values of multiple variables
     */
    protected double[][] getMultipleVariableValueArray() {
        SGVariable[] cVars = new SGVariable[] { this.getSingleVariable() };
        SGVariable multVar = this.getMultipleVariable();
        String dimName = this.mDimension.getName();
        double[][] array = new double[this.mDimensionIndices.length][];
        for (int ii = 0; ii < array.length; ii++) {
            Map<String, Integer> originMap = new HashMap<String, Integer>(this.mOriginMap);
            originMap.put(dimName, this.mDimensionIndices[ii]);
            array[ii] = this.getValueArray(multVar, cVars, originMap);
        }
        return array;
    }

    /**
     * Returns an array of SXYData objects.
     * 
     * @return an array of SXYData objects
     */
    @Override
    public SGISXYTypeData[] getSXYDataArray() {
        final int len = this.mDimensionIndices.length;
        final List<Dimension> dimList = this.mNCFile.getDimensions();
        final String dimName = this.mDimension.getName();
        final String nameMultiple = this.hasMultipleYValues() ? this.mYVariable.getName() : this.mXVariable.getName();
        final String lName = (this.mLowerErrorVariable != null) ? this.mLowerErrorVariable.getName() : null;
        final String uName = (this.mUpperErrorVariable != null) ? this.mUpperErrorVariable.getName() : null;
        final String tName = (this.mTickLabelVariable != null) ? this.mTickLabelVariable.getName() : null;
        final String timeName = (this.mTimeVariable != null) ? this.mTimeVariable.getName() : null;
        final String indexName = (this.mSerialNumberVariable != null) ? this.mSerialNumberVariable.getName() : null;
        final SGISXYTypeData[] dataArray = new SGISXYTypeData[len];
        for (int ii = 0; ii < dataArray.length; ii++) {
            SGNetCDFData data;
            if (this.mXVariable instanceof SGDateVariable || this.mYVariable instanceof SGDateVariable) {
                data = new SGSXYNetCDFDateData(this.mNCFile, this.mObserver, 
                        this.mXVariable.getName(), this.mYVariable.getName(), 
                        lName, uName, nameMultiple, tName, nameMultiple, timeName, indexName);
            } else {
                data = new SGSXYNetCDFData(this.mNCFile, this.mObserver, 
                        this.mXVariable.getName(), this.mYVariable.getName(), 
                        lName, uName, nameMultiple, tName, nameMultiple, timeName, indexName);
            }
            for (Dimension dim : dimList) {
                String name = dim.getName();
                Integer origin;
                if (name.equals(dimName)) {
                    origin = this.mDimensionIndices[ii];
                } else {
                    origin = this.mOriginMap.get(name);
                    if (origin==null) {
                        origin = Integer.valueOf(0);
                    }
                }
                data.setOrigin(name, origin.intValue());
            }
            
            // sets the sub cache
            SGSXYNetCDFData d = (SGSXYNetCDFData) data;
            d.setCacheMap(this.mVariableCache);
            
            dataArray[ii] = (SGISXYTypeData) data;
        }
        return dataArray;
    }
    
    @Override
    public int getOrigin(final String name) {
        if (SGDataUtility.isSameNetCDFGroup(name, this.getDimensionName())) {
            return super.getOrigin(name);
        } else {
            return 0;
        }
    }

    /**
     * Properties for netCDF SXY data with multiple dimensions.
     */
    public static class SXYMultipleDimensionNetCDFDataProperties extends NetCDFDataProperties {

        protected String xName = null;
        
        protected String yName = null;
        
        protected String lName = null;
        
        protected String uName = null;
        
        protected String tName = null;
        
        protected String dimName = null;
        
        protected Integer dimStart = null;
        
        protected Integer dimEnd = null;

        protected Integer dimStep = null;

        /**
         * The default constructor.
         */
        public SXYMultipleDimensionNetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.xName = null;
            this.yName = null;
            this.lName = null;
            this.uName = null;
            this.tName = null;
            this.dimName = null;
            this.dimStart = null;
            this.dimEnd = null;
            this.dimStep = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * 
         * @param obj
         *            an object to be compared
         * @return true if two objects are equal
         */
        @Override
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYMultipleDimensionNetCDFDataProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            SXYMultipleDimensionNetCDFDataProperties p = (SXYMultipleDimensionNetCDFDataProperties) obj;
            if (SGUtility.equals(this.xName, p.xName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.yName, p.yName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.lName, p.lName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.uName, p.uName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.tName, p.tName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.dimName, p.dimName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.dimStart, p.dimStart) == false) {
            	return false;
            }
            if (SGUtility.equals(this.dimEnd, p.dimEnd) == false) {
                return false;
            }
            if (SGUtility.equals(this.dimStep, p.dimStep) == false) {
            	return false;
            }
            return true;
        }
    }

    /**
     * Returns a map which has data information.
     * 
     * @return a map which has data information
     */
    @Override
    public Map<String, Object> getInfoMap() {
        Map<String, Object> infoMap = super.getInfoMap();
        infoMap.put(KEY_MULTIPLE, Boolean.TRUE);
        infoMap.put(KEY_MULTIPLE_VARIABLE, Boolean.FALSE);
        infoMap.put(KEY_MULTIPLE_DIMENSION_START, this.mDimensionStart);
        infoMap.put(KEY_MULTIPLE_DIMENSION_END, this.mDimensionEnd);
        infoMap.put(KEY_MULTIPLE_DIMENSION_STEP, this.mDimensionStep);
        return infoMap;
    }

    /**
     * Returns an array of multiple SXYData objects.
     * 
     * @return an array of multiple SXYData objects
     */
    public SGISXYTypeMultipleData[] getSXYTypeMultipleDataArray() {
        SGISXYTypeData[] array = this.getSXYDataArray();
        SGISXYTypeMultipleData[] ret = new SGISXYTypeMultipleData[array.length];
        for (int ii = 0; ii < array.length; ii++) {
            SGSXYNetCDFData data = (SGSXYNetCDFData) array[ii];
            ret[ii] = data.toMultiple(this.mDimension, this.mDimensionIndices[ii]);
            
            // disposes data
            data.dispose();
        }
        return ret;
    }

    @Override
    public boolean isSplitEnabled() {
        final int len = this.mDimensionIndices.length;
        if (len > 1) {
            return true;
        }
        return false;
    }
    
    public static SGISXYTypeMultipleData merge(final List<SGData> dataList) {
        if (dataList.size()==0) {
            return null;
        }
        for (SGData data : dataList) {
            if ((data instanceof SGSXYMultipleDimensionNetCDFData)==false) {
                return null;
            }
        }
        
        SGSXYMultipleDimensionNetCDFData dataLast = (SGSXYMultipleDimensionNetCDFData)dataList.get(dataList.size()-1);
        SGNetCDFFile ncFile = dataLast.getNetcdfFile();
        SGDataSourceObserver obs = dataLast.mObserver;
        String xName = dataLast.mXVariable.getName();
        String yName = dataLast.mYVariable.getName();
        String lName = (null!=dataLast.mLowerErrorVariable) ? dataLast.mLowerErrorVariable.getName() : null;
        String uName = (null!=dataLast.mUpperErrorVariable) ? dataLast.mUpperErrorVariable.getName() : null;
        String tName = (null!=dataLast.mTickLabelVariable) ? dataLast.mTickLabelVariable.getName() : null;
        String dimName = (null!=dataLast.mDimension) ? dataLast.mDimension.getName() : null;
        String timeName = dataLast.isTimeVariableAvailable() ? dataLast.getTimeVariable().getName() : null;
        String indexName = dataLast.isSerialNumberAvailable() ? dataLast.getSerialNumberVariable().getName() : null;
        
        int dimStart = 0;
        int dimEnd = 0;
        int dimStep = 0;
        for (int i = 0; i < dataList.size(); i++) {
            SGSXYMultipleDimensionNetCDFData dataMult = (SGSXYMultipleDimensionNetCDFData)dataList.get(i);
            int start = dataMult.getDimensionStart();
            int end = dataMult.getDimensionEnd();
            int step = dataMult.getDimensionStep();
            if (i==0) {
                dimStart = start;
                dimEnd = end;
                dimStep = step;
            }
            if (end < start) {
                int itemp = start;
                start = end;
                end = itemp;
            }
            if (start<dimStart) {
                dimStart = start;
            }
            if (end>dimEnd) {
                dimEnd = end;
            }
            if (dimStep < step) {
                dimStep = step;
            }
        }
        
        SGSXYMultipleDimensionNetCDFData data = new SGSXYMultipleDimensionNetCDFData(
                ncFile, obs,
                xName, yName, lName, uName, tName, dimName,
                dimStart, dimEnd, dimStep,
                timeName, indexName);
        data.mOriginMap = new HashMap<String, Integer>(dataLast.mOriginMap);
        data.setDecimalPlaces(dataLast.getDecimalPlaces());
        data.setExponent(dataLast.getExponent());
        
        return data;
    }
    
    @Override
    public boolean isXVariableCoordinate() {
        return this.getXVariable().isCoordinateVariable();
    }
}
