package jp.riken.brain.ni.samuraigraph.data;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;

import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityNumber;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.base.SGValueRange;
import ucar.ma2.Array;
import ucar.nc2.Dimension;

/**
 * The class of scalar XY type data with netCDF data.
 *
 */
public class SGSXYNetCDFData extends SGNetCDFData implements 
    SGISXYTypeData, SGIDataColumnTypeConstants, SGIDataColumnNameConstants {

    /**
     * The variable for x-values.
     */
    protected SGVariable mXVariable = null;
    
    /**
     * The variable for y-values.
     */
    protected SGVariable mYVariable = null;
    
    /**
     * The variable for lower error values.
     */
    protected SGVariable mLowerErrorVariable = null;

    /**
     * The variable for upper error values.
     */
    protected SGVariable mUpperErrorVariable = null;
    
    /**
     * The variable for values that holds error values.
     */
    protected SGVariable mErrorBarHolderVariable = null;
    
    /**
     * The variable for tick labels.
     */
    protected SGVariable mTickLabelVariable = null;

    /**
     * The variable for values that holds tick labels.
     */
    protected SGVariable mTickLabelHolderVariable = null;

    /**
     * The decimal places for the tick labels.
     */
    protected int mDecimalPlaces = 0;

    /**
     * The exponent for the tick labels.
     */
    protected int mExponent = 0;

    /**
     * The default constructor.
     *
     */
    public SGSXYNetCDFData() {
        super();
    }

    /**
     * Builds a data object with given netCDF data.
     * 
     * @param ncfile
     *            netCDF data file
     * @param obs
     *            netCDF data observer
     * @param xName
     *            the name of x variable
     * @param yName
     *            the name of y variable
     * @param lName
     *            the name of lower error variable
     * @param uName
     *            the name of upper error variable
     * @param ehName
     *            the name of error bar holder variable
     * @param tName
     *            the name of tick label variable
     * @param thName
     *            the name of tick label holder variable
     * @param timeName
     *            the name of time variable
     * @param serialNumberName
     *          the name of serial number variable
     */
    public SGSXYNetCDFData(final SGNetCDFFile ncfile, SGDataSourceObserver obs,
            final String xName, final String yName, final String lName, final String uName,
            final String ehName, final String tName, final String thName,
            final String timeName, final String serialNumberName) {
        
        super(ncfile, obs, timeName, serialNumberName);
        
        if (xName == null || yName == null) {
            throw new IllegalArgumentException("xName == null || yName == null");
        }
        if (!(lName == null && uName == null) 
        		&& !(lName != null && uName != null)) {
            throw new IllegalArgumentException("lName = " + lName + ", uName = " + uName);
        }
        if (lName != null && ehName == null) {
            throw new IllegalArgumentException("Error bar holder name is null.");
        }
        if (tName != null && thName == null) {
            throw new IllegalArgumentException("Tick label holder name is null.");
        }
        
        // get variables
        SGVariable xVar = this.getVariable(ncfile, xName);
        SGVariable yVar = this.getVariable(ncfile, yName);
        SGVariable lVar = this.getVariable(ncfile, lName);
        SGVariable uVar = this.getVariable(ncfile, uName);
        SGVariable ehVar = this.getVariable(ncfile, ehName);
        SGVariable tVar = this.getVariable(ncfile, tName);
        SGVariable thVar = this.getVariable(ncfile, thName);
        
        //
        // check variables
        //
        List<Dimension> cDimList = new ArrayList<Dimension>();
        
        // time variable
    	if (this.mTimeVariable != null) {
    		Dimension tDim = this.mTimeVariable.getDimension(0);
    		cDimList.add(tDim);
    	}

        if (this.isSerialNumberAvailable()) {
        	// x and y variables
            if (xVar.isCoordinateVariable() || yVar.isCoordinateVariable()) {
                throw new IllegalArgumentException("Both of x and y variables must not be coordinate variables.");
            }
            
        	// index variable
            Dimension iDim = this.mSerialNumberVariable.getDimension(0);
            cDimList.add(iDim);
            
            // x and y variables
            this.checkNotCoordinateVariable(xVar, cDimList);
            this.checkNotCoordinateVariable(yVar, cDimList);
        } else {
        	// x and y variables
            if ((xVar.isCoordinateVariable() && yVar.isCoordinateVariable())
            		|| (!xVar.isCoordinateVariable() && !yVar.isCoordinateVariable())) {
                throw new IllegalArgumentException("Either of x and y variables must be a coordinate variable.");
            }
            SGVariable cVar = null;
            SGVariable oVar = null;
            if (xVar.isCoordinateVariable()) {
            	cVar = xVar;
            	oVar = yVar;
            } else {
            	cVar = yVar;
            	oVar = xVar;
            }
            Dimension cDim = cVar.getDimension(0);
            cDimList.add(cDim);
            
            // time variable
        	if (this.mTimeVariable != null) {
        		Dimension tDim = this.mTimeVariable.getDimension(0);
                if (cDim.equals(tDim)) {
                    throw new IllegalArgumentException(
            			"The dimension of a coordinate variable is equal to the dimension for time.");
                }
        	}
        	
            // y variables
            this.checkNotCoordinateVariable(oVar, cDimList);
        }
        
        // other variables
    	this.checkNotCoordinateVariable(lVar, cDimList);
    	this.checkNotCoordinateVariable(uVar, cDimList);
    	this.checkNotCoordinateVariable(ehVar, cDimList);
    	this.checkNotCoordinateVariable(tVar, cDimList);
    	this.checkNotCoordinateVariable(thVar, cDimList);
        
        // set to attributes
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mLowerErrorVariable = lVar;
        this.mUpperErrorVariable = uVar;
        this.mErrorBarHolderVariable = ehVar;
        this.mTickLabelVariable = tVar;
        this.mTickLabelHolderVariable = thVar;
    }
    
    /**
     * Disposes of this data object.
     * Do not call any method of this object after called this method.
     */
    public void dispose() {
        super.dispose();
        
        // Note: Do not clear the caches!
        
        this.mXVariable = null;
        this.mYVariable = null;
        this.mLowerErrorVariable = null;
        this.mUpperErrorVariable = null;
        this.mErrorBarHolderVariable = null;
        this.mTickLabelVariable = null;
        this.mTickLabelHolderVariable = null;
    }
    
    private SGVariable getCoordinateVariable() {
    	if (this.isSerialNumberAvailable()) {
    		return this.mSerialNumberVariable;
    	} else {
            if (this.mXVariable.isCoordinateVariable()) {
                return this.mXVariable;
            } else {
                return this.mYVariable;
            }
    	}
    }

    /**
     * Returns the properties of this data.
     * 
     * @return the properties of this data
     */
    public SGProperties getProperties() {
        SGProperties p = new SXYNetCDFDataProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }

    /**
     * Gets the properties of this data.
     * @param p
     *      the properties of this data
     * @return
     *      true if succeeded   
     */
    public boolean getProperties(SGProperties p) {
        if (!(p instanceof SXYNetCDFDataProperties)) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        SXYNetCDFDataProperties sp = (SXYNetCDFDataProperties) p;
        sp.xName = this.getName(this.mXVariable);
        sp.yName = this.getName(this.mYVariable);
        sp.lName = this.getName(this.mLowerErrorVariable);
        sp.uName = this.getName(this.mUpperErrorVariable);
        sp.ehName = this.getName(this.mErrorBarHolderVariable);
        sp.tName = this.getName(this.mTickLabelVariable);
        sp.thName = this.getName(this.mTickLabelHolderVariable);
        return true;
    }
    
    // This method is not already called anywhere.
//    /**
//     * Read properties.
//     * 
//     * @param el
//     *           an Element object
//     * @return true if succeeded
//     */
//    public boolean readProperty(Element el) {
//        
//        if (super.readProperty(el) == false) {
//            return false;
//        }
//        
//        String str = null;
//        SGVariable xVar = null;
//        SGVariable yVar = null;
//        SGVariable lVar = null;
//        SGVariable uVar = null;
//        SGVariable ehVar = null;
//        SGVariable tVar = null;
//        SGVariable thVar = null;
//        
//        // x variable
//        str = el.getAttribute(KEY_X_VALUE_NAME);
//        if (str.length() != 0) {
//            xVar = this.mNCFile.findVariable(str);
//            if (xVar == null) {
//                return false;
//            }
//        } else {
//            return false;
//        }
//        
//        // y variable
//        str = el.getAttribute(KEY_Y_VALUE_NAME);
//        if (str.length() != 0) {
//            yVar = this.mNCFile.findVariable(str);
//            if (yVar == null) {
//                return false;
//            }
//        } else {
//            return false;
//        }
//
//        // lower error variable
//        str = el.getAttribute(KEY_LOWER_ERROR_VALUE_NAME);
//        if (str.length() != 0) {
//            lVar = this.mNCFile.findVariable(str);
//            if (lVar == null) {
//                return false;
//            }
//        }
//        
//        // error bar holder variable
//        str = el.getAttribute(KEY_ERROR_BAR_HOLDER_NAME);
//        if (str.length() != 0) {
//            ehVar = this.mNCFile.findVariable(str);
//            if (ehVar == null) {
//                return false;
//            }
//        }
//
//        if (!(lVar == null && uVar == null && ehVar == null) 
//        		&& !(lVar != null && uVar != null && ehVar != null)) {
//            return false;
//        }
//        
//        // tick label variable
//        str = el.getAttribute(KEY_TICK_LABEL_NAME);
//        if (str.length() != 0) {
//            tVar = this.mNCFile.findVariable(str);
//            if (tVar == null) {
//                return false;
//            }
//        }
//        
//        // tick label holder variable
//        str = el.getAttribute(KEY_TICK_LABEL_HOLDER_NAME);
//        if (str.length() != 0) {
//            thVar = this.mNCFile.findVariable(str);
//            if (thVar == null) {
//                return false;
//            }
//        }
//
//        if (!(tVar == null && thVar == null) && !(tVar != null && thVar != null)) {
//            return false;
//        }
//
//        // set to attribute
//        this.mXVariable = xVar;
//        this.mYVariable = yVar;
//        this.mLowerErrorVariable = lVar;
//        this.mUpperErrorVariable = uVar;
//        this.mErrorBarHolderVariable = ehVar;
//        this.mTickLabelVariable = tVar;
//        this.mTickLabelHolderVariable = thVar;
//
//        return true;
//    }

//    /**
//     * Save values to given file.
//     * 
//     * @param data
//     *             a data
//     * @return
//     *             true if succeeded
//     */
//    public boolean saveData(final File file) {
//        boolean status = false;
//        Variable var = this.getVariable();
//        Variable cVar = this.getCoordinateVariable();
//        Variable[] vArray;
//        if (this.mLowerErrorVariable != null && this.mUpperErrorVariable != null) {
//            vArray = new Variable[]{cVar, var, this.mLowerErrorVariable, this.mUpperErrorVariable};
//        } else {
//            vArray = new Variable[]{cVar, var};                
//        }
//        try {
//            FileWriter fw = new FileWriter(file);
//            for (int ii = 0; ii < vArray.length; ii++) {
//                String title = this.getNameWithUnit(vArray[ii]);
//                if (title == null) {
//                    title = "";
//                }
//                String str = SGUtilityText.getCSVString(title);
//                if (ii != 0) {
//                    fw.write(",");
//                }
//                fw.write(str);
//            }
//            fw.write("\n");
//            status = this.writeData(fw);
//            fw.close();
//        } catch (IOException ex) {
//            return false;
//        }
//        return status;
//    }

    /**
     * Sets the data.
     * 
     * @param data
     *             data set to this object
     * @return
     *             true if succeeded            
     */
    public boolean setData(SGData data) {
        if (!(data instanceof SGSXYNetCDFData)) {
            throw new IllegalArgumentException("Given data is not the instance of SGSXYNetCDFData class.");
        }
        if (super.setData(data) == false) {
            return false;
        }
        SGSXYNetCDFData nData = (SGSXYNetCDFData) data;
        this.mXVariable = nData.mXVariable;
        this.mYVariable = nData.mYVariable;
        this.mLowerErrorVariable = nData.mLowerErrorVariable;
        this.mUpperErrorVariable = nData.mUpperErrorVariable;
        this.mErrorBarHolderVariable = nData.mErrorBarHolderVariable;
        this.mTickLabelVariable = nData.mTickLabelVariable;
        this.mTickLabelHolderVariable = nData.mTickLabelHolderVariable;
        return true;
    }

    /**
     * Set properties to this data.
     * 
     * @param p
     *          properties to be set
     * @return true if succeeded
     */
    public boolean setProperties(SGProperties p) {
        if (!(p instanceof SXYNetCDFDataProperties)) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        SXYNetCDFDataProperties sp = (SXYNetCDFDataProperties) p;
        this.mXVariable = this.findVariable(sp.xName);
        this.mYVariable = this.findVariable(sp.yName);
        this.mLowerErrorVariable = this.findVariable(sp.lName);
        this.mUpperErrorVariable = this.findVariable(sp.uName);
        this.mErrorBarHolderVariable = this.findVariable(sp.ehName);
        this.mTickLabelVariable = this.findVariable(sp.tName);
        this.mTickLabelHolderVariable = this.findVariable(sp.thName);
        return true;
    }

    // This method is not already called anywhere.
//    /**
//     * Write this data to a given output stream.
//     * 
//     * @param w
//     *          an output stream
//     * @return
//     *          true if succeeded
//     */
//    protected boolean writeData(Writer w) throws IOException {
//        double[] cvArray = this.getCoordinateValueArray();
//        double[] vArray = this.getValueArray();
//        double[] lArray = this.getLowerErrorValueArray();
//        double[] uArray = this.getUpperErrorValueArray();
//        final boolean bError = (lArray != null && uArray != null);
//        StringBuffer sb = new StringBuffer();
//        for (int ii = 0; ii < cvArray.length; ii++) {
//            sb.append(cvArray[ii]);
//            sb.append(',');
//            sb.append(vArray[ii]);
//            if (bError) {
//                sb.append(',');
//                sb.append(lArray[ii]);
//                sb.append(',');
//                sb.append(uArray[ii]);                
//            }
//            sb.append('\n');
//        }
//        w.write(sb.toString());
//        return true;
//    }

    /**
     * Returns an array of values to save to a data set.
     * 
     * @return an array of values to save to a data set
     */
    protected double[][] getDataSetValues() {
        final double[] xArray = this.getXValueArray();
        final double[] yArray = this.getYValueArray();
        final double[] lArray = this.getLowerErrorValueArray();
        final double[] uArray = this.getUpperErrorValueArray();
        final int pNum = xArray.length;
        
        final boolean bError = (lArray != null && uArray != null);
        final int colNum = bError ? 4 : 2;
        double[][] colArray = null;
        if (bError) {
            colArray = new double[][]{xArray, yArray, lArray, uArray};
        } else {
            colArray = new double[][]{xArray, yArray};
        }
        
        double[][] values = new double[pNum][];
        for (int ii = 0; ii < pNum; ii++) {
            values[ii] = new double[colNum];
            for (int jj = 0; jj < colNum; jj++) {
                values[ii][jj] = colArray[jj][ii];
            }
        }
// TODO this method will be deleted        
        return values;
    }

//    /**
//     * Write properties of this object to the Element.
//     * 
//     * @param el
//     *            the Element object
//     * @param type
//     *            type of the method to save properties
//     * @return true if succeeded
//     */
//    public boolean writeProperty(Element el, final int type) {
//        if (super.writeProperty(el, type) == false) {
//            return false;
//        }
//        el.setAttribute(KEY_X_VALUE_NAME, this.mXVariable.getName());
//        el.setAttribute(KEY_Y_VALUE_NAME, this.mYVariable.getName());
//        if (this.isErrorBarAvailable()) {
//            el.setAttribute(KEY_LOWER_ERROR_VALUE_NAME, this.mLowerErrorVariable.getName());
//            el.setAttribute(KEY_UPPER_ERROR_VALUE_NAME, this.mUpperErrorVariable.getName());
//            el.setAttribute(KEY_ERROR_BAR_HOLDER_NAME, this.mErrorBarHolderVariable.getName());
//        }
//        if (this.isTickLabelAvailable()) {
//            el.setAttribute(KEY_TICK_LABEL_NAME, this.mTickLabelVariable.getOriginalName());
//            el.setAttribute(KEY_TICK_LABEL_HOLDER_NAME, this.mTickLabelHolderVariable.getName());
//        }
//        return true;
//    }

    /**
     * Returns a text string of data type.
     * 
     * @return a text string of data type
     */
    public String getDataType() {
        return SGDataTypeConstants.SXY_NETCDF_DATA;
    }

    /**
     * Returns the lower error value at given index.
     * 
     * @param n
     *          array index of the lower error values
     * @return the lower error value at given index
     */
    public Double getLowerErrorValue(int n) {
        double[] array = this.getLowerErrorValueArray();
        if (array == null) {
            return null;
        } else {
            return Double.valueOf(array[n]);
        }
    }

    /**
     * Returns an array of lower error values.
     * 
     * @return an array of lower error values
     */
    public double[] getLowerErrorValueArray() {
        if (this.isErrorBarAvailable()) {
            double[] eArray = SGUtility.copyDoubleArray(this.getValueArray(this.mLowerErrorVariable, 
                    new SGVariable[]{ this.getCoordinateVariable() }));
            for (int ii = 0; ii < eArray.length; ii++) {
                eArray[ii] = - Math.abs(eArray[ii]);
            }
            return eArray;
        } else {
            return null;
        }
    }

    /**
     * Returns the number of points.
     * 
     * @return the number of points
     */
    public int getPointsNumber() {
    	SGVariable var = this.getCoordinateVariable();
    	Dimension dim = var.getDimension(0);
    	return dim.getLength();
    }

    @Override
    public String getString(int n) {
        final int len = this.getPointsNumber();
        if (n < 0 || n >= len) {
            throw new IllegalArgumentException("given index is out of bounds: " + n);
        }
        if (this.isTickLabelAvailable()) {
            String[] array = this.getTextArray();
            if (array == null) {
                return "";
            } else {
                return array[n];
            }
        }
        return "";
    }
    
    /**
     * Returns an array of strings.
     * @return
     *         an array of strings
     */
    protected String[] getTextArray() {
        if (this.isTickLabelAvailable()) {
            if (this.mTickLabelVariable instanceof SGTextVariable) {
                return ((SGTextVariable) this.mTickLabelVariable).getStringArray();
            } else if (this.mTickLabelVariable instanceof SGDateVariable) {
                return ((SGDateVariable)this.mTickLabelVariable).getStringArray();
            } else if (this.mTickLabelVariable.getVariable()!=null) {
                if (this.mTickLabelVariable.getVariable().getDataType().isString()) {
                    return null;
                } else {
                    double[] dArray = this.getValueArray(
                            this.mTickLabelVariable, 
                            new SGVariable[]{ this.getCoordinateVariable() });
                    String[] tArray = SGUtilityNumber.getStringArray(dArray, this.mDecimalPlaces, this.mExponent);
                    return tArray;
                }
            }
        }
        return null;
    }

    public String[] getStringArray() {
        return this.getTextArray();
    }

    /**
     * Returns the upper error value at given index.
     * 
     * @param n
     *          array index of the upper error values
     * @return the upper error value at given index
     */
    public Double getUpperErrorValue(int n) {
        double[] array = this.getUpperErrorValueArray();
        if (array == null) {
            return null;
        } else {
            return Double.valueOf(array[n]);
        }
    }

    /**
     * Returns an array of upper error values.
     * 
     * @return an array of upper error values
     */
    public double[] getUpperErrorValueArray() {
        if (this.isErrorBarAvailable()) {
            double[] eArray = SGUtility.copyDoubleArray(this.getValueArray(this.mUpperErrorVariable, 
                    new SGVariable[]{ this.getCoordinateVariable() }));
            for (int ii = 0; ii < eArray.length; ii++) {
                eArray[ii] = Math.abs(eArray[ii]);
            }
            return eArray;
        } else {
            return null;
        }
    }

    /**
     * Returns the n-th value of X-values.
     * 
     * @param n
     *          the index of value
     * @return the X-value at given index
     */
    public Double getXValue(int n) {
        double[] array = this.getXValueArray();
        if (array == null) {
            return null;
        } else {
            return Double.valueOf(array[n]);
        }
    }

    /**
     * Returns an array of X-values.
     * 
     * @return an array of X-values
     */
    public double[] getXValueArray() {
        if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mXVariable, new SGVariable[]{ this.mSerialNumberVariable });
        } else {
            if (this.mXVariable.isCoordinateVariable()) {
                return this.getCoordinateValueArray(this.mXVariable);
            } else {
                return this.getValueArray(this.mXVariable, new SGVariable[]{ this.mYVariable });
            }
        }
    }

    /**
     * Returns the n-th value of Y-values.
     * 
     * @param n
     *          the index of value
     * @return the Y-value at given index
     */
    public Double getYValue(int n) {
        double[] array = this.getYValueArray();
        if (array == null) {
            return null;
        } else {
            return Double.valueOf(array[n]);
        }
    }

    /**
     * Returns an array of Y-values.
     * 
     * @return an array of Y-values
     */
    public double[] getYValueArray() {
    	if (this.isSerialNumberAvailable()) {
            return this.getValueArray(this.mYVariable, new SGVariable[]{ this.mSerialNumberVariable });
    	} else {
            if (this.mYVariable.isCoordinateVariable()) {
                return this.getCoordinateValueArray(this.mYVariable);
            } else {
                return this.getValueArray(this.mYVariable, new SGVariable[]{ this.mXVariable });
            }
    	}
    }

    /**
     * Returns whether error bars are available.
     * 
     * @return true if error bars are available
     */
    public boolean isErrorBarAvailable() {
        return (this.mLowerErrorVariable != null && this.mUpperErrorVariable != null);
    }

    /**
     * Returns whether tick labels are available.
     * 
     * @return true if tick labels are available
     */
    public boolean isTickLabelAvailable() {
        return (this.mTickLabelVariable != null);
    }

    /**
     * Returns whether a given data object has the same tick label resources.
     * 
     * @param data
     *           a data to compare
     * @return true if a given data object has the same tick label resources
     */
    public boolean hasEqualTickLabelResource(SGISXYTypeData data) {
    	if (!(data instanceof SGSXYNetCDFData)) {
    		return false;
    	}
    	SGSXYNetCDFData dataSXY = (SGSXYNetCDFData) data;
    	if (this.mTickLabelVariable == null) {
    		return (dataSXY.mTickLabelVariable == null);
    	} else {
        	return this.mTickLabelVariable.equals(dataSXY.mTickLabelVariable);
    	}
    }
    
    /**
     * Returns the bounds of x-values.
     * 
     * @return the bounds of x-values
     */
    public SGValueRange getBoundsX() {
        return SGDataUtility.getBoundsX(this);
    }

    /**
     * Returns the bounds of y-values.
     * 
     * @return the bounds of y-values
     */
    public SGValueRange getBoundsY() {
        return SGDataUtility.getBoundsY(this);
    }

    /**
     * Returns the title for the X-axis.
     * 
     * @return the title for the X-axis
     */
    public String getTitleX() {
        return this.getNameWithUnit(this.mXVariable);
    }

    /**
     * Returns the title for the Y-axis.
     * 
     * @return the title for the Y-axis
     */
    public String getTitleY() {
        return this.getNameWithUnit(this.mYVariable);
    }
    
    /**
     * A class of scalar XY netCDF data properties.
     */
    public static class SXYNetCDFDataProperties extends NetCDFDataProperties {

        protected String xName = null;
        
        protected String yName = null;
        
        protected String lName = null;
        
        protected String uName = null;
        
        protected String ehName = null;
        
        protected String tName = null;
        
        protected String thName = null;
        
        /**
         * The default constructor.
         */
        public SXYNetCDFDataProperties() {
            super();
        }

        /**
         * Dispose this object.
         */
        public void dispose() {
            super.dispose();
            this.xName = null;
            this.yName = null;
            this.lName = null;
            this.uName = null;
            this.ehName = null;
            this.tName = null;
            this.thName = null;
        }
        
        /**
         * Returns whether this object is equal to given object.
         * @param obj
         *            an object to be compared
         * @return
         *            true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof SXYNetCDFDataProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            SXYNetCDFDataProperties p = (SXYNetCDFDataProperties) obj;
            if (SGUtility.equals(this.xName, p.xName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.yName, p.yName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.lName, p.lName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.uName, p.uName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.ehName, p.ehName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.tName, p.tName) == false) {
            	return false;
            }
            if (SGUtility.equals(this.thName, p.thName) == false) {
            	return false;
            }
            return true;
        }
    }

    /**
     * Returns an array of current column types.
     * 
     * @return an array of current column types
     */
    @Override
    public String[] getCurrentColumnType() {
        List<SGVariable> varList = this.mNCFile.getVariables();
        final int varNum = varList.size();
        String[] array = new String[varNum];
        
        for (int ii = 0; ii < varNum; ii++) {
            SGVariable var = varList.get(ii);
            if (var.equals(this.mXVariable)) {
                array[ii] = X_VALUE;
            } else if (var.equals(this.mYVariable)) {
                array[ii] = Y_VALUE;
            } else if (var.equals(this.mLowerErrorVariable) 
                    || var.equals(this.mUpperErrorVariable)) {
                String name = null;
                if (this.mLowerErrorVariable.equals(this.mUpperErrorVariable)) {
                    name = LOWER_UPPER_ERROR_VALUE;
                } else {
                    if (var.equals(this.mLowerErrorVariable)) {
                        name = LOWER_ERROR_VALUE;
                    } else if (var.equals(this.mUpperErrorVariable)) {
                        name = UPPER_ERROR_VALUE;
                    }
                }
                if (name == null) {
                    return null;
                }
                final int index = this.mNCFile.getVariableIndex(this.mErrorBarHolderVariable.getName());
                if (index == -1) {
                    return null;
                }
                array[ii] = SGDataUtility.appendColumnTitle(name, this.mErrorBarHolderVariable.getName());
            } else if (var.equals(this.mTickLabelVariable)) {
                final int index = this.mNCFile.getVariableIndex(this.mTickLabelHolderVariable.getName());
                if (index == -1) {
                    return null;
                }
                array[ii] = SGDataUtility.appendColumnTitle(TICK_LABEL, this.mTickLabelHolderVariable.getName());
            } else if (var.equals(this.mTimeVariable)) {
                array[ii] = TIME;
            } else if (var.equals(this.mSerialNumberVariable)) {
				array[ii] = SERIAL_NUMBER;
            } else {
                array[ii] = "";
            }
        }        
        return array;
    }
    
    /**
	 * Sets the type of data columns.
	 * 
	 * @param column
	 *            an array of column types
	 * @return true if succeeded
	 */
    public boolean setColumnType(String[] columns) {
    	
    	SGVariable xVar = null;
    	SGVariable yVar = null;
    	SGVariable lVar = null;
    	SGVariable uVar = null;
    	SGVariable tVar = null;
    	SGVariable timeVar = null;
    	SGVariable serialNumberVar = null;
    	String lCol = null;
    	String uCol = null;
    	String tCol = null;
        
        // set variables
        List<SGVariable> varList = this.mNCFile.getVariables();
        for (int ii = 0; ii < columns.length; ii++) {
            SGVariable var = varList.get(ii);
            if (SGUtilityText.isEqualString(X_VALUE, columns[ii])) {
                xVar = var;
            } else if (SGUtilityText.isEqualString(Y_VALUE, columns[ii])) {
                yVar = var;
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_ERROR_VALUE)) {
                lVar = var;
                lCol = columns[ii];
            } else if (SGUtilityText.startsWith(columns[ii], UPPER_ERROR_VALUE)) {
                uVar = var;
                uCol = columns[ii];
            } else if (SGUtilityText.startsWith(columns[ii], LOWER_UPPER_ERROR_VALUE)) {
                lVar = var;
                uVar = var;
                lCol = columns[ii];
                uCol = columns[ii];
            } else if (SGUtilityText.startsWith(columns[ii], TICK_LABEL)) {
                tVar = var;
                tCol = columns[ii];
            } else if (SGUtilityText.isEqualString(columns[ii], TIME)) {
                timeVar = var;
            } else if (SGUtilityText.isEqualString(columns[ii], SERIAL_NUMBER)) {
                serialNumberVar = var;
            } else if ("".equals(columns[ii])) {
            	continue;
            } else {
            	return false;
            }
        }
        if (xVar == null || yVar == null) {
            return false;
        }
        if (!((lVar != null) && (uVar != null)) && !((lVar == null) && (uVar == null))) {
            return false;
        }

        SGVariable ehVar = null;
        if (lCol != null && uCol != null) {
            Integer ehle = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(lCol, this.mNCFile);
            Integer ehue = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(uCol, this.mNCFile);
            if (ehle == null || ehue == null || ehle.equals(ehue)==false) {
                return false;
            }
            ehVar = varList.get(ehle.intValue());
        }

        SGVariable thVar = null;
        if (tCol != null) {
            Integer th = SGDataUtility.getColumnIndexOfNetCDFAppendedColumnTitle(tCol, this.mNCFile);
            if (th == null) {
                return false;
            }
            thVar = varList.get(th.intValue());
        }
    	
        // set variables
        this.mXVariable = xVar;
        this.mYVariable = yVar;
        this.mLowerErrorVariable = lVar;
        this.mUpperErrorVariable = uVar;
        this.mErrorBarHolderVariable = ehVar;
        this.mTickLabelVariable = tVar;
        this.mTickLabelHolderVariable = thVar;
        this.mTimeVariable = timeVar;
        this.mSerialNumberVariable = serialNumberVar;
        
        return true;
    }
    
    /**
     * Returns the data type for the data set.
     * 
     * @return the data type
     */
    protected String getDataTypeForDataSet() {
        return SGDataTypeConstants.SXY_DATA;
    }

    /**
     * Returns an array of all variables.
     * 
     * @return an array of all variables
     */
    public SGVariable[] getVariables() {
    	List<SGVariable> varList = new ArrayList<SGVariable>();
    	varList.add(this.mXVariable);
    	varList.add(this.mYVariable);
    	final boolean be = this.isErrorBarAvailable();
    	final boolean bt = this.isTickLabelAvailable();
        if (be && bt) {
        	varList.add(this.mLowerErrorVariable);
        	varList.add(this.mUpperErrorVariable);
        	varList.add(this.mTickLabelVariable);
        } else if (be && !bt) {
        	varList.add(this.mLowerErrorVariable);
        	varList.add(this.mUpperErrorVariable);
        } else if (!be && bt) {
        	varList.add(this.mTickLabelVariable);
        }
        if (this.mTimeVariable != null) {
        	varList.add(this.mTimeVariable);
        }
        if (this.mSerialNumberVariable != null) {
        	varList.add(this.mSerialNumberVariable);
        }
        SGVariable[] varArray = new SGVariable[varList.size()];
        return (SGVariable[]) varList.toArray(varArray);
    }

    public SGVariable getXVariable() {
    	return this.mXVariable;
    }

    public SGVariable getYVariable() {
    	return this.mYVariable;
    }

    public SGVariable getLowerErrorVariable() {
    	return this.mLowerErrorVariable;
    }

    public SGVariable getUpperErrorVariable() {
    	return this.mUpperErrorVariable;
    }

    public SGVariable getErrorHolderVariable() {
    	return this.mErrorBarHolderVariable;
    }

    public SGVariable getTickLabelVariable() {
    	return this.mTickLabelVariable;
    }

    public SGVariable getTickLabelHolderVariable() {
    	return this.mTickLabelHolderVariable;
    }
    
    /**
     * Sets the decimal places for the tick labels.
     * 
     * @param dp
     *          a value to set to the decimal places
     */
    @Override
    public void setDecimalPlaces(final int dp) {
        if (dp < 0) {
            throw new IllegalArgumentException("Decimal places must not be negative: " + dp);
        }
        this.mDecimalPlaces = dp;
    }

    /**
     * Sets the exponent for the tick labels.
     * 
     * @param exp
     *           a value to set to the exponent
     */
    @Override
    public void setExponent(final int exp) {
        this.mExponent = exp;
    }
    
    /**
     * Returns the decimal places for the tick labels.
     * 
     * @return the decimal places for the tick labels
     */
    @Override
    public int getDecimalPlaces() {
        return this.mDecimalPlaces;
    }

    /**
     * Returns the exponent for tick labels.
     * 
     * @return the exponent for tick labels
     */
    @Override
    public int getExponent() {
        return this.mExponent;
    }
    
    protected Double[] arrangeValueArray(Array data2D, SGVariable var, SGVariable[] cVars) {
        return this.arrange1Dim(data2D);
    }

    /**
	 * Returns whether the error bars are vertical. If this data does not have
	 * error values, returns null.
	 * 
	 * @return true if error bars are vertical, false if they are horizontal and
	 *         null if this data do not have error values
	 */
    public Boolean isErrorBarVertical() {
    	if (this.isErrorBarAvailable()) {
    		return (this.mErrorBarHolderVariable.equals(this.mYVariable));
    	} else {
    		return null;
    	}
    }

    /**
	 * Returns whether the tick labels align horizontally. If this data does not have
	 * tick labels, returns null.
	 * 
	 * @return true if tick labels align horizontally, false if they do not so and
	 *         null if this data do not have tick labels
	 */
	public Boolean isTickLabelHorizontal() {
    	if (this.isTickLabelAvailable()) {
    		return (this.mTickLabelHolderVariable.equals(this.mYVariable));
    	} else {
    		return null;
    	}
	}
	
    /**
     * Return whether the XVariable is a coordinate variable.
     * 
     * @return whether the XVariable is a coordinate variable.
     */
    public boolean isXVariableCoordinate() {
        return this.getXVariable().isCoordinateVariable();
    }

    /**
     * Transforms this data to a multiple type data.
     * 
     * @return a transformed data
     */
    public SGISXYTypeMultipleData toMultiple() {
        // Convert this class (SGSXYNetCDFData) to class (SGSXYMultipleVariableNetCDFData).
        final boolean be = this.isErrorBarAvailable();
        final boolean bt = this.isTickLabelAvailable();
        String[] x = new String[] { this.mXVariable.getName() };
        String[] y = new String[] { this.mYVariable.getName() };
        String[] le = be ? new String[] { this.mLowerErrorVariable.getName() } : null;
        String[] ue = be ? new String[] { this.mUpperErrorVariable.getName() } : null;
        String[] eh = be ? new String[] { this.mErrorBarHolderVariable.getName() } : null;
        String[] t = bt ? new String[] { this.mTickLabelVariable.getName() } : null;
        String[] th = bt ? new String[] { this.mTickLabelHolderVariable.getName() } : null;
        String time = (this.mTimeVariable != null) ? this.mTimeVariable.getName() : null;
        String index = this.isSerialNumberAvailable() ? this.mSerialNumberVariable.getName() : null;
        SGSXYMultipleVariableNetCDFData data = new SGSXYMultipleVariableNetCDFData(
                this.mNCFile, this.mObserver, x, y, le, ue, eh, t, th, time, index);
        data.mOriginMap = new HashMap<String, Integer>(this.mOriginMap);
        data.setDecimalPlaces(this.mDecimalPlaces);
        data.setExponent(this.mExponent);
        data.mVariableCache = new HashMap<String, Object>(this.mVariableCache);
        return data;
    }

    /**
     * Transforms this data to a multiple type data.
     * 
     * @param dim
     *            the dimension
     * @param index
     *            the index of dimension
     * @return a transformed data
     */
    public SGISXYTypeMultipleData toMultiple(final Dimension dim, final int index) {
        final boolean be = this.isErrorBarAvailable();
        final boolean bt = this.isTickLabelAvailable();
        String le = be ? this.mLowerErrorVariable.getName() : null;
        String ue = be ? this.mUpperErrorVariable.getName() : null;
        String t = bt ? this.mTickLabelVariable.getName() : null;
        String time = (this.mTimeVariable != null) ? this.mTimeVariable.getName() : null;
        String id = this.isSerialNumberAvailable() ? this.mSerialNumberVariable.getName() : null;
        SGSXYMultipleDimensionNetCDFData data = new SGSXYMultipleDimensionNetCDFData(
                this.mNCFile, this.mObserver, this.mXVariable.getName(),
                this.mYVariable.getName(), le, ue, t, dim.getName(),
                index, index, 0, time, id);
        data.mVariableCache = new HashMap<String, Object>(this.mVariableCache);
        return data;
    }

}
