package jp.riken.brain.ni.samuraigraph.data;

import java.io.IOException;
import java.util.List;

import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import ucar.ma2.Array;
import ucar.ma2.InvalidRangeException;
import ucar.nc2.Attribute;
import ucar.nc2.Dimension;
import ucar.nc2.Variable;

/**
 * The wrapper class for the netCDF variable.
 *
 */
public class SGVariable implements SGINetCDFConstants {

	/**
	 * The variable of netCDF.
	 * 
	 */
	protected Variable mVariable = null;

	/**
	 * The default constructor.
	 *
	 */
	protected SGVariable() {
		super();
	}
	
	/**
	 * Builds an object with a given variable.
	 *
	 * @param var
	 *           the variable
	 */
	public SGVariable(Variable var) {
		super();
		if (var == null) {
			throw new IllegalArgumentException("var == null");
		}
		this.mVariable = var;
	}

	/**
	 * Returns the variable.
	 * 
	 * @return the variable
	 */
	public Variable getVariable() {
		return this.mVariable;
	}

	/**
	 * Returns the name of this variable.
	 * 
	 * @return the name of this variable
	 */
	public String getName() {
		return this.mVariable.getName();
	}
	
	/**
	 * Returns the long name if it exists.
	 * 
	 * @return the long name
	 */
	public String getLongName() {
		Attribute attr = this.findAttribute(ATTR_LONG_NAME);
		if (attr != null) {
			return attr.getStringValue();
		} else {
			return null;
		}
	}

	/**
	 * Returns the standard name if it exists.
	 * 
	 * @return the standard name
	 */
	public String getStandardName() {
		Attribute attr = this.findAttribute(ATTR_STANDARD_NAME);
		if (attr != null) {
			return attr.getStringValue();
		} else {
			return null;
		}
	}

	/**
	 * Returns the name of this variable in order of priority:
	 * long_name, standard_name and variable name.
	 * 
	 * @return the name of this variable
	 */
	public String getNameInPriorityOrder() {
		String name = null;
		if ((name = this.getLongName()) != null) {
			return name;
		} else if ((name = this.getStandardName()) != null) {
			return name;
		} else {
			return this.getName();
		}
	}
	
	/**
     * Returns the name of this netCDF variable.
     * 
     * @return the name of this netCDF variable
	 */
	public String getOriginalName() {
	    return this.getName();
	}
	
	public int findDimensionIndex(final String name) {
		return this.mVariable.findDimensionIndex(name);
	}
	
	public List<Dimension> getDimensions() {
		return this.mVariable.getDimensions();
	}
	
	public Dimension getDimension(final int index) {
		return this.mVariable.getDimension(index);
	}
	
	public boolean isCoordinateVariable() {
		return this.mVariable.isCoordinateVariable();
	}
	
	public Attribute findAttribute(final String name) {
		return this.mVariable.findAttribute(name);
	}

	public String getUnitsString() {
		return this.mVariable.getUnitsString();
	}
	
	public boolean isUnlimited() {
		return this.mVariable.isUnlimited();
	}
	
	public Array read(int[] a, int[] b) throws InvalidRangeException, IOException {
		return this.mVariable.read(a, b);
	}
	
	public List<Attribute> getAttributes() {
		return this.mVariable.getAttributes();
	}
	
	public String toString() {
		if (this.mVariable == null) {
			return "null";
		} else {
			return this.mVariable.toString();
		}
	}
	
	public boolean equals(Object obj) {
		if (!(obj instanceof SGVariable)) {
			return false;
		}
		SGVariable var = (SGVariable) obj;
		return SGUtility.equals(var.mVariable, this.mVariable);
	}

	/**
	 * Returns the name to identify this variable in order of priority:
	 * long_name, standard_name and variable name.
	 * 
	 * @return the name to identify this variable
	 */
	public String getIdentifiableNameInPriorityOrder() {
		return this.getNameInPriorityOrder();
	}
	
	/**
	 * Returns the name of variable.
	 * 
	 * @return the name of variable
	 */
	public String getVariableName() {
		return this.mVariable.getName();
	}
	
	/**
	 * Returns the value type.
	 * 
	 * @return the value type
	 */
	public String getValueType() {
		return SGIDataColumnTypeConstants.VALUE_TYPE_NUMBER;
	}
	
}
