package jp.riken.brain.ni.samuraigraph.figure.java2d;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Area;
import java.awt.geom.Rectangle2D;

import jp.riken.brain.ni.samuraigraph.base.SGColorBarModel;
import jp.riken.brain.ni.samuraigraph.base.SGDrawingElement;
import jp.riken.brain.ni.samuraigraph.base.SGFillPaint;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2f;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.figure.SGIColorMapConstants;

import org.w3c.dom.Element;

/**
 * The color map.
 * 
 * @author kuromaru
 *
 */
public abstract class SGElementGroupColorMap extends SGElementGroupRectangle implements 
        SGIColorMapConstants, SGIElementGroupGridSXY {

    /**
     * A value for rectangle width with related axis.
     */
    protected double mWidthValue;
    
    /**
     * A value for rectangle height with related axis.
     */
    protected double mHeightValue;
    
    /**
     * The type of color map.
     */
    protected SGColorBarModel mColorBarModel = null;
    
    /**
     * The array of x-coordinates.
     */
    protected float[] mXCoordinateArray = null;
    
    /**
     * The array of y-coordinates.
     */
    protected float[] mYCoordinateArray = null;

    /**
     * The array of rectangle width.
     */
    protected float[] mWidthArray = null;
    
    /**
     * The array of rectangle height.
     */
    protected float[] mHeightArray = null;

    /**
     * The array of z-values.
     */
    protected double[] mZValuesArray = null;
    
    /**
     * The bounds of this color map.
     */
    protected Rectangle2D mBounds = new Rectangle2D.Float();

    /**
     * A flag whether the color map is under the grid mode.
     */
    protected boolean mGridMode = false;

    // the margin for single data value
    protected static final float SINGLE_VALUE_MARGIN = 10.0f;

    /**
     * Build this object.
     *
     */
    public SGElementGroupColorMap() {
        super();
    }

    /**
     * Disposes this object.
     * 
     */
    public void dispose() {
    	super.dispose();
    	this.mColorBarModel = null;
        this.mXCoordinateArray = null;
        this.mYCoordinateArray = null;
        this.mZValuesArray = null;
        this.mBounds = null;
    }
    
    /**
     * Sets the color bar model.
     * 
     * @param model
     *           the color bar model
     */
    public void setColorBarModel(SGColorBarModel model) {
        if (model == null) {
            throw new IllegalArgumentException("model == null");
        }
        this.mColorBarModel = model;
    }

    /**
     * Returns the color bar model.
     * 
     * @return the color bar model
     */
    public SGColorBarModel getColorBarModel() {
        return this.mColorBarModel;
    }
    
    /**
     * Returns the value for rectangle width.
     * 
     * @return the value for rectangle width
     */
    public double getWidthValue() {
        return this.mWidthValue;
    }

    /**
     * Sets the value for rectangle width.
     * 
     * @param value
     *           rectangle width to set
     * @return true if succeeded
     */
    public boolean setWidthValue(final double value) {
    	if (value < 0.0) {
    		return false;
    	}
        this.mWidthValue = value;
        return true;
    }

    /**
     * Returns the value for rectangle height.
     * 
     * @return the value for rectangle height
     */
    public double getHeightValue() {
        return this.mHeightValue;
    }

    /**
     * Sets the value for rectangle height.
     * 
     * @param value
     *           rectangle height to set
     * @return true if succeeded
     */
    public boolean setHeightValue(final double value) {
    	if (value < 0.0) {
    		return false;
    	}
        this.mHeightValue = value;
        return true;
    }
    
    /**
     * Sets the location of drawing elements.
     * 
     * @param pointArray
     *           an array of coordinates
     * @return true if succeeded
     */
    public boolean setLocation(final SGTuple2f[] pointArray) {

        if (this.mDrawingElementArray == null) {
            return true;
        }

        if (pointArray.length != this.mDrawingElementArray.length) {
        	this.initDrawingElement(pointArray);
        }

        SGDrawingElement[] array = this.mDrawingElementArray;
        for (int ii = 0; ii < array.length; ii++) {
            ColorMapRectangle rect = (ColorMapRectangle) array[ii];
            final boolean eff = !(pointArray[ii].isInfinite() || pointArray[ii]
                    .isNaN());
            rect.setVisible(eff);
            if (eff) {
                rect.setLocation(pointArray[ii]);
            }
        }

        return true;
    }

    /**
     * Sets the location of drawing elements.
     * 
     * @param xCoordinateArray
     *           an array of x-coordinates
     * @param yCoordinateArray
     *           an array of y-coordinates
     * @return true if succeeded
     */
    public boolean setLocation(float[] xCoordinateArray, float[] yCoordinateArray) {
        if (xCoordinateArray == null || yCoordinateArray == null) {
            throw new IllegalArgumentException(
                    "xCoordinateArray == null || yCoordinateArray == null");
        }
        
        final int xNum = xCoordinateArray.length;
        final int yNum = yCoordinateArray.length;
        
        // get the bounds
        float xMin, xMax, yMin, yMax;
        final float xMinMargin;
        final float xMaxMargin;
        final float yMinMargin;
        final float yMaxMargin;

        if (xNum == 1) {
            xMin = xCoordinateArray[0];
            xMax = xMin;
            xMinMargin = SINGLE_VALUE_MARGIN;
            xMaxMargin = SINGLE_VALUE_MARGIN;
        } else {
            if (xCoordinateArray[0] < xCoordinateArray[xNum - 1]) {
                xMin = xCoordinateArray[0];
                xMax = xCoordinateArray[xNum - 1];
                xMinMargin = Math.abs(xCoordinateArray[1] - xCoordinateArray[0]) / 2.0f;
                xMaxMargin = Math.abs(xCoordinateArray[xNum - 1] - xCoordinateArray[xNum - 2]) / 2.0f;
            } else {
                xMin = xCoordinateArray[xNum - 1];
                xMax = xCoordinateArray[0];
                xMinMargin = Math.abs(xCoordinateArray[xNum - 1] - xCoordinateArray[xNum - 2]) / 2.0f;
                xMaxMargin = Math.abs(xCoordinateArray[1] - xCoordinateArray[0]) / 2.0f;
            }
        }
        
        if (yNum == 1) {
            yMin = yCoordinateArray[0];
            yMax = yMin;
            yMinMargin = SINGLE_VALUE_MARGIN;
            yMaxMargin = SINGLE_VALUE_MARGIN;
        } else {
            if (yCoordinateArray[0] < yCoordinateArray[yNum - 1]) {
                yMin = yCoordinateArray[0];
                yMax = yCoordinateArray[yNum - 1];
                yMinMargin = Math.abs(yCoordinateArray[1] - yCoordinateArray[0]) / 2.0f;
                yMaxMargin = Math.abs(yCoordinateArray[yNum - 1] - yCoordinateArray[yNum - 2]) / 2.0f;
            } else {
                yMin = yCoordinateArray[yNum - 1];
                yMax = yCoordinateArray[0];
                yMinMargin = Math.abs(yCoordinateArray[yNum - 1] - yCoordinateArray[yNum - 2]) / 2.0f;
                yMaxMargin = Math.abs(yCoordinateArray[1] - yCoordinateArray[0]) / 2.0f;
            }
        }
        
        xMin -= xMinMargin;
        xMax += xMaxMargin;
        yMin -= yMinMargin;
        yMax += yMaxMargin;
        final float w = xMax - xMin;
        final float h = yMax - yMin;
        this.mBounds = new Rectangle2D.Float(xMin, yMin, w, h);
        
        // calculate the width array and the height array
        final float[] widthArray = new float[xNum];
        for (int xx = 0; xx < xNum; xx++) {
            final float width;
            if (xx == 0) {
                width = Math.abs(xCoordinateArray[1] - xCoordinateArray[0]);
            } else if (xx == xNum - 1) {
                width = Math.abs(xCoordinateArray[xNum - 1] - xCoordinateArray[xNum - 2]);
            } else {
                width = Math.abs(xCoordinateArray[xx + 1] - xCoordinateArray[xx - 1]) / 2.0f;
            }
            widthArray[xx] = width;
        }
        
        final float[] heightArray = new float[yNum];
        for (int yy = 0; yy < yNum; yy++) {
            final float height;
            if (yy == 0) {
                height = Math.abs(yCoordinateArray[1] - yCoordinateArray[0]);
            } else if (yy == yNum - 1) {
                height = Math.abs(yCoordinateArray[yNum - 1] - yCoordinateArray[yNum - 2]);
            } else {
                height = Math.abs(yCoordinateArray[yy + 1] - yCoordinateArray[yy - 1]) / 2.0f;
            }
            heightArray[yy] = height;
        }

        // set to the attributes
        this.mXCoordinateArray = xCoordinateArray;
        this.mYCoordinateArray = yCoordinateArray;
        this.mWidthArray = widthArray;
        this.mHeightArray = heightArray;

        return true;
    }

    /**
     * Returns the tag name.
     * 
     * @return the tag name
     */
    public String getTagName() {
        return TAG_NAME_COLOR_MAP;
    }

    /**
     * Paint the rectangles.
     * 
     * @param g2d
     *          graphic context
     * @param clipRect
     *          a rectangle for clipping
     * @return true if succeeded
     */
    public boolean paintElement(Graphics2D g2d, Rectangle2D clipRect) {
    	
    	if (this.mGridMode) {
            Area clipArea = null;
            if (clipRect != null) {
                clipArea = new Area(clipRect);
            }
            final int xNum = this.mXCoordinateArray.length;
            final int yNum = this.mYCoordinateArray.length;
            
            for (int yy = 0; yy < yNum; yy++) {
                final float yCur = this.mYCoordinateArray[yy] - this.mHeightArray[yy] / 2.0f;
                for (int xx = 0; xx < xNum; xx++) {
                    final int index = xNum * yy + xx;
                    final double value = this.mZValuesArray[index];
                    if (Double.isNaN(value) || Double.isInfinite(value)) {
                        continue;
                    }
                    final float xCur = this.mXCoordinateArray[xx] - this.mWidthArray[xx] / 2.0f;
                    final Rectangle2D dRect = new Rectangle2D.Double(xCur, yCur, 
                            this.mWidthArray[xx], this.mHeightArray[yy]);

                    // paint
                    Color cl = this.mColorBarModel.evaluate(value);
                    g2d.setPaint(cl);

                    if (clipArea != null) {
                        Area inner = new Area(dRect);
                        inner.intersect(clipArea);
                        g2d.fill(inner);
                    } else {
                        g2d.fill(dRect);
                    }
                }
            }
    	} else {
            SGDrawingElement[] array = this.mDrawingElementArray;
            if (array != null) {
                for (int ii = 0; ii < array.length; ii++) {
                    ColorMapRectangle el = (ColorMapRectangle) array[ii];
                    el.paint(g2d, clipRect);
                }
            }
    	}
    	
        return true;
    }

    /**
     * Initializes the drawing elements.
     * 
     * @param array
     *           the location of drawing elements
     */
    protected boolean initDrawingElement(SGTuple2f[] array) {
//    	if (this.mGridMode) {
//            // do nothing
//    	} else {
//    		if (super.initDrawingElement(array) == false) {
//    			return false;
//    		}
//    	}
		if (super.initDrawingElement(array) == false) {
			return false;
		}
        return true;
    }

    /**
     * Initializes the drawing elements.
     * 
     * @param array
     *           the location of drawing elements
     */
    public boolean initDrawingElement(float[] xArray, float[] yArray) {
        // do nothing
        return true;
    }

    /**
     * Creates and returns an instance of a drawing element.
     */
    protected SGDrawingElement createDrawingElementInstance(final int index) {
//    	if (this.mGridMode) {
//    		return null;
//    	} else {
//            return new ColorMapRectangle(this, index);
//    	}
        return new ColorMapRectangle(this, index);
    }

    /**
     * Read properties from the Element object.
     * 
     * @param el
     *          the Element object
     */
    public boolean readProperty(Element el) {
    	
    	if (this.mGridMode) {
    		// do nothing
    	} else {
            String str = null;
            Number num = null;

            // width of the rectangle
            str = el.getAttribute(KEY_RECTANGLE_WIDTH_VALUE);
            if (str.length() != 0) {
                num = SGUtilityText.getDouble(str);
                if (num == null) {
                    return false;
                }
                this.setWidthValue(num.doubleValue());
            }
            
            // height of the rectangle
            str = el.getAttribute(KEY_RECTANGLE_HEIGHT_VALUE);
            if (str.length() != 0) {
                num = SGUtilityText.getDouble(str);
                if (num == null) {
                    return false;
                }
                this.setHeightValue(num.doubleValue());
            }
    	}
    	
        return true;
    }

    /**
     * Write properties to the Element object.
     * 
     * @param el
     *          the Element object
     */
    public boolean writeProperty(Element el) {
    	if (this.mGridMode) {
    		// do nothing
    	} else {
            el.setAttribute(KEY_RECTANGLE_WIDTH_VALUE, Double.toString(this.mWidthValue));
            el.setAttribute(KEY_RECTANGLE_HEIGHT_VALUE, Double.toString(this.mHeightValue));
    	}
        return true;
    }

    
    /**
     * A class of rectangle for the color map.
     * @author kuromaru
     *
     */
    protected static class ColorMapRectangle extends RectInGroup implements
        SGIDrawingElementJava2D {

        /**
         * Z-value.
         */
        private double mZValue;
        
        /**
         * Builds a rectangle in a color map.
         * 
         * @param group
         *           a group of rectangles
         */
        public ColorMapRectangle(final SGElementGroupColorMap group, final int index) {
            super(group, index);
        }

        /**
         * Sets the z-value.
         * 
         * @param value
         *           a value to set
         */
        public void setZValue(final double value) {
            this.mZValue = value;
        }
        
        /**
         * Returns the z-value.
         * 
         * @return the z-value
         */
        public double getZValue() {
            return this.mZValue;
        }
        
        /**
         * Paint this object.
         * 
         * @param g2d
         *            graphics context
         */
        public void paint(final Graphics2D g2d) {
            if (this.isVisible() == false) {
                return;
            }
            final Rectangle2D dRect = this.getElementBounds();
            Color cl = this.getColor();
            g2d.setPaint(cl);
            g2d.fill(dRect);
            
//            g2d.setStroke(new BasicStroke(1.0f));
//            g2d.setPaint(Color.BLACK);
//            g2d.draw(dRect);
        }

        public boolean contains(int x, int y) {
            return this.getElementBounds().contains(x, y);
        }

        public Rectangle2D getElementBounds() {
            final float x = this.getX();
            final float y = this.getY();
            final float w = this.getWidth();
            final float h = this.getHeight();

            final float ww = Math.abs(w);
            final float hh = Math.abs(h);

            final float xx = x - 0.50f * ww;
            final float yy = y - 0.50f * hh;

            Rectangle2D rect = new Rectangle2D.Float(xx, yy, ww, hh);
            return rect;
        }

        public void paint(Graphics2D g2d, Rectangle2D clipRect) {
            if (clipRect == null) {
                this.paint(g2d);
            } else {
                final Rectangle2D dRect = this.getElementBounds();
                Area clipArea = new Area(clipRect);
                Area inner = new Area(dRect);

                inner.intersect(clipArea);

                // paint
                g2d.setPaint(this.getColor());
                g2d.fill(inner);
            }
        }

        public Color getColor() {
            SGColorBarModel model = ((SGElementGroupColorMap) this.mGroup).getColorBarModel();
            return model.evaluate(this.mZValue);
        }
    }
    
    /**
     * Properties of the color map.
     * 
     * @author kuromaru
     *
     */
    public static class ColorMapProperties extends ElementGroupProperties {
        double widthValue;
        double heightValue;
        ColorMapProperties() {
            super();
        }
        public Object copy() {
            Object obj = super.copy();
            ColorMapProperties p = (ColorMapProperties) obj;
            p.widthValue = this.widthValue;
            p.heightValue = this.heightValue;
            return p;
        }
        public boolean equals(final Object obj) {
            if ((obj instanceof ColorMapProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }
            ColorMapProperties p = (ColorMapProperties) obj;
            if (this.widthValue != p.widthValue) {
                return false;
            }
            if (this.heightValue != p.heightValue) {
                return false;
            }
            return true;
        }
    }


    private boolean mFocusedFlag = false;

    public void setFocused(final boolean b) {
        this.mFocusedFlag = b;
    }

    public boolean isFocused() {
        return this.mFocusedFlag;
    }

    private SGElementGroupSetInGraph mGroupSet = null;

    public boolean setElementGroupSet(SGElementGroupSetInGraph gs) {
        this.mGroupSet = gs;
        return true;
    }

    /**
     * Sets the color to each rectangle.
     * 
     * @param values
     *           values for colors
     */
    public void setColors(final double[] values) {
        if (values == null) {
            throw new IllegalArgumentException("values == null");
        }
        if (values.length != this.mDrawingElementArray.length) {
            throw new IllegalArgumentException("Illegal array length: " + values.length 
                    + " != " + this.mDrawingElementArray.length);
        }
        for (int ii = 0; ii < values.length; ii++) {
            ColorMapRectangle rect = (ColorMapRectangle) this.mDrawingElementArray[ii];
            rect.setZValue(values[ii]);
            if (Double.isNaN(values[ii]) || Double.isInfinite(values[ii])) {
                rect.setVisible(false);
            } else {
                rect.setVisible(true);
                Color cl = this.mColorBarModel.evaluate(values[ii]);
                rect.setInnerPaint(new SGFillPaint(cl));
            }
        }
    }
    
    /**
     * Sets the value range.
     * 
     * @param min
     *           the minimum value
     * @param max
     *           the maximum value
     */
    public void setValueRange(final double min, final double max) {
        this.mColorBarModel.setValueRange(min, max);
    }
    
    public SGProperties getProperties() {
        ColorMapProperties p = new ColorMapProperties();
        if (this.getProperties(p) == false) {
            return null;
        }
        return p;
    }
    
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof ColorMapProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }
        ColorMapProperties cp = (ColorMapProperties) p;
        cp.widthValue = this.mWidthValue;
        cp.heightValue = this.mHeightValue;
        return true;
    }
    
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof ColorMapProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }
        ColorMapProperties cp = (ColorMapProperties) p;
        this.mWidthValue = cp.widthValue;
        this.mHeightValue = cp.heightValue;
        return true;
    }
    
    /**
     * Sets the z-values to the rectangles.
     * 
     * @param zValues
     *          an array of z-values
     * @return true if succeeded
     */
    protected boolean setZValues(final double[] zValues) {
        if (zValues == null) {
            throw new IllegalArgumentException("zValues == null");
        }
        if (this.mGridMode) {
            this.mZValuesArray = zValues;
        } else {
            if (zValues.length != this.mDrawingElementArray.length) {
                throw new IllegalArgumentException(
                        "The number of z values is different from that of the rectangles: "
                                + this.mDrawingElementArray.length + " != "
                                + zValues.length);
            }
            for (int ii = 0; ii < this.mDrawingElementArray.length; ii++) {
                ColorMapRectangle rect = (ColorMapRectangle) this.mDrawingElementArray[ii];
                rect.setZValue(zValues[ii]);
            }
        }
        return true;
    }

    /**
     * Sets the edge line width in a given unit.
     * 
     * @param lw
     *           the edge line width to set
     * @param unit
     *           the unit for given line width
     * @return true if succeeded
     */
    public boolean setEdgeLineWidth(float lw, String unit) {
        // do nothing
        return true;
    }

    /**
     * Returns whether this color map is under the grid mode.
     * 
     * @return true if this color map is under the grid mode
     */
    public boolean isGridMode() {
    	return this.mGridMode;
    }

    /**
     * Sets the grid mode.
     * 
     * @param b
     *          true to set the grid mode
     * @return true if succeeded
     */
    public boolean setGridMode(final boolean b) {
    	this.mGridMode = b;
    	return true;
    }
    
    /**
     * Returns the x-coordinate of the rectangle at given index.
     * 
     * @param index
     *           the array index
     * @return the x-coordinate
     */
    public float getX(final int index) {
        return this.mXCoordinateArray[index];
    }

    /**
     * Returns the y-coordinate of the rectangle at given index.
     * 
     * @param index
     *           the array index
     * @return the y-coordinate
     */
    public float getY(int index) {
        return this.mYCoordinateArray[index];
    }
}
