package jp.riken.brain.ni.samuraigraph.figure.java2d;

import java.awt.Color;
import java.awt.Graphics2D;
import java.awt.geom.Rectangle2D;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import jp.riken.brain.ni.samuraigraph.base.SGAxis;
import jp.riken.brain.ni.samuraigraph.base.SGData;
import jp.riken.brain.ni.samuraigraph.base.SGDataColumnInfo;
import jp.riken.brain.ni.samuraigraph.base.SGIData;
import jp.riken.brain.ni.samuraigraph.base.SGIFigureElementGraph;
import jp.riken.brain.ni.samuraigraph.base.SGProperties;
import jp.riken.brain.ni.samuraigraph.base.SGPropertyMap;
import jp.riken.brain.ni.samuraigraph.base.SGPropertyResults;
import jp.riken.brain.ni.samuraigraph.base.SGTuple2f;
import jp.riken.brain.ni.samuraigraph.base.SGUtility;
import jp.riken.brain.ni.samuraigraph.base.SGUtilityText;
import jp.riken.brain.ni.samuraigraph.data.SGDataUtility;
import jp.riken.brain.ni.samuraigraph.data.SGIVXYTypeData;
import jp.riken.brain.ni.samuraigraph.data.SGNetCDFData;
import jp.riken.brain.ni.samuraigraph.figure.SGElementGroup;
import jp.riken.brain.ni.samuraigraph.figure.SGIArrowConstants;
import jp.riken.brain.ni.samuraigraph.figure.SGUtilityForFigureElement;

import org.w3c.dom.Element;

/**
 * @author okumura
 */
public class SGElementGroupSetInGraphVXY extends SGElementGroupSetInGraph
        implements SGIElementGroupSetVXY, SGIVXYDataDialogObserver, SGIArrowConstants {

	/**
	 * The array of coordinate of the start points.
     * These values are used only in the non-grid mode.
	 */
	protected SGTuple2f[] mStartPointsArray = null;

    /**
     * The array of x-coordinates.
     * These values are used only in the grid mode.
     */
    protected float[] mXArray = null;
    
    /**
     * The array of y-coordinates.
     * These values are used only in the grid mode.
     */
    protected float[] mYArray = null;

	/**
	 * The array of coordinate of the end points.
     * These values are used in both of the grid mode and non-grid mode.
	 */
    protected SGTuple2f[] mEndPointsArray = null;

    /**
     * The scaling factor for the magnitude of vectors.
     */
    private float mMagnitudePerCM = 1.0f;

    /**
     * Builds a group set object.
     * 
     * @param data
     *           the data
     * @param graph
     *           the graph that this object belongs.
     */
    protected SGElementGroupSetInGraphVXY(SGData data, SGFigureElementGraph graph) {
        super(data, graph);
    }

    /**
     * Dispose of this object.
     */
    public void dispose() {
        super.dispose();
        this.mStartPointsArray = null;
        this.mEndPointsArray = null;
        this.mXArray = null;
        this.mYArray = null;
    }

    // initialize the points
    protected void initPointsArray(final int num) {
        SGTuple2f[] sArray = new SGTuple2f[num];
        SGTuple2f[] eArray = new SGTuple2f[num];
        for (int ii = 0; ii < num; ii++) {
            sArray[ii] = new SGTuple2f();
            eArray[ii] = new SGTuple2f();
        }
        this.mStartPointsArray = sArray;
        this.mEndPointsArray = eArray;
    }

    /**
     * Initialize the points.
     * 
     * @param numX
     *          the number of x-coordinates
     * @param numY
     *          the number of y-coordinates
     */
    protected void initPointsArray(final int numX, final int numY) {
        this.mXArray = new float[numX];
        this.mYArray = new float[numY];
        final int num = numX * numY;
        this.mEndPointsArray = new SGTuple2f[num];
        for (int ii = 0; ii < num; ii++) {
            this.mEndPointsArray[ii] = new SGTuple2f();
        }
    }

    /**
     * Returns the magnitude of vectors per centimeter.
     * 
     * @return the magnitude of vectors per centimeter
     */
    public float getMagnitudePerCM() {
        return this.mMagnitudePerCM;
    }

    /**
     * Sets the magnitude of vectors per centimeter.
     * 
     * @param mag
     *           the magnitude of vectors per centimeter
     * @return true if succeeded
     */
    public boolean setMagnitudePerCM(final float mag) {
        if (mag <= 0.0) {
            return false;
        }
        SGIVXYTypeData vData = (SGIVXYTypeData) this.mData;
        this.mMagnitudePerCM = SGUtilityForFigureElement.roundMagnitudePerCM(mag, vData);
        this.updateDrawingElementsLocation(this.mGraph.getData(this));
        return true;
    }

    /**
     * 
     * @return
     */
    public String getClassDescription() {
        return "";
    }

    protected boolean createDrawingElements() {
        return true;
    }

    public boolean addDrawingElementGroup(final int type) {
    	SGElementGroupArrow group = null;
        if (type == SGIElementGroupConstants.ARROW_GROUP) {
            group = new SGElementGroupArrowInGraph(this.mGraph);
            group.setGridMode(this.isGridMode());
        } else {
            return false;
        }

        // add a group
        this.addElementGroup(group);

        return true;
    }

    private boolean addElementGroup(final SGElementGroupArrow group) {
        // set the group set
        SGIElementGroupInGraph iGroup = (SGIElementGroupInGraph) group;
        iGroup.setElementGroupSet(this);

        // initialize the drawing element
        if (group.initDrawingElement(this.mStartPointsArray,
                this.mEndPointsArray) == false) {
            throw new Error();
        }
        
        // add the new group to the list
        this.mDrawingElementGroupList.add(group);

        return true;
    }

    /**
     * A flag for the invariance.
     */
    private boolean mDirectionFixedFlag;

    /**
     * Returns whether the direction is invariant when the scale of the x-axis direction 
     * and the y-axis direction is different.
     * 
     * @return a flag for the invariance
     */
    public boolean isDirectionInvariant() {
        return this.mDirectionFixedFlag;
    }

    /**
     * Sets whether the direction is invariant when the scale of the x-axis direction 
     * and the y-axis direction is set to be different.
     * 
     * @param b
     *          a flag to set to the invariance
     * @return true if succeeded
     */
    public boolean setDirectionInvariant(final boolean b) {
        this.mDirectionFixedFlag = b;
        return true;
    }

    /**
     * Called when the location of data points have changed.
     */
    public boolean updateDrawingElementsLocation(final SGIData data) {
        if ((data instanceof SGIVXYTypeData) == false) {
            return false;
        }
        
        SGIVXYTypeData dataVXY = (SGIVXYTypeData) data;

        if (this.isGridMode()) {
        	
        	// get x and y values
            final double[] xValues = dataVXY.getXCoordinateArray();
            final double[] yValues = dataVXY.getYCoordinateArray();

            // initialize the points
            this.initPointsArray(xValues.length, yValues.length);
    		
            // set the location of the data points
            if (this.mGraph.calcLocationOfPoints(xValues,
                    this.getXAxis(), true, this.mXArray) == false) {
                return false;
            }
            if (this.mGraph.calcLocationOfPoints(yValues, 
                    this.getYAxis(), false, this.mYArray) == false) {
                return false;
            }

            List<SGElementGroup> gList = this.mDrawingElementGroupList;
            for (int ii = 0; ii < gList.size(); ii++) {
                SGElementGroup group = (SGElementGroup) gList.get(ii);
                if (group.isVisible()) {
                    if (group instanceof SGElementGroupArrow) {
                    	SGElementGroupArrow arrows = (SGElementGroupArrow) group;
                        if (arrows.setLocation(this.mXArray, this.mYArray) == false) {
                            return false;
                        }
                    }
                }
            }
            
        } else {
        	
        	// get the number of the points
            final int num = dataVXY.getPointsNumber();
            
            // initialize the points
            this.initPointsArray(num);

            // set the location of the start points
            SGAxis xAxis = this.getXAxis();
            SGAxis yAxis = this.getYAxis();
            if (this.mGraph.calcLocationOfPoints(dataVXY.getCoordinateArray(), xAxis, yAxis,
                    this.mStartPointsArray) == false) {
                return false;
            }

            // calculate the location of the end points and set to the groups
            if (SGArrowUtility.calcEndLocation(dataVXY, this.mGraph, this, 
            		this.getMagnitudePerCM(), this.isDirectionInvariant(), null, null,
            		this.mStartPointsArray, this.mEndPointsArray) == false) {
            	return false;
            }
            
            for (int ii = 0; ii < this.mDrawingElementGroupList.size(); ii++) {
                SGElementGroup group = (SGElementGroup) this.mDrawingElementGroupList.get(ii);
                if (group.isVisible()) {
                    if (group instanceof SGElementGroupArrowInGraph) {
                        SGElementGroupArrowInGraph vGroup = (SGElementGroupArrowInGraph) group;
                        vGroup.setLocation(this.mStartPointsArray, this.mEndPointsArray);
                    }
                }
            }
        }
        
        return true;
    }

    public void paintGraphics2D(final Graphics2D g2d) {
        Rectangle2D gRect = null;
        if (!this.getClipFlag()) {
            gRect = this.mGraph.getGraphRect();
        }

        List<SGElementGroup> list = this.mDrawingElementGroupList;
        for (int ii = 0; ii < list.size(); ii++) {
            final SGElementGroupArrowInGraph group = (SGElementGroupArrowInGraph) list
                    .get(ii);
            if (group.isVisible()) {
                group.paintElement(g2d, gRect);
            }
        }

    }

    /**
     * Returns whether the legend is visible.
     */
    public boolean getLegendVisibleFlag() {
        return this.isVisibleInLegend();
    }

    /**
     * Returns the location of x-axis.
     */
    public int getXAxisLocation() {
        return this.mGraph.getAxisElement().getLocationInPlane(this.getXAxis());
    }

    /**
     * Returns the location of y-axis.
     */
    public int getYAxisLocation() {
        return this.mGraph.getAxisElement().getLocationInPlane(this.getYAxis());
    }

    public boolean prepare() {
        this.mTemporaryProperties = this.getProperties();
        return true;
    }

    /**
     * Write properties of this object to the Element.
     * 
     * @param el
     *            the Element object
     * @param type
     *            type of the method to save properties
     * @return true if succeeded
     */
    public boolean writeProperty(final Element el, final int type) {
        if (super.writeProperty(el, type) == false) {
            return false;
        }

        final String strX = this.mGraph.getAxisElement().getLocationName(
                this.getXAxis());
        final String strY = this.mGraph.getAxisElement().getLocationName(
                this.getYAxis());

        el.setAttribute(SGIFigureElementGraph.KEY_X_AXIS_POSITION, strX);
        el.setAttribute(SGIFigureElementGraph.KEY_Y_AXIS_POSITION, strY);

        // write polar flag
        SGIVXYTypeData data = (SGIVXYTypeData) this.mGraph.getData(this);
        final boolean polar = data.isPolar();
        el.setAttribute(SGIFigureElementGraph.KEY_POLAR, Boolean
                .toString(polar));

        // write magnification
        final float factor = this.getMagnitudePerCM();
        el.setAttribute(SGIFigureElementGraph.KEY_MAGNITUDE_PER_CM, Float
                .toString(factor));

        final boolean invariant = this.isDirectionInvariant();
        el.setAttribute(SGIFigureElementGraph.KEY_DIRECTION_INVARIANT, Boolean
                .toString(invariant));

        return true;
    }

    /**
     * Returns the properties of this object.
     */
    public SGProperties getProperties() {
        ElementGroupSetInVXYGraphProperties ep = new ElementGroupSetInVXYGraphProperties();
        if (this.getProperties(ep) == false) {
            return null;
        }
        return ep;
    }

    /**
     * Get the properties of this object.
     */
    public boolean getProperties(final SGProperties p) {
        if ((p instanceof ElementGroupSetInVXYGraphProperties) == false) {
            return false;
        }
        if (super.getProperties(p) == false) {
            return false;
        }

        ElementGroupSetInVXYGraphProperties ep = (ElementGroupSetInVXYGraphProperties) p;
        ep.mMagnitudeScalingFactor = this.getMagnitudePerCM();
        ep.mDirectionInvariant = this.isDirectionInvariant();
        
        return true;
    }

    /**
     * Set properties to this object.
     */
    public boolean setProperties(final SGProperties p) {
        if ((p instanceof ElementGroupSetInVXYGraphProperties) == false) {
            return false;
        }
        if (super.setProperties(p) == false) {
            return false;
        }

        ElementGroupSetInVXYGraphProperties ep = (ElementGroupSetInVXYGraphProperties) p;
        this.setMagnitudePerCM(ep.mMagnitudeScalingFactor);
        this.setDirectionInvariant(ep.mDirectionInvariant);
        
        this.updateGridMode();
        
        return true;
    }

    public float getLineWidth(final String unit) {
        return this.getArrowGroup().getLineWidth(unit);
    }

    public int getLineType() {
        return this.getArrowGroup().getLineType();
    }

    public Color getColor() {
        return this.getArrowGroup().getColor();
    }

    public float getHeadSize(final String unit) {
        return this.getArrowGroup().getHeadSize(unit);
    }

    public float getHeadOpenAngle() {
        return this.getArrowGroup().getHeadOpenAngle();
    }

    public float getHeadCloseAngle() {
        return this.getArrowGroup().getHeadCloseAngle();
    }

    public int getStartHeadType() {
        return this.getArrowGroup().getStartHeadType();
    }

    public int getEndHeadType() {
        return this.getArrowGroup().getEndHeadType();
    }

    public boolean setLineWidth(final float lw, final String unit) {
        return this.getArrowGroup().setLineWidth(lw, unit);
    }

    public boolean setLineType(final int type) {
        return this.getArrowGroup().setLineType(type);
    }

    public boolean setColor(final Color cl) {
        return this.getArrowGroup().setColor(cl);
    }

    public boolean setHeadSize(final float size, final String unit) {
        return this.getArrowGroup().setHeadSize(size, unit);
    }

    public boolean setHeadAngle(final Float openAngle, final Float closeAngle) {
        return this.getArrowGroup().setHeadAngle(openAngle, closeAngle);
    }

    public boolean setStartHeadType(final int type) {
        return this.getArrowGroup().setStartHeadType(type);
    }

    public boolean setEndHeadType(final int type) {
        return this.getArrowGroup().setEndHeadType(type);
    }

    public boolean hasValidAngle(final Number open, final Number close) {
        final float openAngle = (open != null) ? open.floatValue() : this
                .getHeadOpenAngle();
        final float closeAngle = (close != null) ? close.floatValue() : this
                .getHeadCloseAngle();
        return (openAngle < closeAngle);
    }

    /**
     * A class of properties of a group set in VXY graph element.
     */
    public static class ElementGroupSetInVXYGraphProperties extends
            ElementGroupSetPropertiesInFigureElement {

        float mMagnitudeScalingFactor;

        boolean mDirectionInvariant;
        
        /**
         * The default constructor.
         */
        public ElementGroupSetInVXYGraphProperties() {
            super();
        }

        /**
         * Returns whether this object is equal to given object.
         * 
         * @param obj
         *            an object to be compared
         * @return true if two objects are equal
         */
        public boolean equals(final Object obj) {
            if ((obj instanceof ElementGroupSetInVXYGraphProperties) == false) {
                return false;
            }
            if (super.equals(obj) == false) {
                return false;
            }

            ElementGroupSetInVXYGraphProperties p = (ElementGroupSetInVXYGraphProperties) obj;
            if (p.mMagnitudeScalingFactor != this.mMagnitudeScalingFactor) {
                return false;
            }
            if (p.mDirectionInvariant != this.mDirectionInvariant) {
                return false;
            }

            return true;
        }

        /**
         * Returns string representation of this object.
         * 
         * @return string representation of this object
         */
        public String toString() {
            String str = super.toString();
            StringBuffer sb = new StringBuffer();
            sb.append("[");
            sb.append(str);
            sb.append(", Magnitude Scaling Factor=");
            sb.append(this.mMagnitudeScalingFactor);
            sb.append(", Direction Invariant=");
            sb.append(this.mDirectionInvariant);
            sb.append("]");
            return sb.toString();
        }

    }

    /**
     * Retruns the location of the start point at a given index.
     * 
     * @param index
     *           the index
     * @return the location
     */
    public SGTuple2f getStartLocation(final int index) {
    	if (this.isGridMode()) {
            if (this.mXArray == null || this.mYArray == null) {
                return null;
            }
            final int yIndex = index / this.mXArray.length;
            final int xIndex = index % this.mXArray.length;
            if (yIndex < 0 || yIndex >= this.mYArray.length) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
            return new SGTuple2f(this.mXArray[xIndex], this.mYArray[yIndex]);
    	} else {
            if (this.mStartPointsArray == null) {
                return null;
            }
            if (index < 0 || index >= this.mStartPointsArray.length) {
                throw new IllegalArgumentException("Index out of bounds: " + index);
            }
            return this.mStartPointsArray[index];
    	}
    }

    /**
     * Retruns the location of the end point at a given index.
     * 
     * @param index
     *           the index
     * @return the location
     */
    public SGTuple2f getEndLocation(final int index) {
        if (this.mEndPointsArray == null) {
            return null;
        }
        if (index < 0 || index >= this.mEndPointsArray.length) {
            throw new IllegalArgumentException("Index out of bounds: " + index);
        }
        return this.mEndPointsArray[index];
    }

    protected SGPropertyResults setProperties(SGPropertyMap map,
            SGPropertyResults iResult) {

        SGPropertyResults result = (SGPropertyResults) iResult.clone();
        result = super.setProperties(map, result);
        if (result == null) {
            return null;
        }
        
        Iterator<String> itr = map.getKeyIterator();
        while (itr.hasNext()) {
            String key = itr.next();
            String value = map.getValueString(key);
            
            if (COM_DATA_ARROW_MAGNITUDE_PER_CM.equalsIgnoreCase(key)) {
                Float mag = SGUtilityText.getFloat(value);
                if (mag == null) {
                    result.putResult(COM_DATA_ARROW_MAGNITUDE_PER_CM, SGPropertyResults.INVALID_INPUT_VALUE);
                    continue;
                }
                if (SGUtility.isValidPropertyValue(mag.floatValue()) == false) {
                    result.putResult(COM_DATA_ARROW_MAGNITUDE_PER_CM, SGPropertyResults.INVALID_INPUT_VALUE);
                    continue;
                }
                if (this.setMagnitudePerCM(mag.floatValue()) == false) {
                    result.putResult(COM_DATA_ARROW_MAGNITUDE_PER_CM, SGPropertyResults.INVALID_INPUT_VALUE);
                    continue;
                }
                result.putResult(COM_DATA_ARROW_MAGNITUDE_PER_CM, SGPropertyResults.SUCCEEDED);
            } else if (COM_DATA_ARROW_DIRECTION_INVARIANT.equalsIgnoreCase(key)) {
                Boolean b = SGUtilityText.getBoolean(value);
                if (b == null) {
                    result.putResult(COM_DATA_ARROW_DIRECTION_INVARIANT, SGPropertyResults.INVALID_INPUT_VALUE);
                    continue;
                }
                if (this.setDirectionInvariant(b.booleanValue()) == false) {
                    result.putResult(COM_DATA_ARROW_DIRECTION_INVARIANT, SGPropertyResults.INVALID_INPUT_VALUE);
                    continue;
                }
                result.putResult(COM_DATA_ARROW_DIRECTION_INVARIANT, SGPropertyResults.SUCCEEDED);
            }
        }

        List<SGElementGroupArrow> arrowGroups = this.getArrowGroups();
        for (int ii = 0; ii < arrowGroups.size(); ii++) {
            SGElementGroupArrowForData arrows = (SGElementGroupArrowForData) arrowGroups.get(ii);
            result = arrows.setProperties(map, result);
            if (result == null) {
                return null;
            }
        }
        
        return result;
    }

	/**
	 * Returns a list of arrow groups.
	 * 
	 * @return a list of arrow groups
	 */
	public List<SGElementGroupArrow> getArrowGroups() {
		List<SGElementGroupArrow> retList = new ArrayList<SGElementGroupArrow>();
		List<SGElementGroup> list = SGUtilityForFigureElement.getGroups(
				SGElementGroupArrow.class, this.mDrawingElementGroupList);
		for (int ii = 0; ii < list.size(); ii++) {
			retList.add((SGElementGroupArrow) list.get(ii));
		}
		return retList;
	}

	/**
	 * Returns an arrow group which is the first element of an array.
	 * 
	 * @return the first element of an array of arrow groups, or null when this
	 *         group set does not have any arrow groups
	 */
	public SGElementGroupArrow getArrowGroup() {
		return (SGElementGroupArrow) SGUtilityForFigureElement.getGroup(
				SGElementGroupArrow.class, this.mDrawingElementGroupList);
	}

    /**
     * Sets the properties of element groups.
     * 
     * @param elementGroupPropertiesList
     * @return true if succeeded
     * 
     */
    protected boolean setElementGroupProperties(List elementGroupPropertiesList) {
        for (int ii = 0; ii < elementGroupPropertiesList.size(); ii++) {
            SGProperties gp = (SGProperties) elementGroupPropertiesList.get(ii);
            SGElementGroup group = null;
            if (gp instanceof SGElementGroupArrow.ArrowProperties) {
                group = this.getArrowGroup();
            } else {
                throw new Error("Illegal group property: " + gp);
            }
            if (group == null) {
                continue;
            }
            if (group.setProperties(gp) == false) {
                return false;
            }
        }
        return true;
    }

    /**
     * Returns true if this group set is under the grid mode.
     * 
     * @return true if this group set is under the grid mode
     */
    public boolean isGridMode() {
    	return this.isGridMode(this.getData());
    }

    /**
     * Returns true if a given data is under the grid mode.
     * 
     * @param data
     *           the data
     * @return true if a given data is under the grid mode
     */
    public boolean isGridMode(SGData data) {
    	if (SGDataUtility.isNetCDFData(data)) {
    		SGNetCDFData nData = (SGNetCDFData) data;
    		return !nData.isSerialNumberAvailable();
    	} else {
    		return false;
    	}
    }
    
    private boolean updateGridMode() {
    	final boolean b = this.isGridMode();
    	List<SGElementGroupArrow> groupList = this.getArrowGroups();
    	for (int ii = 0; ii < groupList.size(); ii++) {
    		SGElementGroupArrow group = groupList.get(ii);
    		if (group.setGridMode(b) == false) {
    			return false;
    		}
    	}
    	return true;
    }

    /**
     * Sets the information of data columns.
     * @param columns
     *                information of data columns
     */
    public boolean setColumnInfo(SGDataColumnInfo[] columns, String message) {
    	if (super.setColumnInfo(columns, message) == false) {
    		return false;
    	}

    	// initialize the points
    	SGIVXYTypeData data = (SGIVXYTypeData) this.getData();
    	this.initPointsArray(data.getPointsNumber());
    	
    	// update the grid mode
    	if (this.updateGridMode() == false) {
    		return false;
    	}
    	
    	return true;
    }

}
